/*****************************************************************************/
//    ContextNode.h
/*****************************************************************************/

#ifndef _ContextNode_H_
#define _ContextNode_H_

#include "Context.h"
#include "GoalNode.h"
#include "GoalNodeLink.h"

/*****************************************************************************/
//    Codependent classes

class corContext;

/*****************************************************************************/

class corContextNode {
public:
    corGoalNode *goals;

    // unmarked goals as a singly linked list
    corGoalNodeLink *first_unmarked_goal;    
    corGoalNodeLink *last_unmarked_goal;

    // marked goals as a singly linked list
    corGoalNodeLink *first_marked_goal;
    corGoalNodeLink *last_marked_goal;

    int mark;            // if any goal is marked, so is the node
    int node_number;    // used for linear ordering of nodes

    int negated;    
    // CHECK IF NEEDED??
    // if any goal is negated, so is the node

    corContextNode *next;    // link to next node in context
    corContextNode *prev;    // link to previous node in context
    corContextNode *next_marked;    // link to next marked node
    corContextNode *prev_marked;    // link to previous marked node

    corContext *context;    // the context to which this belongs to 

    corContextNode(corGoalNode *new_goal, corContext *c);
    
    inline int is_marked()        { return mark; }
    inline void mark_node()        { mark = 1; }

    int any_unmarked_goals();
    corGoalNode *get_unmarked_goal();

    void move_goals_to_done();

    void move_to_marked_list(corGoalNodeLink *g_ptr);
    void merge_nodes(corContextNode *tbm);

    void update_goalnode_info();
    void print_all(FILE *file);
    void free_all();
    // ~ContextNode();
};

/*****************************************************************************/

#endif _ContextNode_H_

