// *************************************************************************
//    HashIndex.h
// *************************************************************************

#ifndef _HashIndex_H_
#define _HashIndex_H_

#include "GenericIndex.h"

// *************************************************************************

#define UnmatchedHashValue ((HashVal) -3)

// *************************************************************************
/*
    We separate the index from the relation that is being indexed.
    class HashNode and class HashIndex form the index part;
    class HashSimpleRelation in hash-rel.h reflects this separation.
*/

class corHashNode {
    friend class corHashIndex;
    friend class corHashIndexSet;

protected:
    HashVal hash_val;
    corHashNode *next;
    corTuple *value;
    corHashNode(HashVal hash, corTuple*val) {
        hash_val = hash; 
        value = val; 
        next = NULL;      
    }
public:
    inline corTuple *tuple()    { return value; }
};

typedef corHashNode *HashNodePtr;

//GArrayDeclare(HashNodePtr)

// *************************************************************************

class corHashIndex : public corGenericIndex {
    friend class corHashIndexSet;
protected:
    int hash_size;                  // number of buckets in the hash index
    int limit;                      // Grow hash table if there are more
                                    // than this many tuples

    double incr_ratio, max_occupancy;

    // The following is to be used when linear hashing is introduced.
    // int rehash-point;            // Tuples in buckets with numbers less
                                    // than this //  have been rehashed.

    int hash_count;                 // number of Tuples in hash_table,
                                    // but not in var_chain

    HashNodePtr *hash_table;
//  GArray(HashNodePtr) hash_table;
    HashNodePtr var_chain;          // tuples whose hash value is VarHashValue
 
    // HashNodePtr unmatched;  tuple which do not match pattern

    // The next two fields implement a scheme of linking the non-empty
    // buckets of the relation into a list, so that empty buckets are
    // never considered during a file scan.

    int *non_empty_buckets;         // array of same size as hash table.
                                    // a non-negative entry is the index
                                    // of the next non-empty bucket
    int non_empty_bucket_head;      // index of first non-empty bucket


    int **prev_bucket_ptrs;         // array of pointers to previous
                                    // entry in non-empty bucket link

public: 
    corHashIndex( corStorageRelation* rel, corBitVector* aset, int buckets = -1,
        double incr_ratio1 = -1.0, double max_occupancy1 = -1.0);
    corHashIndex( corStorageRelation* rel, corArgList* pat, int n_var, corArgList* v_nm,
        int buckets = -1, double incr_ratio1 = -1.0,
        double max_occupancy1 = -1.0);

    void initialize_iterator(corTupleIterator&);

    corTuple *get_next_tuple(corTupleIterator& it) ;

    corBindEnv *get_next(corTupleIterator&);

    HashVal hash(corArgList *args, corBindEnv *bindenv);
    void insert_tuple(corTuple *tuple);
    virtual int is_subsumed(corTuple *tuple);

    virtual int index_type()            {return COR_HASH_INDEX;}
    void insert_entry(HashNodePtr entry);
    void rehash_bigger();
    void free_all_entries();
    void move_all_entries(corGenericIndex *ind);
    void delete_all_tuples();
    void reset_index();    

    virtual ~corHashIndex();
};

typedef corHashIndex *HashIndexPtr;

// *************************************************************************

#define DONT_USE_STATIC_ENV   1
#define MAX_VARS 64
#define VAR_INCR 8

// *************************************************************************

#endif _HashIndex_H_
