/*****************************************************************************/
//    StorageRelation.h
/*****************************************************************************/

#ifndef _StorageRelation_H_
#define _StorageRelation_H_

#include "Relation.h"
#include "TupleList.h"
#include "RMark.h"
//#include "GenericIndex.h"
//#include "GenericIndexSet.h"
#include "ArgList.h"
#include "BitVector.h"
#include "PredAnnotations.h"
#include "AggSelInfo.h"
#include "Tuple.h"
#include "BindEnv.h"
//#include "TupleIterator.h"
//#include "QFModuleData.h"
//#include "GoalNode.h"
#include "SymbolArg.h"

/*****************************************************************************/

class corRelation;
class corRelationList;
class corGenericIndex;
class corGenericIndexSet;
class corContext;
class corGoalNode;
class corTupleIterator;
class corQFModuleData;

/*****************************************************************************/

/*  typedef class StorageRelation StorageRelation;  */

class corStorageRelation : public corRelation, public corTupleList {
public:
    // auxiliary relation used only with magic relations, for generating
    // goal identifiers if required.
    corStorageRelation *for_supp_index;

    //     relation delta fields
    corRMark headMark;
    corRMark* free_marks;        // List of free rmarks. Used to reduce calls to new
    corRelationList *delta_relations;
    // above two fields used for determining if the delta relations
    // will be a chain of tuples or a chain of indexed relations

    // List of free relations that can be reused as delta relations.
    corRelationLink *free_reln_list;

    inline int there_are_marks() {  return headMark.next != &headMark; }

    // index and semi-naive related methods
    virtual corGenericIndexSet *get_index_set() {
        return NULL;
    }
    virtual corGenericIndex *create_index_object(corBitVector *) {
        return NULL;
    }
    virtual corGenericIndex *create_index_object(corArgList *, int, corArgList *) {
        return NULL;
    }

    int add_index(char *s);
    int add_index(corBitVector *s);
    virtual int add_index(corArgList *args);
    virtual int add_index(corArgList *pat, int n_var, corArgList *v_nm);
    int add_all_indices(corStorageRelation *reln);
    int add_all_tuples (corStorageRelation *_reln, int marks_into_account = 0,
        int check_dupl = 0, int hide_em = 0);
    virtual int add_all_goals (corStorageRelation *reln, corContext *context,
        corStorageRelation *owner_rel);

    virtual void add_agg_sel_anno(corPredAnnotations *anno);
    virtual void add_agg_sel_info(corAggSelInfo *newinfo);


    // insert and related methods

    // this is called to actually perform a physical insert of a tuple
    virtual int insert_raw(corTuple *tuple);

    // insert_new is called to insert a _new_ tuple
    // it therefore has to perform subsumption checks, etc
    // insert if not subsumed.
    virtual int insert_new(corArgList& args, corBindEnv* env,
        corBindEnv *dont_rename_env = NULL, corTuple *parent_goal = NULL,
        int dont_rename = 0); 

    // this checks to see if there are deltas
    virtual int insert_tuple(corTuple *tuple); 
    int insert(corTuple *tuple);            // this is probably redundant
                                         // this one is used by ordered search

    int insert_in_context(corGoalNode *g_node, corGoalNode *parent_goal);

    int insert_grouping(corArgList&args, corBindEnv *env,
        corBindEnv *dont_rename_env = NULL, corTuple *parent_goal= NULL,
        int dont_rename=0); 

    virtual corRMark* getMark();
    virtual void freeMark(corRMark *, int hide_em=0);
    virtual void free_allMarks();

    virtual int is_subsumed(corTuple *tuple, corRMark *startmark= NULL,
        corRMark *endmark = NULL);
    virtual int is_agg_subsumed(corTuple *tuple, corRMark *startmark= NULL,
        corRMark *endmark = NULL);

    virtual void free_all();
    virtual void free_indices();
    virtual int check_subsumption()        { return check_subsum; }
    void set_subsumption(int subsum_flag) {
        check_subsum = subsum_flag;
    }

    virtual corBindEnv* get_next(corTupleIterator&) = 0;
    virtual corTuple* get_next_tuple(corTupleIterator&) = 0;
    virtual ~corStorageRelation();
    corStorageRelation( int arity, int delta_indexed = -1 );
    corStorageRelation();
};

/*****************************************************************************/

// A Tuple iterator is used to loop through (part of a) Relation
// Only return those tuples "between" start_mark and end_mark,
// and only those which match 

typedef corStorageRelation * (*InterpDerivedSolver)(corQFModuleData *md,
    corStorageRelation *magic, int old_magic_relation,
    corTupleIterator* iterator=NULL);

corStorageRelation* corAllocateRelation(Name name, int arity, int delta_indexed);

/*****************************************************************************/

#endif _StorageRelation_H_

