#include "RdbSystem.h"
#include "RdbDatabase.h"


RdbDatabase* RdbSystem::LookupDatabase(const String dbName, 
				       const String serverName)
{
    if( currentDb && currentDb->DatabaseName() == dbName 
       && currentDb->ServerName() == serverName )
      return currentDb;
    int i = FindDatabase(dbName, serverName);
    if( i >= 0 )
      return dbList[i];
    return NULL;
}


void RdbSystem::SetFocusToDatabase(RdbDatabase* db)
{
    if( currentDb )
      currentDb->RemoveFocus();
    int i = FindDatabase(db);
    assert(i>=0);		// better be found!
    currentDb = db;
}


void RdbSystem::DisconnectAllDatabases()
{
    for( int i = dbList.Size() -1; i >= 0 ; i-- ) {
	dbList[i]->Disconnect();
    }
}


void RdbSystem::CommitAllDatabases()
{
    for( int i = dbList.Size() -1; i >= 0 ; i-- ) {
	dbList[i]->Commit();
    }
}


void RdbSystem::RollbackAllDatabases()
{
    for( int i = dbList.Size() -1; i >= 0 ; i-- ) {
	dbList[i]->Rollback();
    }
}


void RdbSystem::RegisterDatabase(RdbDatabase* db)
{
    dbList.Append(db);
}


void RdbSystem::DeregisterDatabase(RdbDatabase* db)
{
    int i = FindDatabase(db);
    assert(i >= 0);
    if( currentDb == db )
      currentDb = NULL;
    dbList.Delete(i);
}


int RdbSystem::FindDatabase(const RdbDatabase* db) const
{
    for( int i = dbList.Size() -1; i >= 0 ; i-- )
      if( dbList[i] == db ) {
	  return i;
      }
    return -1;
}


int RdbSystem::FindDatabase(const String dbName,
			    const String serverName) const
{
    for( int i = dbList.Size() -1; i >= 0 ; i-- )
      if( dbList[i]->DatabaseName() == dbName &&
	  dbList[i]->ServerName() == serverName ) {
	  return i;
      }
    return -1;
}

