#ifndef _RDBSYSTEM_H_
#define _RDBSYSTEM_H_

// A RdbSystem represents one particular DBMS.
// It controls the RdbDatabase's.

#include "RdbError.h"
#include "List.h"
#include "String.h"
#include "Boolean.h"

class RdbDatabase;		// forward declaration

typedef PtrList<RdbDatabase> RdbDatabaseList;


class RdbSystem
: public RdbError
{
  public:
    
    RdbSystem();	       

    virtual ~RdbSystem();

    virtual String SystemName() const = 0;

    Boolean IsCurrentDatabase(const RdbDatabase* db) const;
				// tests to see if db is the database which
				// currently has the focus
    
    RdbDatabase* LookupDatabase(const String dbName,
				const String serverName = String());
				// find the database by name, if never 
				// connected to, return NULL
    
    virtual RdbDatabase* OpenDatabase(const String dbName,
				      const String userid = String(),
				      const String password = String(),
				      const String serverName = String() ) = 0;
				// open database in backend

    // currently, two-phase commit is not used, but it will be shortly...
    // shouldn't these return some error codes??

    void CommitAllDatabases();	// issue a commit to all databases

    void RollbackAllDatabases(); // issue a rollback to all databases

    void DisconnectAllDatabases(); // disconnect from all databases (& commit)

    virtual void const * Implementor() const = 0;
				// access to implementor (child) class 'this'
    
  protected:
    
    int FindDatabase(const RdbDatabase* db) const;
				// get the subscript into dbList by pointer

    int FindDatabase(const String dbName,
		     const String serverName) const;
				// get the subscript into dbList by name

  private:
    
    RdbDatabase* currentDb;	// current database

    RdbDatabaseList dbList;	// all the databases that are instantiated
    
    void RegisterDatabase(RdbDatabase* db);
				// called by RdbDatabase constructor

    void DeregisterDatabase(RdbDatabase* db);
				// called by RdbDatabase destructor

    void SetFocusToDatabase(RdbDatabase* db);
				// removes focus from current db & 
				// makes db current called by 
				// RdbDatabase::SetFocus
    
    RdbSystem(const RdbSystem&); // not allowed

    RdbSystem& operator =(const RdbSystem&); // not allowed

    friend class RdbDatabase;	// to register database,
				// deregister database,
				// set the focus
};


#include "RdbSystem.i"

#endif // _RDBSYSTEM_H_
