/*****************************************************************************/
//	ArgList.C
/*****************************************************************************/

#include "ArgList.h"

#include "ObjectCount.h"		// REMOVE WHEN FIXED
#include "BitVector.h"

/*****************************************************************************/
corArgList::corArgList()
{
	numArgs = 0;
	argsSize = 0;
	args = NULL;
}

corArgList::corArgList(const corArgList & arglist)
{
 numArgs = arglist.numArgs;
 argsSize = arglist.argsSize;
 args = new corCArg[argsSize];
 //wX-ATTN: is sharing ok here?
 for ( int i = 0 ; i < argsSize; i++) 
   args[i] = arglist.args[i];
}

void corArgList::operator=(const corArgList & arglist)
{
 if (args !=NULL) 
   delete []args;
 numArgs = arglist.numArgs;
 argsSize = arglist.argsSize;
 args = new corCArg[argsSize];
 for ( int i = 0 ; i < argsSize; i++) 
   args[i] = arglist.args[i];
}

corArgList::corArgList(int size)
{
	numArgs = argsSize = size;
	args = new corCArg[size];
	
#ifdef DEBUG
		corC_NewArgListCount++;
		corC_NewArgListSize+=numArgs;		// check
#endif
}


corArgList::~corArgList()
{
 delete []args;
}

/*****************************************************************************/
void corArgList::set_count(int size)
{
	int i;

	if (size == numArgs) return;

	if (size <= 0) {
		numArgs = argsSize = 0;
		delete args;
		args = NULL;
		
		return;
	}

	if (args == NULL) {
		numArgs = argsSize = size;
		//args = new ArgPtr[size];
		//for (i=0; i<size; i++) args[i] = NULL;
		args = new corCArg[size];
		//for (i=0; i<size; i++) args[i].init();
		return;
	}

	if (size < numArgs) {
		numArgs = size;
		return;
	}

	//	size > numArgs
	if (size <= argsSize) {
		//for (i=numArgs; i<size; i++) args[i] = NULL;
		//for (i=numArgs; i<size; i++) args[i].other = NULL;
		for (i=numArgs; i<size; i++) args[i].init();
		numArgs = size;
	}
	else {
	  // newArgs = new ArgPtr[size];
	  // 		for (i=0; i<numArgs; i++) newArgs[i] = args[i];
	  // 		for (i=numArgs; i<size; i++) newArgs[i] = NULL;
		corCArg *newArgs = new corCArg[size];
		for (i=0; i<numArgs; i++) newArgs[i] = args[i];
		//for (i=numArgs; i<size; i++) newArgs[i].other = NULL;
		for (i=numArgs; i<size; i++) newArgs[i].init();
		numArgs = argsSize = size;
		delete args;
		args = newArgs;
	}
}
 
/*****************************************************************************/
void corArgList::print_dump(corBindEnv *env, FILE *file) const
{
 print(env, file);
}

/*****************************************************************************/
void corArgList::print(corBindEnv *env, FILE *file) const
{
	int bFirst_arg = TRUE;
	for( int j =0; j < numArgs; j++) {
		if (bFirst_arg==FALSE)
		  fprintf(file, ",");
		else
		  bFirst_arg= FALSE;
		args[j].print(env, file);
	} 
}

/*****************************************************************************/
void corArgList::print(corBindEnv *env, FILE *file, char *) const
{
	print(env, file);
}

/*****************************************************************************/
void corArgList::sprint(char* str, int* pos, corBindEnv *env) const
{
	int bFirst_arg = TRUE;
	FOR_EACH_ARG(arg, *this) {
		if (bFirst_arg == FALSE) {
			sprintf(str + *pos, ",");
			(*pos)++ ;
		       }
		else
		  bFirst_arg = FALSE;
		arg->sprint(str, pos, env);
	} END_EACH_ARG
}

/*****************************************************************************/
int corArgList::isConstant()
{
	FOR_EACH_ARG(arg, *this) {
		if (! arg->isConstant()) return (0); 
	} END_EACH_ARG
	return (1) ;
}

/*****************************************************************************/
corVarLink *corArgList::var_list(corVarLink *rest)
{
// 	FOR_EACH_ARG(targ,*this) {
// 		rest = targ->var_list(rest);
// 	} END_EACH_ARG;
 for (int j=0; j < numArgs; j++) {
   rest = args[j].var_list(rest);
  }
 return rest;
}

/*****************************************************************************/
Variable corArgList::max_vars() 
{
	corVarLink *rest, *prev, *cur;
	Variable max_var_num=-1;

	rest = this->var_list(NULL);
	for (cur=rest; cur;) {
		if (cur->arg->var > max_var_num)
			max_var_num = cur->arg->var;
		prev = cur;
		cur = cur->next;
		delete prev;
	}
	return max_var_num+1;
}

/*****************************************************************************/
/*
		Computes the hash value of the set of arguments specified by
		bitvector bv.
*/

HashVal corArgList::hash(corBitVector& bv, corBindEnv* env)
{
	HashVal ihash = 0;
	HashVal arg_hash;
	register int bv_i;
	for (bv_i=0; bv_i < count(); bv_i++) {
		if (bv.test(bv_i)) {
			// should probably also check that "bv" has the right count.
			//arg_hash = (*this)[bv_i]->hash(env);
			arg_hash = (*this)[bv_i].hash(env);
			if (arg_hash == VarHashValue)
				return arg_hash;
			ihash = (ihash * 31) ^ arg_hash;
		}
	} 
	return corIntToHash(ihash);
}

/*****************************************************************************/
HashVal corArgList::hash(corBindEnv* env)
{
	HashVal ihash = 0;
	HashVal arg_hash;
	register int bv_i;
	for (bv_i=0; bv_i < count(); bv_i++) {
		//arg_hash = (*this)[bv_i]->hash(env);
	  arg_hash = (*this)[bv_i].hash(env);
		if (arg_hash == VarHashValue)
			return arg_hash;
		ihash = (ihash * 31) ^ arg_hash;
	} 
	return corIntToHash(ihash);
}

/*****************************************************************************/

