/*****************************************************************************/
//	ExprArg.C
/*****************************************************************************/

#include "ExprArg.h"

#include "BuiltinSymbols.h"	// SubOpSymbol
#include "Globals.h"
#include "Suspension.h"
#include "Operators.h"


/*****************************************************************************/
corVarLink * corExprArg::var_list(corVarLink *rest)
{
	rest = left_arg.var_list(rest);
	if (right_arg.IsEmpty() == 0) 
		rest = right_arg.var_list(rest);
	return rest;
}

/*****************************************************************************/
corCArg corExprArg::simplify(corBindEnv *context, corTermLink *&renamed_vars, 
					corBindEnv *const_env, corBindEnv *rename_only_env)
{
	corCArg temp;
	corCArg temp1;
	corCArg tmp_carg(this);

	if (op_arity == 1) {
		if (op == corSubOpSymbol) {
			temp = left_arg.simplify(context,renamed_vars, const_env,
				rename_only_env);
			tmp_carg= (corSubOp(corZeroCArg, temp));
			return tmp_carg;
		}
		else if (op == corBitnotOpSymbol) {
			temp = left_arg.simplify(context,renamed_vars, const_env,
				rename_only_env);
			tmp_carg=corBitxorOp(corMinusOneCArg, temp);
			return tmp_carg;
		}
		else {
			fprintf(corexEnv->error_file, "Unknown operator %s", corSymbolString(op));
			fflush(corexEnv->error_file);
			tmp_carg=  corZeroCArg;
			return tmp_carg;
			// if NULL is returned, it might crash straightaway
			// this way it continues execution
		}
	}
	else {
		ASSERT((op_arity == 2));
		if (op == corAddOpSymbol) {
			temp = left_arg.simplify(context, renamed_vars, const_env,
				rename_only_env);
			temp1 = right_arg.simplify(context, renamed_vars, const_env,
				rename_only_env);
			tmp_carg= corAddOp(temp, temp1);
			return tmp_carg;
		}
		else if (op == corSubOpSymbol) {
			temp = left_arg.simplify(context,renamed_vars, const_env,
				rename_only_env);
			temp1 = right_arg.simplify(context,renamed_vars, const_env,
				rename_only_env);
			tmp_carg=corSubOp(temp, temp1);
			return tmp_carg;
		}
		else if (op == corMulOpSymbol) {
			temp = left_arg.simplify(context,renamed_vars, const_env,
				rename_only_env);
			temp1 = right_arg.simplify(context,renamed_vars, const_env,
				rename_only_env);
			tmp_carg= corMulOp(temp, temp1);
			return tmp_carg;
		}
		else if (op == corDivOpSymbol) {
			temp = left_arg.simplify(context,renamed_vars, const_env,
				rename_only_env);
			temp1 = right_arg.simplify(context,renamed_vars, const_env,
				rename_only_env);
			tmp_carg=corDivOp(temp, temp1);
			return tmp_carg;
		}
		else if (op == corIDivOpSymbol) {
			temp = left_arg.simplify(context,renamed_vars, const_env,
				rename_only_env);
			temp1 = right_arg.simplify(context,renamed_vars, const_env,
				rename_only_env);
			tmp_carg= corDivOp(temp, temp1);
			return tmp_carg;
			// WARNING: this may have to be changed if IDivOp is defined!
		}
		else if (op == corLShiftOpSymbol) {
			temp = left_arg.simplify(context,renamed_vars, const_env,
				rename_only_env);
			temp1 = right_arg.simplify(context,renamed_vars, const_env,
				rename_only_env);
			tmp_carg= corLShiftOp(temp, temp1);
			return tmp_carg;
		}
		else if (op == corRShiftOpSymbol) {
		  temp = left_arg.simplify(context,renamed_vars, const_env, rename_only_env);
		  temp1 = right_arg.simplify(context,renamed_vars, const_env, rename_only_env);
		  tmp_carg= corRShiftOp(temp, temp1);
		  return tmp_carg;
		}
		else if (op == corBitxorOpSymbol) {
			temp = left_arg.simplify(context,renamed_vars, const_env,
				rename_only_env);
			temp1 = right_arg.simplify(context,renamed_vars, const_env,
				rename_only_env);
			tmp_carg= corBitxorOp(temp, temp1);
			return tmp_carg;
		}
		else if (op == corBitandOpSymbol) {
			temp = left_arg.simplify(context,renamed_vars, const_env,
				rename_only_env);
			temp1 = right_arg.simplify(context,renamed_vars, const_env,
				rename_only_env);
			tmp_carg= corBitandOp(temp, temp1);
			return tmp_carg;
		}
		else if (op == corBitorOpSymbol) {
			temp = left_arg.simplify(context,renamed_vars, const_env,
				rename_only_env);
			temp1 = right_arg.simplify(context,renamed_vars, const_env,
				rename_only_env);
			tmp_carg = corBitorOp(temp, temp1);
			return tmp_carg;
		}
		else if (op == corModOpSymbol) {
			temp = left_arg.simplify(context,renamed_vars, const_env,
				rename_only_env);
			temp1 = right_arg.simplify(context,renamed_vars, const_env,
				rename_only_env);
			tmp_carg= corModOp(temp, temp1);
			return tmp_carg;
		}
		else {
			fprintf(corexEnv->error_file, "Unknown operator %s", corSymbolString(op));
			fflush(corexEnv->error_file);
			tmp_carg=  corZeroCArg;
			return tmp_carg;
		}
	}
}

/*****************************************************************************/
HashVal corExprArg::hash(corBindEnv *context)
{
	corTermLink *renamed_vars = NULL;
	corCArg a1 = simplify(context, renamed_vars, context, NULL);
	if (a1.IsEmpty() == 0)
	  return a1.hash(context);
	else 
	  return VarHashValue;
}

/*****************************************************************************/
void corExprArg::print_dump(corBindEnv *env, FILE *file) const
{
	fprintf(file, "\"");
	print(env, file);
	fprintf(file, "\"");
}

/*****************************************************************************/
void corExprArg::print(corBindEnv *env, FILE *file) const
{
	fprintf(file, "(");
	if (op_arity == 1) {
		if (op == corSubOpSymbol) {
			fprintf(file, "-");
			left_arg.print(env, file);
		}
		else {
			fprintf(corexEnv->error_file, "Unknown operator %s", corSymbolString(op));
			fflush(corexEnv->error_file);
			corZeroCArg.print(env, file);
		}
	}
	else {
		ASSERT((op_arity == 2));
		left_arg.print(env, file);
		if (op == corAddOpSymbol) {
			fprintf(file, "+");
		}
		else if (op == corSubOpSymbol) {
			fprintf(file, "-");
		}
		else if (op == corMulOpSymbol) {
			fprintf(file, "*");
		}
		else if (op == corDivOpSymbol) {
			fprintf(file, "/");
		}
		else if (op == corIDivOpSymbol) {
			fprintf(file, "//");
		}
		else if (op == corLShiftOpSymbol) {
			fprintf(file, "<<");
		}
		else if (op == corRShiftOpSymbol) {
			fprintf(file, ">>");
		}
		else if (op == corBitxorOpSymbol) {
			fprintf(file, corSymbolString(corBitxorOpSymbol));
		}
		else if (op == corBitandOpSymbol) {
			fprintf(file, corSymbolString(corBitandOpSymbol));
		}
		else if (op == corBitorOpSymbol) {
			fprintf(file, corSymbolString(corBitorOpSymbol));
		}
		else if (op == corBitnotOpSymbol) {
			fprintf(file, corSymbolString(corBitnotOpSymbol));
		}
		else if (op == corModOpSymbol) {
			fprintf(file, " mod ");
		}
		else {
			fprintf(corexEnv->error_file, "Unknown operator %s", corSymbolString(op));
			fflush(corexEnv->error_file);
		}
		right_arg.print(env, file);
	}
	fprintf(file, ")");
}

/*****************************************************************************/
void corExprArg::printon(FILE *file) const
{
	corsuspension.byte_count -=80; // WX needs furthur work
	
	fprintf(file, "(");
	if (op_arity == 1) {
		if (op == corSubOpSymbol) {
			fprintf(file, "-");
			left_arg.printon(file);
		}
		else {
			fprintf(corexEnv->error_file, "Unknown operator %s", corSymbolString(op));
			fflush(corexEnv->error_file);
			corZeroCArg.printon(file);
		}
	}
	else {
		ASSERT((op_arity == 2));
		left_arg.printon(file);
		if (op == corAddOpSymbol) {
			fprintf(file, "+");
		}
		else if (op == corSubOpSymbol) {
			fprintf(file, "-");
		}
		else if (op == corMulOpSymbol) {
			fprintf(file, "*");
		}
		else if (op == corDivOpSymbol) {
			fprintf(file, "/");
		}
		else if (op == corIDivOpSymbol) {
			fprintf(file, "//");
		}
		else if (op == corLShiftOpSymbol) {
			fprintf(file, "<<");
		}
		else if (op == corRShiftOpSymbol) {
			fprintf(file, ">>");
		}
		else if (op == corBitxorOpSymbol) {
			fprintf(file, corSymbolString(corBitxorOpSymbol));
		}
		else if (op == corBitandOpSymbol) {
			fprintf(file, corSymbolString(corBitandOpSymbol));
		}
		else if (op == corBitorOpSymbol) {
			fprintf(file, corSymbolString(corBitorOpSymbol));
		}
		else if (op == corBitnotOpSymbol) {
			fprintf(file, corSymbolString(corBitnotOpSymbol));
		}
		else if (op == corModOpSymbol) {
			fprintf(file, " mod ");
		}
		else {
			fprintf(corexEnv->error_file, "Unknown operator %s", corSymbolString(op));
			fflush(corexEnv->error_file);
		}
		right_arg.printon(file);
	}
	fprintf(file, ")");

}

/*****************************************************************************/

void corExprArg::printon(FILE *, char *) const
{
}

/*****************************************************************************/
void corExprArg::sprint(char* str, int* pos, corBindEnv *) const
{
	sprintf(str + *pos, "%s", "<some number>") ;
	*pos += strlen("<some number>") ;
}

/*****************************************************************************/

