/*****************************************************************************/
//	NumArg.C
/*****************************************************************************/

#include "NumArg.h"

#include "Globals.h"

/*****************************************************************************/
/*
const NumArg *Zero = new NumArg(0);
const NumArg *One = new NumArg(1);
const NumArg *MinusOne = new NumArg(-1);
*/

static long print_width = -1;
static long print_precision = -1;
static char print_double_format = 'g';
static char print_int_format = 'd';

int corC_NewIntCount = 0;
int corC_NewDoubleCount = 0;

extern "C" corNumArg * corNewDouble(double d);
/* EXTERN Double * NewDouble ARGS((double d)); */

/*****************************************************************************/
/*
 * It is important for NumArgs to return copies when simplifying. This way
 * the original arg can be deleted without affecting the simplified copy.
 * Consider the case when a tuple from relation A is being added to relation
 * B. CopyArgs is called on the args of the A tuple, and this calls simplify.
 * The arglist returned by CopyARgs is inserted into a B tuple. Later, the
 * A tuple is deleted, and trying to be memory efficient, we reclaim the
 * args in it. This screws up B unless we make a copy. Right now we do the
 * reclaiming only for NumArgs, hence the copying for NumArgs
 */

corCArg corNumArg::simplify(corBindEnv *, corTermLink *&,
					  corBindEnv *, corBindEnv *)
{
/****
		if (_num_kindof == COR_INTEGER)
				return new NumArg(i_val);
		else 
				return new NumArg(d_val); 
*****/
 corCArg c(this);
		return c;
}

/*****************************************************************************/
int corNumArg::equals(corArg *arg2)
{
	if (arg2->kindof() != COR_NUM_CONST)
		return (0);
	return( ! compare(*((corNumArg*)arg2)) );
}

/*****************************************************************************/
HashVal corNumArg::hash(corBindEnv *)
{
	int *ptr ;
	ptr = (int *)&d_val;
	return corIntToHash((_num_kindof == COR_INTEGER) ? i_val : (ptr[0] ^ ptr[1]));
	// return IntToHash((_num_kindof == COR_INTEGER) ? i_val : ( (*(ptr = (int*)&d_val)) ^ ptr[1])); 
}

/*****************************************************************************/
int corNumArg::compare(const Root& other) const
{
	if (_num_kindof == COR_INTEGER && other._num_kindof == COR_INTEGER) {
		return (other.i_val < i_val ? 1 : other.i_val==i_val ? 0 : -1); 
	}

	if (_num_kindof == COR_DOUBLE && other._num_kindof == COR_DOUBLE) {
		return (other.d_val < d_val ? 1 : other.d_val==d_val ? 0 : -1); 
	}

	return (other.doubval() < doubval() ?
		1 : other.doubval()==doubval() ? 0 : -1);  
}

/*****************************************************************************/
void corNumArg::printon(FILE *file) const
{
	char format[20]; char buffer[100];
	char *fptr = format;

	*fptr++ = '%';
	if (print_width >= 0) {
		sprintf(fptr, "%ld", print_width);
		fptr += strlen(fptr);
	}
	if (print_precision >= 0) {
		sprintf(fptr, ".%ld", print_precision);
		fptr += strlen(fptr);
 	}
	*fptr++ = (_num_kindof == COR_INTEGER ? print_int_format :
		print_double_format);
	*fptr = '\0';
	if (_num_kindof == COR_INTEGER)
		sprintf(buffer, format, i_val);
	else
		sprintf(buffer, format, d_val);
 
	// Kludge to make sure the last char isn't '.'.
	fptr = buffer + strlen(buffer);
	if (fptr[-1] == '.') {
		fptr[0] = '0';
		fptr[1] = 0;
	}

	fputs(buffer, file);
}

/*****************************************************************************/
void corNumArg::printon(FILE *file, char *fmt_str) const
{
	char buffer[100];
	if (_num_kindof == COR_INTEGER)
		sprintf(buffer, fmt_str, i_val);
	else
		sprintf(buffer, fmt_str, d_val);
	
	char *temp = fmt_str;
	
	if (*temp == '\0')
		fprintf(corexEnv->error_file,
			"Incorrect format statement provided to print float value.\n");
	
	while (*temp != '\0') {
		if (_num_kindof == COR_INTEGER &&
			(((*temp) == 'd') || ((*temp) == 'D'))) break;
		if (_num_kindof == COR_DOUBLE && 
			(*temp == 'f' || *temp == 'g' || *temp == 'G' || *temp == 'e' ||
				*temp == 'E')) break;
		temp++;
	}
	
	// Kludge to make sure the last char isn't '.'.
	char * fptr = buffer + strlen(buffer);
	if (fptr[-1] == '.')
	{
		fptr[0] = '0';
		fptr[1] = 0;
	}
	
	fputs(buffer, file);
}

/*****************************************************************************/
void corNumArg::sprint(char* str, int* pos, corBindEnv *) const
{
	if (_num_kindof == COR_INTEGER) sprintf(str + *pos, "%d", i_val);
	else  if (_num_kindof == COR_DOUBLE) sprintf(str + *pos, "%f", d_val);

	*pos = strlen(str);
}

/*****************************************************************************/
// This is extern "C", so it can be called by lex.

corNumArg* corMakeFixInt(int i)
{
	corC_NewIntCount++;
	return new corNumArg(i);
}

/*****************************************************************************/
corNumArg * corNewDouble(double d) 
{
	corC_NewDoubleCount++;
	return new corNumArg(d);
}

/*****************************************************************************/

