/*****************************************************************************/
//	VarArg.C
/*****************************************************************************/

#include "VarArg.h"

#include "Globals.h"
#include "SymbolArg.h"
#include "BindEnv.h"

/*****************************************************************************/
void corVarArg::print(corBindEnv* env, FILE *file) const
{
	if (env == NULL || env->lookup(var).expr.IsEmpty())
		fprintf(file, "%s", corSymbolString(var_name()));
	else env->lookup(var).printon(file) ;		
}

/*****************************************************************************/
void corVarArg::print(corBindEnv* env, FILE *file, char *fmt_str) const
{
	char * temp = fmt_str;

	while (*temp != '\0') {
		if (*temp == 's') break;
		temp++;
	}

	if (*temp == '\0') {
		fprintf( corexEnv->error_file,
			"Incorrect format statement provided to print string.\n");
		print(env, file);
	}
	else {
		if (env == NULL || env->lookup(var).expr.IsEmpty())
			fprintf(file, fmt_str, corSymbolString(var_name()));
		else env->lookup(var).printon(file) ;
	}
}

/*****************************************************************************/
int corVarLink::contains(corVarArg *arg)
{
	for (corVarLink* cur=this; cur!= NULL; cur = cur->next) {
		if (cur->arg->var == arg->var)
			return 1;
	}
	return 0;
}

/*****************************************************************************/
void corVarArg::sprint(char* str, int* pos, corBindEnv *env) const
{
	if (env == NULL || env->lookup(var).expr.IsEmpty()) {
		sprintf(str + *pos, "%s", corSymbolString(var_name()));
		*pos += SymbolLength(var_name()) ;
	}
	else {
		(env->lookup(var).expr).sprint(str, pos, env->lookup(var).bindenv);
	}
}

/*****************************************************************************/
int corVarArg::isConstant() 
{
	return 0;
}

/*****************************************************************************/
corCArg corVarArg::simplify(corBindEnv* env, corTermLink*& renamed_vars, 
				corBindEnv* const_env, corBindEnv *rename_only_env) 
{		
	corTerm term;
	corCArg tmp_carg(this);

	if (env != NULL) 
		term = env->lookup(var);

	if (rename_only_env && env != rename_only_env && env != const_env) {
	  corCArg nullCarg;
	  return nullCarg;
	 }

	if (env == const_env && (env == NULL || term.expr.IsEmpty()) )
		return tmp_carg;

	if (env == NULL || term.expr.IsEmpty()) {
		corTermLink *tlink = renamed_vars;
		int new_i = 0;
		for ( ; tlink != NULL; tlink = tlink->next, new_i++)
			if (tlink->old_var == var && tlink->old_bindenv == env) 
				return tlink->new_arg;
		tlink = new corTermLink;
		tlink->next = renamed_vars;
		renamed_vars = tlink;
		tlink->old_var = var;
		tlink->old_bindenv = env;
		char varname[100];
		sprintf(varname, "_X%d", const_env->max_var_number()+new_i);
		tlink->new_arg = new corVarArg(corEnterSymbol(varname),
			const_env->max_var_number()+new_i);
		return tlink->new_arg;
	}

	return term.expr.simplify( term.bindenv, renamed_vars, const_env,
		rename_only_env);
}

/*****************************************************************************/
corVarLink * corVarArg::var_list(corVarLink *rest)
{
	/* Check for duplicates. */
	corVarLink *link;

	for (link = rest; link != NULL; link = link->next) {
		if (link->arg->var == var) return rest;
	}
	return new corVarLink(this, rest);
}

/*****************************************************************************/
int corVarArg::equals(corArg *arg2)
{
	if (arg2->kindof() != COR_VARIABLE)
		return(0);
	return (var == ((corVarArg *)arg2)->var);
}

/*****************************************************************************/
HashVal corVarArg::hash(corBindEnv *env /*= 0*/) 
{
	if (env == NULL) return VarHashValue;
	corTerm term = env->lookup(var);
	if (term.expr.IsEmpty()) return VarHashValue;
	return term.expr.hash(term.bindenv);
}

/*****************************************************************************/
int corVarArg::scan_vars(void *data, VarFunc *func)
{
	return func(this, data);
}

/*****************************************************************************/

