/*****************************************************************************/
//    BindEnv.h
/*****************************************************************************/

#ifndef _BindEnv_H_
#define _BindEnv_H_

#include "Arg.h"
#include "ObjectCount.h"

/*****************************************************************************/
// A bindenv is a set of substitutions, mapping variables to Terms.
// Variables in a Clause are numbered sequentially,
// with each variable's number stored in the 'var' field of a VarArg.
// A binding can be implemented as an array, mapping
// 'var' numbers to Terms.

/***  WARNING:
    IF      a bindenv is allocated within a function and deleted on 
            function exit, 
    THEN    1) you MUST create a StackMark first.
        2) before returning from the function you MUST pop back to 
            the stackmark.  Otherwise the trail may contain 
            references to the defunct bindenv, and cause 
            a crash later.
****/

/*****************************************************************************/

#define NOBINDENVHACK

#define USE_TRIES 

/*****************************************************************************/

class corTerm;

/*****************************************************************************/

class corBindEnv {
    friend class corArrayBindEnv;
    friend class corVersionedBindEnv;
    friend class corGArrayBindEnv;

    int _limit;
    int _is_versioned;

public:
    // The following can be restored as inline functions 
    // if we have no subclass of BindEnv other than ArrayBindEnv
#ifdef NOBINDENVHACK
    virtual inline corTerm lookup(Variable var) const = 0;
    virtual inline int bind(Variable var, corTerm binding)=0;
#else
    inline corTerm lookup(Variable var) const;
    inline int bind(Variable var, corTerm binding);
#endif
    virtual corBindEnv *make_version()=0;
    virtual corBindEnv *make_version_and_swap()=0;
    virtual corBindEnv *copy_shell()=0;
    virtual void restore(corBindEnv *env)=0;
    virtual void resize(int newsize)=0;
    virtual int all_free()=0;
    inline int max_var_number() { return _limit;}
    inline int is_versioned() { return _is_versioned;}
};

/*****************************************************************************/

#endif _BindEnv_H_

