// *************************************************************************
//    Trail.h
// *************************************************************************

#ifndef _Trail_H_
#define _Trail_H_

#include "BindEnv.h"
#include "VarArg.h"
#include "Term.h"

// *************************************************************************

struct corBaVarPair {
    corBindEnv *bindenv;
    Variable var;
};

typedef struct corBaVarPair TrailElement;

// *************************************************************************
// The Trail is a means of allowing environments to be created (during
// rule evaluation) and then discarded when they are not needed any more.
// It is a stack of environments (BindEnv/ Variable) pairs which can
// be pushed to grow it, and popped back to a fixed StackMark, to bring
// the environment back to its old state.

class corTrail {
    TrailElement *Stack;
    int Limit;

public:
    // Size of current Trail stack.
    int TrailSize;

    // Creates a trail with the specified initial sizes.
    corTrail(int trailsize);

    // Return specified element of trail
    inline TrailElement& operator [](int i)
    {
        return (Stack[i]);
    }

    // Pop trail to specified size. Clear (in BindArray) variables that are 
    // found on the popped portion of the trail.
    inline void PopTrail(int new_size)
    {
        while (TrailSize > new_size) {
            TrailSize--;
            if (Stack[TrailSize].var >= 0) {
                ((Stack[TrailSize]).bindenv)->bind(Stack[TrailSize].var,
                    corNullTerm);
            }
            else {
                delete Stack[TrailSize].bindenv ;
            }
        }
    }
                
    inline void PushTrail(corBindEnv *env, Variable var )
    {
        if (TrailSize >= Limit) {
            if (Limit == 0) {
                  Limit = 255;
                  //Stack = (TrailElement*)malloc(Limit * sizeof(TrailElement));
                  Stack = new TrailElement[Limit];
            }
            else {
                  TrailElement *tmpStack = Stack;
                  int len = Limit;
                  Limit += Limit;
                  //Stack = (TrailElement*)realloc((char *)Stack,
                  //  Limit * sizeof(TrailElement));
                  Stack = new TrailElement[Limit];
                  for (int i=0; i<len; i++)
                      Stack[i] = tmpStack[i];
                  delete [] tmpStack;
            }
        }
        Stack[TrailSize].bindenv = env;
        Stack[TrailSize].var = var;
        TrailSize++;
    }

    // Push a bindenv on trail; 
    // The argument -1 indicates the whole bindenv is to be 
    // deleted on backtracking.
    inline void PushTrail(corBindEnv *env)
    {
        PushTrail(env, -1);
    }
};

// *************************************************************************

extern corTrail *corCurTrail;

// *************************************************************************

#endif _Trail_H_

