/*****************************************************************************/
//    Unify.C
/*****************************************************************************/

#include "Unify.h"

#include "Config.h"
#include "Trail.h"
#include "FuncArg.h"
#include "CoralError.h"
#include "Globals.h"
#include "TupleIterator.h"
#include "HashSimpleRelation.h"
#include "Execute.h"

/*****************************************************************************/

corTerm corGlobalNewt;

#define DO_OCCUR_CHECK 0

/*****************************************************************************/
/*
    unify_binding:  
    This is used to bind a variable to a term.  The variable must be free.
    The term may actually be a variable interpreted in a bindenv.
    If so, it is dereferenced completely. This is actually essential 
    so that a variable is never bound to itself -- that would be a silly 
    (and wrong) thing to do.
*/

/*******************************
%     if (DO_OCCUR_CHECK) {
%     // WARNING:: Incomplete implementation 
%         // Simple check - don't bind variable to itself, but return success. 
%         if (!(base_term->expr->kindof() == COR_VARIABLE && 
%             base_term->varnum() == var && base_term->bindenv == bindenv) ) 
%         return COR_U_SUCCEED;
%     // Fill in the rest of the occur check code here ....
%     }
********************************/

/*****************************************************************************/
inline int corinline_unify_binding(corBindEnv* bindenv, Variable var, corTerm term)
{
#ifdef DO_TRACE
    if (corexEnv->dbg_unify) {
    fprintf(corexEnv->trace_file,
        "[unify_binding(bindenv=0x%x, var#=%d, arg=0x%x=",
        bindenv, var, &term.expr);
    term.expr.printon(corexEnv->trace_file);
    fprintf(corexEnv->trace_file, ", bindenv=0x%x)]\n", term.bindenv);
    }
#endif


    corfull_dereference_term(term);

    if ( ! (term.bindenv == bindenv 
        && term.expr.kindof() == COR_VARIABLE
        && (term.expr.VarArgP())->var == var )) {
      bindenv->bind(var, term);
      corCurTrail->PushTrail((bindenv), (var));
    }    
    
    /* else:  don't bind variable to itself -- simply return success.
       This limited form of occur check is critical for 
       answer-return unification with non-ground facts and 
       return-unification optimizations. */
    
#ifdef DO_TRACE
    if (corexEnv->dbg_unify) fprintf(corexEnv->trace_file, "unify_binding succeeded.\n");
#endif
    return COR_U_SUCCEED;
}

/*****************************************************************************/
int corunify_binding(corBindEnv* bindenv, Variable var, corTerm term)
{
    return corinline_unify_binding(bindenv, var, term);
}

/*****************************************************************************/
int corsubsumes_binding(corBindEnv* bindenv, Variable var, corTerm term, int can_bind)
{
#ifdef DO_TRACE
    if (corexEnv->dbg_unify) {
    fprintf(corexEnv->trace_file,
        "[subsumes_binding(bindenv=0x%x, var=0x%x, arg=0x%x=",
        bindenv, var, &term.expr);
    term.expr.printon(corexEnv->trace_file);
    fprintf(corexEnv->trace_file, ", bindenv=0x%x)]\n", term.bindenv);
    }
#endif
    ASSERT((bindenv != NULL && bindenv->lookup(var).expr == NULL))

    corTerm base_term = term;
    corfull_dereference_term(base_term);

    if (base_term.expr.kindof() == COR_VARIABLE && 
        base_term.varnum() == var && base_term.bindenv == bindenv) 
    return COR_U_SUCCEED;

    if(!can_bind) 
        return(COR_U_FAIL);

    if (DO_OCCUR_CHECK) {
    // WARNING: Incomplete implementation
    }
    
    bindenv->bind(var, base_term);
    corCurTrail->PushTrail(bindenv, var);

#ifdef DO_TRACE
    if (corexEnv->dbg_unify) fprintf(corexEnv->trace_file,
                 "subsumes_binding succeeded.\n");
#endif
    return COR_U_SUCCEED;
}

/**************************************************************************/
/*
    Unify the terms 'left' and 'right'.
    Return COR_U_SUCCEED or COR_U_FAIL.

    WARNING:  COR_DONTCARE is dangerous:  make sure there is no underscore
    within a structire.  If you unify f(_) with f(a), the result on the lhs
    will be f(_), not f(a).  Thus the result may depend on which side you
    choose!!  This is shady.  In particular, the result of unifying 
    g(X,X,X) with g(f(_),f(a),Y) is order dependent!!
*/

corunify_args(corTerm left, corTerm right)
{
 arg_kind left_kindof, right_kindof;

#ifdef DO_PROFILE
  if (corT_Stack.count > 0)
    (corT_Stack.top()->moduledata->module_info.corGlobalProfileInfo.unifies)++ ;
  else
    (corGlobalProfileInfo.unifies)++ ;
#endif
 
#ifdef DO_TRACE
  if (corexEnv->dbg_unify) {
    fprintf(corexEnv->trace_file, "[unify_args(left=0x%x=", &left.expr);
    left.printon(corexEnv->trace_file);
    fprintf(corexEnv->trace_file, ", right=0x%x=", &right.expr);
    right.printon(corexEnv->trace_file);
    fprintf(corexEnv->trace_file,
        ", left_bindings=0x%x, right_bindings=0x%x)]\n",
        left.bindenv, right.bindenv);
  }
#endif
  corfull_dereference_term(left);
  
  for(;;) {
    if (left.expr == right.expr && left.bindenv == right.bindenv)
      return COR_U_SUCCEED;

    left_kindof = left.kindof();
    right_kindof = right.kindof();

    switch (left_kindof) {
    case COR_VARIABLE:
      if ( right_kindof != COR_DONTCARE )
    return corinline_unify_binding(left.bindenv, left.varnum(), 
                    right);
      /* Left was fully dereferenced earlier*/
      else return COR_U_SUCCEED;

    case COR_NUM_CONST:
      switch (right_kindof) {
      case COR_NUM_CONST:
    if (right.expr.cNum_kindof() == COR_ARITH_EXPR) {
      corarith_dereference_term(right);
    }
    if (left.expr.compare_num(right.expr))
      return COR_U_FAIL;
    return COR_U_SUCCEED;
      case COR_VARIABLE:
    if ( corisfree_term(&right) ) 
      return corinline_unify_binding(right.bindenv, right.varnum(), left);
    else {
      corfull_dereference_term(right);
      continue;
    }
      case COR_DONTCARE: return (COR_U_SUCCEED);
      default : return (COR_U_FAIL);
      }

    case COR_SYMBOL_CONST:
      switch (right_kindof) {
      case COR_SYMBOL_CONST:
    /**if ((Symbol*)left.expr != (Symbol*)right.expr)
      return COR_U_FAIL;
    return COR_U_SUCCEED; **/
        if ((left.expr.SymbolP())->equals(right.expr.SymbolP())) //wX-ATTN
      return COR_U_SUCCEED;
    else return COR_U_FAIL;

      case COR_VARIABLE:
    if ( corisfree_term(&right) ) 
      return corinline_unify_binding(right.bindenv, right.varnum(), left);
    else {
      corfull_dereference_term(right);
      continue;
    }
    
      case COR_DONTCARE: return (COR_U_SUCCEED);
      default : return (COR_U_FAIL);
      }

    case COR_CONST_ARG:
      switch (right_kindof) {
      case COR_CONST_ARG:
    if (! left.expr.equals(right.expr))
      return COR_U_FAIL;
    return COR_U_SUCCEED;
      case COR_VARIABLE:
    if ( corisfree_term(&right) ) 
      return corinline_unify_binding(right.bindenv, right.varnum(), left);
    else {
      corfull_dereference_term(right);
      continue;
    }
      case COR_DONTCARE: return (COR_U_SUCCEED);
      default: return COR_U_FAIL;
      }

    case COR_FUNCTOR:
      switch (right_kindof) {

      case COR_VARIABLE:
    if ( corisfree_term(&right) )
      return corinline_unify_binding(right.bindenv, right.varnum(), left);
    else {
      corfull_dereference_term(right);
      continue;
    }
      case COR_FUNCTOR: {
    corFuncArg* left_func = left.expr.FuncArgP();
    corFuncArg* right_func = right.expr.FuncArgP();
        
    if ((left_func->functor() != right_func->functor()) ||
        (left_func->arity() != right_func->arity()))
      return COR_U_FAIL;
        
    /** NOTE:
      In the following code, after checking if a term is 
      fully ground, we compute the hash value.  isConstant 
      stores VarHashValue if it finds that there is a variable 
      in the term.  If the term is infact
      a constant,  hash stores the hash value.
      
      Hence, so long as this pair is always used together
      no term is traversed more than twice by these functions. 
      If we call only isConstant, ground terms can be traversed  
      many times by these functions.  
      If we call just hash, nonground terms can be traversed 
      many times by these functions.
      **/
    
    if (left_func->isConstant()) {
      HashVal left_hash = left_func->hash(left.bindenv);
      if (corisGroundHashValue(left_hash) && right_func->isConstant()){
        HashVal right_hash = right_func->hash(right.bindenv);
        if ( corisGroundHashValue(right_hash) ) {
          if (left_hash != right_hash)
        return COR_U_FAIL;
#ifdef USE_HASHCONSING
          else return(COR_U_SUCCEED);
#endif
        }
      }
    }
    return corunify_arg_lists(left_func->args, right_func->args,
                   left.bindenv, right.bindenv);
      } // case COR_FUNCTOR
    
      case COR_GROUPING:
    CORAL_error(COR_INTERNAL_ERROR,"Grouping in unification.",
            "unify_args()");
    
      default : return COR_U_FAIL;
      } // left=COR_FUNCTOR, switch (right->kindof())
    
      case COR_DONTCARE: return (COR_U_SUCCEED);

      case COR_RELATION: {
        switch (right_kindof) {
    case COR_VARIABLE: {
      if ( corisfree_term(&right) )
        return corunify_binding(right.bindenv, right.varnum(), left);
      else {
        corfull_dereference_term(right);
        return corunify_args(right, left);
      }
    }
    case COR_RELATION: {
      
      corHashSimpleRelation* left_rel = left.expr.HashSimpleRelP();
      corHashSimpleRelation* right_rel = right.expr.HashSimpleRelP();
      
      if (left_rel->equals(right_rel) == 1) return(COR_U_SUCCEED);
      
      return(COR_U_FAIL);
      
    }
    case COR_GROUPING:
      CORAL_error(COR_INTERNAL_ERROR,"Grouping in unification.",
              "unify_args()");
      
    default : return COR_U_FAIL;
        }
      }
      
    case COR_GROUPING:
      CORAL_error(COR_INTERNAL_ERROR,"Grouping in unification.","unify_args()");
    default:
      fprintf(corexEnv->error_file,"Error! unify: default case yet to be implemented!\n");
      return COR_U_FAIL;      
    } // switch (left->kindof())
      
  } // for (;;) 
      
} // unify_args

/*****************************************************************************/
corunify_arg_lists(corArgList& left, corArgList& right, corBindEnv *left_bindings,
    corBindEnv *right_bindings )
{
#ifdef DO_TRACE
    if (corexEnv->dbg_unify) fprintf(corexEnv->trace_file, "unify_arg_lists called:\n");
#endif

    register int i;
    // might be unnecessary !! : PRAVEEN:EFF1
    // if (right.count() != left.count()) return COR_U_FAIL;

    corTerm left_term, right_term;

    for ( i=0 ;  i < left.count() ; i++ ) {
    left_term.expr = left[i];
    right_term.expr = right[i];
        left_term.bindenv = left_bindings;
        right_term.bindenv = right_bindings;
    if (corunify_args(left_term, right_term) == COR_U_FAIL)
        return COR_U_FAIL;
    }
#ifdef DO_TRACE
    if (corexEnv->dbg_unify) fprintf(corexEnv->trace_file,"unify_arg_lists succeeded\n");
#endif

    return COR_U_SUCCEED;
}

/*****************************************************************************/

extern int corsubsumes_nested_arg_lists(corArgList& left, corArgList& right,
    corBindEnv *left_bindings, corBindEnv *right_bindings, int can_bind);

#define CAN_BIND 1

/*****************************************************************************/
/*
    The following fn. checks if left subsumes right

    WARNING:  The code  assumes that left and right do not share any
      variables! If the do, right would have to be copied
      (recursively) so that it does not share any subterm 
      with left.

    ASSUMPTIONS:
    1) We assume that right and left don't share any variables.
    2) right.bindenv can be NULL - this is fine since right
        cannot be bound in any case.  Some checks for 
        null bindenv are required as a result.
    3) if the flag canbind is not set, then subsumption checks for 
        equality - no variable binding  is allowed.
        The equality is strict in that no variable renaming
        is allowed.
    4) If right.bindenv is not null, we need to know when a 
        left variable is bound to a right variable.
*/

int corsubsumes_args(corTerm left, corTerm right, int can_bind)
{
#ifdef DO_TRACE
    if (corexEnv->dbg_subsumes) {
    fprintf(corexEnv->trace_file, "[subsumes_args(left=0x%x=", &left.expr);
    left.expr.printon(corexEnv->trace_file);
    fprintf(corexEnv->trace_file, ", right=0x%x=", &right.expr);
    right.expr.printon(corexEnv->trace_file);
    fprintf(corexEnv->trace_file,
        ", left_bindings=0x%x, right_bindings=0x%x)]\n",
           left.bindenv, right.bindenv);
    }
#endif
    if (left.expr == right.expr)
    if (left.bindenv == right.bindenv)
        return COR_U_SUCCEED;
    
    switch (left.kindof()) {
      case COR_DONTCARE: return COR_U_SUCCEED;
      case COR_VARIABLE: 
      {
        if (left.bindenv==NULL)  // Can happen is left was bound to a 
                    // right variable and was dereferenced
        return(COR_U_FAIL);
        if ( corisfree_term(&left) )
            return corsubsumes_binding(left.bindenv, left.varnum(), right, 
                can_bind);
#if 1    /* WARNING:  Assumes that the left term has a single level bindenv.
        The condition is satisfied provided it is a fact, since
        our fact representation has only single level bindenvs 
        MAKE SURE that this assumption is valid in other uses!!
    */
        else {
        /* Bug Fix:  the can_bind argument was never changed before.
            It should be changed when dereferencing crosses over
            into the right hand side. */
        corTerm left1 = left.dereference_var();
        if (left1.bindenv == left.bindenv) {    
            /* Still in left -- can bind left variable */
          return corsubsumes_args(left1, right, can_bind);
        }
        else { /* Crossed over to right -- cannot bind left variable */
          return corsubsumes_args(left1, right, 0);
        }
        }
#else
        else return corsubsumes_args(left.dereference_var(), right, 
                right.bindenv?0:can_bind);
            // if right.bindenv is not null, left.bindenv
            // must have been empty, so dereferencing left
            // makes us cross over into right.  Hence set can_bind
            // to 0; 
#endif
      }
      case COR_NUM_CONST:
    if ( left.expr.cNum_kindof() == COR_ARITH_EXPR) {
      corarith_dereference_term(left);
     }
      case COR_SYMBOL_CONST:
      case COR_CONST_ARG:
    switch (right.kindof()) {
      case COR_DONTCARE: return COR_U_FAIL;
      case COR_VARIABLE: {
        if (right.bindenv == NULL || corisfree_term(&right) )
        return (COR_U_FAIL);
        return corsubsumes_args(left, right.dereference_var(), can_bind);
      }
      case COR_NUM_CONST:
        if ( right.expr.cNum_kindof() == COR_ARITH_EXPR) {
          corarith_dereference_term(right);
        }
        corfull_dereference_term(right);
        if (left.kindof() != COR_NUM_CONST
        || (left.expr.compare_num(right.expr)))
	  return COR_U_FAIL;
        return COR_U_SUCCEED;
      case COR_SYMBOL_CONST:
            /**** if (left.kindof() != COR_SYMBOL_CONST ||
                 (Symbol*)left.expr != (Symbol*)right.expr)
                  return COR_U_FAIL;
              return COR_U_SUCCEED; *****/
       if ((right.expr.SymbolP())->equals(left.expr.SymbolP()))
         return COR_U_SUCCEED;
       else return COR_U_FAIL;

      case COR_CONST_ARG:
              if (left.kindof() != COR_CONST_ARG ||
          ! left.expr.equals(right.expr))
                  return COR_U_FAIL;
              return COR_U_SUCCEED;

      case COR_FUNCTOR:
      case COR_RELATION:
         return COR_U_FAIL;
      case COR_GROUPING:
            CORAL_error(COR_INTERNAL_ERROR,"Grouping in subsumption check.",
                        "subsumes_args()");
      default:
         return COR_U_FAIL;
    }

      case COR_RELATION:
        switch (right.kindof()) {
            case COR_DONTCARE: return COR_U_FAIL;
            case COR_VARIABLE: {
               if (right.bindenv == NULL || corisfree_term(&right) )
                   return (COR_U_FAIL);
               return corsubsumes_args(left, right.dereference_var(), can_bind);
            }
          case COR_NUM_CONST:
          case COR_SYMBOL_CONST:
          case COR_CONST_ARG:
          case COR_FUNCTOR:
               return COR_U_FAIL;
          case COR_GROUPING:
            CORAL_error(COR_INTERNAL_ERROR,"Grouping in subsumption check.",
                        "subsumes_args()");
          case COR_RELATION: {

        corRelation *left_rel = left.expr.RelationP();
            corRelation *right_rel = right.expr.RelationP();

            if (left_rel->equals(right_rel) == 1) return COR_U_SUCCEED ;
            return COR_U_FAIL ;
          }
          default:
               return COR_U_FAIL;
        }

      case COR_FUNCTOR:
    switch (right.kindof()) {
      case COR_DONTCARE: return COR_U_FAIL;
      case COR_VARIABLE: {
        if (right.bindenv == NULL || corisfree_term(&right) )
        return (COR_U_FAIL);
        return corsubsumes_args(left, right.dereference_var(), can_bind);
            /* can_bind might start out bound, but not set by subsumes */
      }
      case COR_NUM_CONST:
          case COR_SYMBOL_CONST:
          case COR_CONST_ARG:
      case COR_RELATION:
           return COR_U_FAIL;
      case COR_FUNCTOR: {
        corFuncArg* left_func = left.expr.FuncArgP();
        corFuncArg* right_func = right.expr.FuncArgP();

        if (left_func->functor() != right_func->functor())
        return COR_U_FAIL;
                if (left_func->isConstant()) {
                    HashVal left_hash = left_func->hash(left.bindenv);
                    if (corisGroundHashValue(left_hash) && right_func->isConstant()){
                        HashVal right_hash = right_func->hash(right.bindenv);
                    if ( corisGroundHashValue(right_hash) ) {
                            if (left_hash != right_hash)
                            return COR_U_FAIL;
#if USE_HASHCONSING
                   else return(COR_U_SUCCEED);
#endif
                    }
                    }
                }
         return corsubsumes_nested_arg_lists(
                left_func->args, right_func->args,
                left.bindenv, right.bindenv, can_bind);
        
           } // case COR_FUNCTOR

           case COR_GROUPING:
            CORAL_error(COR_INTERNAL_ERROR,"Grouping in subsumption check.",
                        "subsumes_args()");
           default:
                return(COR_U_FAIL);
           }
      case COR_GROUPING:
        CORAL_error(COR_INTERNAL_ERROR,"Grouping in subsumption check.",
                    "subsumes_args()");

      default:
    fprintf(corexEnv->error_file,"Error! subsumes: default case yet to be implemented!\n");

    }
    return(COR_U_FAIL);
}

/*****************************************************************************/
int corsubsumes_nested_arg_lists(corArgList& left, corArgList& right,
    corBindEnv *left_bindings, corBindEnv *right_bindings, int can_bind)
{
#ifdef DO_TRACE
    if (corexEnv->dbg_subsumes) {
    fprintf(corexEnv->trace_file, 
        "[subsumes_nested_arg_listss(left&=0x%x", &left);
    fprintf(corexEnv->trace_file, ", right&=0x%x", &right);
    fprintf(corexEnv->trace_file, 
        ", left_bindings=0x%x, right_bindings=0x%x)]\n",
           left_bindings, right_bindings);
    }
#endif

    corTerm left_term, right_term;
    left_term.bindenv = left_bindings;
    right_term.bindenv = right_bindings;
    int i = left.count();
    if (right.count() != i) return COR_U_FAIL;
    for ( ; --i >= 0; ) {
    left_term.expr = left[i];
    right_term.expr = right[i];
        if (corsubsumes_args(left_term, right_term, can_bind) == COR_U_FAIL)
            return COR_U_FAIL;
    }
    return COR_U_SUCCEED;
}

/*****************************************************************************/
int corsubsumes_arg_lists(corArgList& left, corArgList& right,
    corBindEnv *left_bindings, corBindEnv *right_bindings)
{
#ifdef DO_PROFILE
      if (corT_Stack.count > 0)
    (corT_Stack.top()->moduledata->module_info.corGlobalProfileInfo.corsubsumes)++ ;
      else
    (corGlobalProfileInfo.corsubsumes)++ ;
#endif
#ifdef DO_TRACE
    if (corexEnv->dbg_subsumes) {
    fprintf(corexEnv->trace_file, "[subsumes_arg_lists(left&=0x%x", &left);
    fprintf(corexEnv->trace_file, ", right&=0x%x", &right);
    fprintf(corexEnv->trace_file, 
        ", left_bindings=0x%x, right_bindings=0x%x)]\n",
           left_bindings, right_bindings);
    }
#endif


    /********  ERROR -- TO BE FIXED
    // WARNING::  ReturnUnify optimizations needed here!!!
    .. fix subsumes to assume that one of the arguments is in 
    ..   a normal form where there are no cross bindenv bindings.
    ..   we can then use this fact to avoid checks for all_free etc.
    ..   Perhaps we can pass in the bindenv that is in normal form as an
    ..   argument when called from subsumes_tuple, so that all_free is 
    ..   used for all current used.
    *******************************/

#if 0
    /* ERROR:  By commenting out the following code, we make an assumption 
    that left bindings never cross into
    right.
    MAKE SURE that this assumption is valid (esp. in ord-search)
    */
    if ( ! ( right_bindings==NULL || 
         (left_bindings != NULL && left_bindings->all_free()) ||
         left.isConstant() // ||
         // left_bindings->is_versioned();
       ) ) {
    fprintf(corexEnv->error_file,"System Error: subsumes_arg_lists works only %s %s",
        "when either right_bindings is null, \n \t "
        "or left_bindings is all_free.\n");
    return(COR_U_FAIL);
    }
#endif 

    return corsubsumes_nested_arg_lists(left, right, left_bindings, 
        right_bindings, CAN_BIND);
}

/*****************************************************************************/
int corsubsumes(corBindEnv* env1, corArgList& arg2, corBindEnv* env2)
{
    return(corsubsumes_arg_lists(arg2, arg2, env1, env2));
}

/*****************************************************************************/
/*
    Checks if tuple subsumes newtuple, and prints a message if display is
    set.  If tuple does not have a bindenv it creates one of the required
    size.
*/

int corsubsumes_tuple( corTuple *tuple, corTuple *newtuple, int display)
{
    corStackMark unify_mark;
    int stat;

static corArrayBindEnv NilEnv (0);
    if (tuple->bindenv) {
    stat = corsubsumes_arg_lists(tuple->args(), newtuple->args(),
            tuple->bindenv, newtuple->bindenv);
    unify_mark.pop_to();
    }
    else if(tuple->env_size!=0) {
    corArrayBindEnv tuple_env(tuple->env_size);
    stat = corsubsumes_arg_lists(tuple->args(), newtuple->args(), 
        &tuple_env, newtuple->bindenv);
    unify_mark.pop_to();
     }
     else {
    stat = corsubsumes_arg_lists(tuple->args(), newtuple->args(), 
            &NilEnv, newtuple->bindenv);
        unify_mark.pop_to();
    }

    if (stat && display) {
    fprintf(corexEnv->trace_file, ") - subsumed by (");
    tuple->printon(corexEnv->trace_file);
    fprintf(corexEnv->trace_file, ")]\n");
    }
    return stat;
}

/*****************************************************************************/
int corsubsumes_value(AggregateKind aggkind, corCArg val1, corCArg val2, corBindEnv *env1,
        corBindEnv *env2) 
{

    // Returns 1 if val1 (new value) subsumes val2 (old value),
    //   -1 if val2 (old value) subsumes val1 (new value) and
    // returns 0 if the values are equal.

    
    corStackMark stackmark;
    corTerm term1(val1, env1);
    corTerm term2(val2, env2);
    corfull_dereference_term(term1);
    corfull_dereference_term(term2);

    // First dispose off any selection
    if (aggkind == AnyAggregation)  {
      /***************
      if (! term1.isConstant() || ! term2.isConstant()) {
            fprintf(exEnv->error_file,
          "Error: Aggregate Selection: aggregated argument not a constant\n");
          return 0;
      }
      ****************/

      int stat = corunify_args(term1,term2);
      stackmark.pop_to();
        if (stat==COR_U_SUCCEED)
          return 0;
      else return -1;  // new value is subsumed if don't unify
    }

    if (term1.expr.kindof() != COR_NUM_CONST ||
        term2.expr.kindof() != COR_NUM_CONST ) { 
      fprintf(corexEnv->error_file,
	      "Aggregate Selection: aggregated argument not a number\n");
    stackmark.pop_to();
        return 0;
     }
    int stat = (term1.expr.compare_num(term2.expr));
     // stat is pos if term1.expr is greater, neg if lesser, and 0 if equal

    if (aggkind == MinAggregation)
        stat = -stat;
     else if (aggkind != MaxAggregation) {
    CORAL_error(COR_BAD_ANNO, "Bad aggregation type", "subsumes_value");
    stat = 0;
     }
   
     stackmark.pop_to();
     return stat;
}

/*****************************************************************************/
//    Checks if each of tuple1 and tuple2 subsumes the other.

int corequal_tuples( corTuple *tuple1, corTuple *tuple2, int display)
{
    return (corsubsumes_tuple(tuple1, tuple2, display) &&
        corsubsumes_tuple(tuple2, tuple1, display));
}

/*****************************************************************************/
static int is_answer_fact( corTupleIterator& iterator, corTuple *)
{
    if (iterator.relation->r_kind == COR_R_ANSWER)
    return 1;
    return 0;
}

/*****************************************************************************/
/*
    Shady check - assumes answer fact is the delta relation, and hence the
    last arg must be a free var.  If supfact is to be fetched, the last arg
    is the goalid arg, and would have been bound. 
*/

static void bind_free_iterator_args (corTupleIterator& iterator, corTuple *tuple) {

    // NOTE:  this is called only for the supplementary literal, and hence 
    //        there should be no nested structures.

    int i;
    corCArg arg;

    for(i=0; i < iterator.arg_list.count(); i++) {
       arg = iterator.arg_list.arg(i);
       if ( arg.kindof() == COR_VARIABLE ) {
             corTerm term (arg, iterator.bindenv);
         corfull_dereference_term(term);
             if (term.kindof() == COR_VARIABLE) {   // Bind it 
                  corTerm tuple_term(tuple->arg(i), iterator.tuple_env);
                  corinline_unify_binding(iterator.bindenv, term.varnum(), 
                tuple_term);
             }
       }
    }
}

/*****************************************************************************/
int corunify_literal(corTupleIterator& iterator, corTuple *tuple, int display)
{

  corStackMark unify_mark;
  
  if (corCurrModuleInfo->NonGroundFacts) {

    if ((USING_RETURN_UNIFY) && (iterator.unification_number == 2) &&
    (iterator.relation->r_kind == COR_R_SUPPLEMENTARY)) { 
      if (iterator.prev_tuple->parid == tuple->contid) {
#ifdef DO_TRACE
    if (corexEnv->dbg_return_unify)
      fprintf(corexEnv->trace_file, 
          "** Return unify - succeeded\n");
#endif
    iterator.tuple_env = iterator.prev_bindenv;
    bind_free_iterator_args(iterator, tuple);
    iterator.reset_no_match(); 
#ifdef DO_TRACE
    if (display) {
      fprintf(corexEnv->trace_file,
          "<subclass>Relation: get_next_tuple: ");
      iterator.relation->name->print(NULL,corexEnv->trace_file);
      fprintf(corexEnv->trace_file,"(");
      tuple->printon(corexEnv->trace_file);
      fprintf(corexEnv->trace_file,
          ") [env_size=%d] \n", tuple->env_size);
    }
#endif
    return 1;
    
      }
    }
  }


  corTuple *newtuple = NULL;  // Renamed version of given tuple

  if (corCurrModuleInfo->NonGroundFacts) {

    if (USING_RETURN_UNIFY && (iterator.unification_number > 2))
    {
    // rename the tuple, and resize iterator.bindenv.
    // Don't do this for the first literal, since we can avoid renaming 
    //    facts for that literal when return-unification succeeds.
    //    If return-unification fails, the renaming is done in a lazy
    //    manner.
    newtuple = tuple->rename(iterator.prev_bindenv);
        iterator.tuple_env = newtuple->bindenv;
    tuple = newtuple;
    }
    else { 
        // why do this if there are only ground facts ?? - PRAVEEN
    iterator.tuple_env = tuple->make_bindenv_version();
    }
  }

  if (corunify_arg_lists(iterator.arg_list, tuple->args(),
              iterator.bindenv, iterator.tuple_env )) {
/****
  if (unify_arg_lists(iterator.temp_arglist, tuple->args(),
              iterator.bindenv, iterator.tuple_env )) {
***/
        // Remove this line ?? ___ EFFICIENCY1
    iterator.reset_no_match(); 
#ifdef DO_TRACE
    if (display) {
        fprintf(corexEnv->trace_file, "<subclass>Relation: get_next_tuple: ");
        iterator.relation->name->print(NULL,corexEnv->trace_file);
        fprintf(corexEnv->trace_file,"(");
        tuple->printon(corexEnv->trace_file);
        fprintf(corexEnv->trace_file, ") [env_size=%d] \n", tuple->env_size);
    }
#endif

    return 1 ;
    }

#ifdef DO_PROFILE
    if (corT_Stack.count > 0)
       (corT_Stack.top()->moduledata->module_info.
         corGlobalProfileInfo.non_unifying_tuples)++ ;
      else
        (corGlobalProfileInfo.non_unifying_tuples)++ ;
#endif

    unify_mark.pop_to();

  if (corCurrModuleInfo->NonGroundFacts) {
    // All this is only for non-ground facts !!
    if (iterator.tuple_env) {
    delete iterator.tuple_env;
    iterator.tuple_env = NULL;
    }

    if (newtuple && newtuple != tuple) {
    // rename created a new tuple, which is no longer required
    delete newtuple;  // Unification failed.
    }
  }
  return 0;
}

/*****************************************************************************/

