/*****************************************************************************/
//    Database.C
/*****************************************************************************/

#include "Database.h"

#include "SymbolTable.h"
#include "CoralError.h"
#include "Globals.h"
#include "RelationMisc.h"
#include "Interface.h"

/*****************************************************************************/

corDatabaseStruct corBuiltinDB(NULL);
corDatabaseStruct *corCurDB = &corBuiltinDB;
struct corSymbolTable *corDatabaseTable = NULL;

/*****************************************************************************/
Association corclear_relation(Pointer , Association* ass)
{
    if (! ((corRelation *)(ass->val))){
        return *ass;
    }

    if (! --(((corRelation *)(ass->val))->ref_count)) {
        // PRAVEEN:CHECK maybe this can be empty_relation(1) ???
        ((corRelation *)(ass->val))->empty_relation(0);
        delete (corRelation *)(ass->val);
        ass->val = NULL;
    }
    return *ass;
}

/*****************************************************************************/
corDatabaseStruct *corlookupDB(Name name)
{
    Association *ptr;
    if (corDatabaseTable == NULL)
        corDatabaseTable = corAllocTab(COR_DATABASE_TABLE_INCR, NULL);
    
    ptr = corSymbolLookup(corDatabaseTable, name);
    
    if (HashNone(ptr->arg))
        return 0;
    else
        return (corDatabaseStruct *)(ptr->val);
}

/*****************************************************************************/
corDatabaseStruct* corcreateDB(Name name)
{
    char err_msg[35];
    corDatabaseStruct *db;

    if (db = corlookupDB(name)) {
        sprintf(err_msg, "%10s.. exists already", corSymbolString(name));
        err_msg[34] = '\0';
        CORAL_error(COR_MULTIPLE_DB_ERR, err_msg, "createDB");
        return NULL;
    }

    Association *ptr;
    ptr = corSymbolLookup(corDatabaseTable, name);

    db = new corDatabaseStruct(name);
    ptr->arg = (Name)name;
    ptr->val = db;
    corTabInserted(corDatabaseTable, ptr);
    db->RelationTable = corAllocTab(COR_RELATION_TABLE_INCR, NULL);
    return db;
}

/*****************************************************************************/
static Association clear_database(Pointer , Association* ass)
{
    if (strcmp(corSymbolString(((corDatabaseStruct *)(ass->val))->name),COR_BUILTINS))
    {
        delete (corDatabaseStruct *)(ass->val);
    }
    return *ass;
}

/*****************************************************************************/
// used by exit_coral to remove all DBs

void corclearDatabases()
{
    if (corDatabaseTable) {
        corForEachTabElement(corDatabaseTable, (Func)clear_database, 0);
        delete corDatabaseTable;
    }
}

/*****************************************************************************/
// Used by init_coral to add the default DB to the DatabaseTable

int coraddDBStruct(corDatabaseStruct *db)
{
    if (db) {
        if (corlookupDB(db->name)) {
            return 0;
        }

        Association *ptr;
        ptr = corSymbolLookup(corDatabaseTable, db->name);

        ptr->arg = (Name)(db->name);
        ptr->val = db;
        corTabInserted(corDatabaseTable, ptr);
        return 1;
    }
    else
        return 0;
}

/*****************************************************************************/
int cordestroyDB(Name name)
{
    char err_msg[35];

    if (name == corCurDB->name) {
        CORAL_error(COR_MULTIPLE_DB_ERR, "cannot destroy current DB",
            "destroyDB");
        return 0;
    }

    if (!strcmp(corSymbolString(name), COR_BUILTINS)) {
        fputs("Cannot destroy the builtin database.\n",corexEnv->error_file);
        return(0);
    }

    corDatabaseStruct *db;
    if (!(db = corlookupDB(name))) {
        sprintf(err_msg, "%10s does not exist", corSymbolString(name));
        err_msg[34] = '\0';
        CORAL_error(COR_MULTIPLE_DB_ERR, err_msg, "destroyDB");
        return(0);
    }

    if (!corDeleteInTab(corDatabaseTable, name)) {
        sprintf(err_msg, "%10s does not exist", corSymbolString(name));
        err_msg[34] = '\0';
        CORAL_error(COR_MULTIPLE_DB_ERR, err_msg, "destroyDB");
        return(0);
    }

    delete db;
    return(1);
}

/*****************************************************************************/
int corcurrentDB(Name name)
{
    char err_msg[35];
    corDatabaseStruct *db;

    if (!strcmp(corSymbolString(name), COR_BUILTINS)) {
        fputs("Cannot make the builtin database the current database\n",
            corexEnv->error_file);
        return(0);
    }

    if (!(db = corlookupDB(name))) {
        sprintf(err_msg, "%10s does not exist", corSymbolString(name));
        err_msg[34] = '\0';
        CORAL_error(COR_MULTIPLE_DB_ERR, err_msg, "currentDB");
        return 0;
    }
    corCurDB = db;
    return 1;
}

/*****************************************************************************/
int correnameDB(Name dbname, Name newName)
{
    char err_msg[35];
    corDatabaseStruct *db;

    if (!strcmp(corSymbolString(dbname), COR_BUILTINS)) {
        fputs("Cannot rename the builtin database\n",corexEnv->error_file);
        return 0;
    }

    if (!(db = corlookupDB(dbname))) {
        sprintf(err_msg, "%10s does not exist", corSymbolString(dbname));
        err_msg[34] = '\0';
        CORAL_error(COR_MULTIPLE_DB_ERR, err_msg, "renameDB");
        return 0;
    }

    if (corlookupDB(newName)) {
        sprintf(err_msg, "%10s already exists", corSymbolString(newName));
        err_msg[34] = '\0';
        CORAL_error(COR_MULTIPLE_DB_ERR, err_msg, "renameDB");
        return 0;
    }

    corDeleteInTab(corDatabaseTable, dbname);
    Association *ptr;
    ptr = corSymbolLookup(corDatabaseTable, newName);
    ptr->arg = (Name)newName;
    ptr->val = db;
    corTabInserted(corDatabaseTable, ptr);
    db->name = newName;
    return 1;
}

/*****************************************************************************/
int correnameRelation(Name relname, Name newName)
{
    Association *ptr;
    corRelation *oldrel;

    ptr = corSymbolLookup(corCurDB->RelationTable, relname);
    if (HashNone(ptr->arg)) {
        fprintf(corexEnv->error_file, "%s does not exist", corSymbolString(relname));
        return 0;
    }
    oldrel = (corRelation *)ptr->val;

    ptr = corSymbolLookup(corCurDB->RelationTable, newName);
    if (!HashNone(ptr->arg)) {
        fprintf(corexEnv->error_file, "%s already exists", corSymbolString(newName));
        return 0;
     }

    corDeleteInTab(corCurDB->RelationTable, relname);

    ptr->arg = (Name)newName;
    ptr->val = oldrel;
    corTabInserted(corDatabaseTable, ptr);
    oldrel->name = newName;
    return 1;
}

/*****************************************************************************/
static Association dump_relation(Pointer outf, Association* ass)
{
    ((corRelation *)(ass->val))->print_facts((FILE *)outf);
    return *ass;
}

/*****************************************************************************/
int corsaveDB(Name dbname, FILE *outf)
{
    char err_msg[35];
    corDatabaseStruct *db;

    if (!dbname)
        db = corCurDB;
    else if (!strcmp(corSymbolString(dbname), COR_BUILTINS)) {
        fputs("Cannot save the builtin database\n",corexEnv->error_file);
        return 0;
    }
    else if (!(db = corlookupDB(dbname))) {
        sprintf(err_msg, "%10s does not exist", corSymbolString(dbname));
        err_msg[34] = '\0';
        CORAL_error(COR_MULTIPLE_DB_ERR, err_msg, "saveDB");
        return 0;
    }

    fprintf(outf, "db_create(%s).\n", corSymbolString(db->name));
    fprintf(outf, "db_current(%s).\n", corSymbolString(db->name));
    corForEachTabElement(db->RelationTable, (Func)dump_relation, (Pointer)outf);

    return 1;
}

/*****************************************************************************/
int corcopyToDB(corFuncArg *srcrel, Name dbname, corFuncArg *dstrel)
{
    corDatabaseStruct *db;
    char err_msg[100];

    if (!(db = corlookupDB(dbname))) {
        sprintf(err_msg, "%10s DB does not exist", corSymbolString(dbname));
        CORAL_error(COR_MULTIPLE_DB_ERR, err_msg, "copyToDB");
        return 0;
    }

    Association *ptr = corSymbolLookup(db->RelationTable, srcrel->functor());
    if (HashNone(ptr->arg)) {
        sprintf(err_msg,"rel %10s does not exist in DB %s",
            corSymbolString(srcrel->functor()), corSymbolString(dbname));
        CORAL_error(COR_BAD_REL_NAME, err_msg, "copyToDB");
        return 0;
    }
    corRelation *rel1 = (corRelation *)(ptr->val);

    ptr = corSymbolLookup(corCurDB->RelationTable, dstrel->functor());
    if (!HashNone(ptr->arg)) {
        sprintf(err_msg,"rel %10s ..already exists in current DB",
            corSymbolString(dstrel->functor()));
        CORAL_error(COR_BAD_REL_NAME, err_msg, "copyToDB");
        return 0;
    }

    corRelation *rel2 = cormake_relation(dstrel->functor(), dstrel->arity());

    corTuple *tuple;
    FOR_EACH_TUPLE(tuple, rel1) {
        rel2->insert_new(tuple);
    } END_EACH_TUPLE(tuple)

    return 1;
}

/*****************************************************************************/
int coraddToDB(corFuncArg *srcrel, Name dbname, corFuncArg *dstrel)
{
    corDatabaseStruct *db;
    char err_msg[100];

    if (!(db = corlookupDB(dbname))) {
        sprintf(err_msg, "%10s DB does not exist", corSymbolString(dbname));
        CORAL_error(COR_MULTIPLE_DB_ERR, err_msg, "addToDB");
        return 0;
    }

    Association *ptr = corSymbolLookup(db->RelationTable, srcrel->functor());
    if (HashNone(ptr->arg)) {
        sprintf(err_msg,"rel %10s does not exist in DB %s",
            corSymbolString(srcrel->functor()), corSymbolString(dbname));
        CORAL_error(COR_BAD_REL_NAME, err_msg, "addToDB");
        return 0;
    }
    corRelation *rel1 = (corRelation *)(ptr->val);

    ptr = corSymbolLookup(corCurDB->RelationTable, dstrel->functor());
    if (HashNone(ptr->arg)) {
        sprintf(err_msg,"rel %10s does not exist in current DB",
            corSymbolString(dstrel->functor()));
        CORAL_error(COR_BAD_REL_NAME, err_msg, "addToDB");
        return 0;
    }
    corRelation *rel2 = (corRelation *)(ptr->val);

    corTuple *tuple;
    FOR_EACH_TUPLE(tuple, rel1) {
        rel2->insert_new(tuple);
    } END_EACH_TUPLE(tuple)

    return 1;
}

/*****************************************************************************/
int corremoveRelFromDB(corFuncArg *rel, Name dbname)
{
    char err_msg[45];
    corDatabaseStruct *db;

    if (dbname == corCurDB->name)
        db = corCurDB;
    else if (!(db = corlookupDB(dbname))) {
        sprintf(err_msg, "db %10s does not exist", corSymbolString(dbname));
        err_msg[44] = '\0';
        CORAL_error(COR_MULTIPLE_DB_ERR, err_msg, "removeRelInDB");
        return 0;
    }

    Association *ptr = corSymbolLookup(db->RelationTable, rel->functor());
    if (HashNone(ptr->arg)) {
        sprintf(err_msg,"rel %10s does not exist",
            corSymbolString(rel->functor()));
        err_msg[44] = '\0';
        CORAL_error(COR_MULTIPLE_DB_ERR, err_msg, "removeRelInDB");
        return 0;
    }
    else {
        if (! --(((corRelation *)(ptr->val))->ref_count)) {
            ((corRelation *)(ptr->val))->empty_relation(0);
            delete (corRelation *)(ptr->val);
        }
        corDeleteInTab(db->RelationTable, rel->functor());
    }
    return 1;
}

/*****************************************************************************/
int corclearRelInDB(corFuncArg *rel, Name dbname)
{
    char err_msg[45];

    corDatabaseStruct *db;
    if (dbname == corCurDB->name)
        db = corCurDB;
    else if (!(db = corlookupDB(dbname))) {
        sprintf(err_msg, "db %10s does not exist", corSymbolString(dbname));
        err_msg[44] = '\0';
        CORAL_error(COR_MULTIPLE_DB_ERR, err_msg, "clearRelInDB");
        return 0;
    }

    Association *ptr = corSymbolLookup(db->RelationTable, rel->functor());
    if (HashNone(ptr->arg)) {
        sprintf(err_msg,"rel %10s does not exist",
            corSymbolString(rel->functor()));
        err_msg[44] = '\0';
        CORAL_error(COR_MULTIPLE_DB_ERR, err_msg, "clearRelInDB");
        return 0;
    }
    else
        ((corRelation *)(ptr->val))->empty_relation(1);

    return 1;
}

/*****************************************************************************/
static void InsertRel(corDatabaseStruct *db, Name n, int ,  corRelation *r)
{
    if (!db->RelationTable)
        db->RelationTable = corAllocTab(COR_RELATION_TABLE_INCR, NULL);

    corInsert(db->RelationTable, n, r);
}


/*****************************************************************************/
int coropenRelInCurDB(corFuncArg *rel, Name dbname)
{
    char err_msg[35];
    corDatabaseStruct *db;

    if (!strcmp(corSymbolString(dbname), COR_BUILTINS)) {
        fputs("Access already exists to builtin relations\n",corexEnv->error_file);
        return 0;
    }
    else if (!(db = corlookupDB(dbname))) {
        sprintf(err_msg, "%10s does not exist", corSymbolString(dbname));
        err_msg[34] = '\0';
        CORAL_error(COR_MULTIPLE_DB_ERR, err_msg, "openRelInDB");
        return 0;
    }

    Association *ptr = corSymbolLookup(db->RelationTable, rel->functor());
    if (!HashNone(ptr->arg)) {
        InsertRel(corCurDB, rel->functor(), rel->arity(), (corRelation *)ptr->val);
        (((corRelation *)(ptr->val))->ref_count)++;
        return 1;
    }
    
    CORAL_error(COR_REL_NOT_FOUND, rel->functor()->string(),"OpenRelInCurDB");
    return 0;
}

/*****************************************************************************/
static Association displayDBName(Pointer , Association *ass)
{
    fprintf(corexEnv->error_file, "%s\n" , corSymbolString((corSymbol *)(ass->arg)));
    return *ass;
}

/*****************************************************************************/
int corlistEveryDB()
{
    corForEachTabElement(corDatabaseTable, (Func)displayDBName, 0);
    return 1;
}


/*****************************************************************************/
Association cordisplay_relation_name(Pointer outf, Association *ass)
{
    ((corRelation *)(ass->val))->print_name((FILE *)outf);

    return *ass;
}

/*****************************************************************************/
int corlistRelsInDB(Name dbname, FILE *outfile )
{
    char err_msg[35];
    corDatabaseStruct *db;

    if (!(db = corlookupDB(dbname))) {
        sprintf(err_msg, "%10s does not exist", corSymbolString(dbname));
        err_msg[34] = '\0';
        CORAL_error(COR_MULTIPLE_DB_ERR, err_msg, "currentDB");
        return 0;
    }

    if (db->RelationTable) {
        corForEachTabElement(db->RelationTable, (Func)cordisplay_relation_name,
            outfile);
    }

    return 1;
}

/*****************************************************************************/

