//**************************************************************************
//    PersistSolvers.C
//**************************************************************************

#ifdef WITH_PERSISTENCE

#include <stdio.h>
#include <string.h>
#include "BuiltinRelation.h"
#include "PersistentRelation.h"

//**************************************************************************

extern void deletePersistentRelation(Name, int);

extern int SM_initialized ;

//**************************************************************************
//
//  Solver function that is called when the builtin relation schema()
//  is invoked.
//
//**************************************************************************

corBindEnv *SchemaSolver(corBuiltinRelation&, corTupleIterator& iterator)
{
    int cnt = 0;
    char err_str[100];

    if (!SM_initialized) {
        CORAL_error(COR_ESM_ERR, "dbserver not initialized", "SchemaSolver");
        iterator.set_no_match();
        return iterator.bindenv;
    }

    if (iterator.arg_list.count() > 0) {
        FOR_EACH_ARG(arg, iterator.arg_list) {
            cnt++;
            if (arg->kindof() != COR_FUNCTOR) {
                sprintf(err_str, "bad argument %d - should be a functor", cnt);
                CORAL_error(COR_BAD_ARG_TYPE, err_str, "SchemaSolver");
                iterator.set_no_match();
                return iterator.bindenv;
            }
            
            // Create a file descriptor for the new relation
            int error = 0;
            corFileDescStruct *fd = new corFileDescStruct(arg->FuncArgP(), error);
            
            // Add the descriptor to the catalogs
            if (!error) {
                SM.addFile((arg->FuncArgP())->functor(),
                    (arg->FuncArgP())->arity(), fd);
            }
            else {
                CORAL_error(error, corSymbolString((arg->FuncArgP())->functor()),
                    "SchemaSolver");
                delete fd;
            }

            // make a new persistent relation
            new corPersistentRel((arg->FuncArgP())->functor(),
                (arg->FuncArgP())->arity(), error);
            delete fd;
        } END_EACH_ARG
    }
    return iterator.bindenv;            
}

//**************************************************************************
//
//  Solver function that is called when the builtin relation del_schema()
//  is invoked.
//
//**************************************************************************

corBindEnv* DelSchemaSolver(corBuiltinRelation&, corTupleIterator& iterator)
{
    int cnt = 0;
    char err_str[100];

    if (!SM_initialized) {
        CORAL_error(COR_ESM_ERR, "dbserver not initialized", "SchemaSolver");
        iterator.set_no_match();
        return iterator.bindenv;
    }

    if (iterator.arg_list.count() > 0) {
        FOR_EACH_ARG(arg, iterator.arg_list) {
            cnt++;
            if (arg->kindof() != COR_FUNCTOR) {
                sprintf(err_str, "bad argument %d - should be a functor", cnt);
                CORAL_error(COR_BAD_ARG_TYPE, err_str, "SchemaSolver");
                iterator.set_no_match();
                return iterator.bindenv;
            }
            corFuncArg *f_arg = arg->FuncArgP();
            deletePersistentRelation(f_arg->functor(), f_arg->arity());
        } END_EACH_ARG
    }
    return iterator.bindenv;            
}

//**************************************************************************
//
//  Solver function that is called when the builtin relation StartBulkLoad()
//  is invoked.
//
//**************************************************************************

corBindEnv *StartBulkLoadSolver(corBuiltinRelation& , corTupleIterator& iterator)
{
    //    COMMENTED OUT CODE IN PREVIOUS VERSION
    return iterator.bindenv;      
}

//**************************************************************************
//
//  Solver function that is called when the builtin relation EndBulkLoad()
//  is invoked.
//
//**************************************************************************

corBindEnv* EndBulkLoadSolver(corBuiltinRelation& , corTupleIterator& iterator)
{
    //    COMMENTED OUT CODE IN PREVIOUS VERSION
    return iterator.bindenv;      
}

//**************************************************************************
corBindEnv* TransBeginSolver(corBuiltinRelation& , corTupleIterator& iterator)
{
    if (!SM_initialized) {
        iterator.set_no_match();
        return NULL;
    }

    SM.startTrans();
    if (!SM_initialized) {
        iterator.set_no_match();
        return NULL;
    }
 
    return iterator.bindenv;
}

//**************************************************************************
corBindEnv* TransAbortSolver(corBuiltinRelation& , corTupleIterator& iterator)
{
    if (!SM_initialized) {
        iterator.set_no_match();
        return NULL;
    }

    SM.abortTrans();

    if (!SM_initialized) {
        iterator.set_no_match();
        return NULL;
    }
 
    return iterator.bindenv;
}

//**************************************************************************
corBindEnv* TransCommitSolver(corBuiltinRelation& , corTupleIterator& iterator)
{
    if (!SM_initialized) {
        iterator.set_no_match();
        return NULL;
    }

    if (!SM.commitTrans() || !SM_initialized) {
        iterator.set_no_match();
        return NULL;
    }
 
    return iterator.bindenv;
}

//**************************************************************************

#endif WITH_PERSISTENCE

