// *************************************************************************
//    Rewriting.C
// *************************************************************************

#include <unistd.h>

#include "Rewriting.h"

#include "Globals.h"
#include "CoralError.h"
#include "ParserStruct.h"
#include "ExportInfo.h"
#include "Unparse.h"
#include "Util.h"
#include "BuiltinUtil.h"

//**************************************************************************

#define TMP_REWRITE_FILE "/tmp/__rewrite.%s"
#define TMP_M_REWRITE_FILE "/tmp/__rewrite.%s.M"
#define TMP_E_REWRITE_FILE "/tmp/__rewrite.%s.E"

//**************************************************************************
#ifdef    SAVE_INTER
static int unlink(char *path){

    static cnt = 0;
    char cmmd[100];
    fprintf(stderr,"Not unlinking %s (copying to rewsave.%d)\n",path,cnt);
    sprintf(cmmd,"/bin/cp %s /tmp/rewsave.%d\n", path, cnt++);
    system(cmmd);
    return(0);
}
#endif    // SAVE_INTER
//**************************************************************************
/*
    Called from yyparse() at the end of a module.  For each exported
    predicate, all the rules are written out to a temporary file,
    magic is called on it, and the magic output file is placed on
    the scanner input stack.
*/

void corperform_rewriting(corParserStruct& corparserStruct)
{
    corExportInfo *export_info;
    
    if (corparserStruct.CurModule.ExportedPreds == NULL) {
        fprintf(corexEnv->error_file,
                "CORAL: Error - No exported query form.\n");
        return;
    }

    // create output file
    FILE *outf;
    int i;
    char tmp_rewrite[MAX_PATH_LEN];
    char tmp_e_rewrite[MAX_PATH_LEN];
    char tmp_m_rewrite[MAX_PATH_LEN];

    sprintf(tmp_rewrite, TMP_REWRITE_FILE, coral_pid);
    sprintf(tmp_e_rewrite, TMP_E_REWRITE_FILE, coral_pid);
    sprintf(tmp_m_rewrite, TMP_M_REWRITE_FILE, coral_pid);


    if (!(outf = fopen(tmp_rewrite, "w"))) {
        fprintf(corexEnv->error_file,
            "CORAL::error : unable to open temporary file\n");
        fprintf(corexEnv->error_file, "Aborting attempt to process module\n");
        return;
    }

    // Process each exported query form independently.
    for (export_info = corparserStruct.CurModule.ExportedPreds, i = 0;
        export_info != NULL; export_info = export_info->next, i++)
    {
        if (corparserStruct.CurModule.name) {
            fprintf(outf, "\nmodule %s%d.\n", 
                corSymbolString(corparserStruct.CurModule.name), i);
        }
        fprintf(outf, "export %s(", export_info->pred_name->string());
        if (export_info->adorn_symbol)
            fputs(export_info->adorn_symbol->string(), outf);
        fputs(").\n",outf);

        corparserStruct.CurModule.ConvertFunctions = 0;
        corprint_Annotations(corparserStruct.CurModule, outf,
            corparserStruct.make_index_annotations,
            corparserStruct.agg_sel_annotations,
            corparserStruct.AllowedAdornList,
            corparserStruct.MultisetAnnotations,
            corparserStruct.prioritize_annotations,
            corparserStruct.diskrel_annotations);

        FOR_EACH_RULE(rule, corparserStruct.rules) {
            corprint_rule(rule, outf);
        } END_EACH_RULE

        fputs("end_module.\n\n",outf) ;
    }

    // close output file
    fclose(outf);

    char magic_input_file[200];
    char term_str_E[200];
    char term_str_M[200];

    if (corparserStruct.CurModule.UseExistOpt) {
        if (corparseStack->cur_file_name())
            sprintf(term_str_E, "%s.E", corparseStack->cur_file_name());
        else 
            sprintf(term_str_E, "%s.E", "stdin");

        if (corexEnv->C_create_magic_default) {
            // check to see if the magic output file for the currently
            // consulted file already exists, else create it.
            if (!(outf = fopen(term_str_E, "r"))) { 
                if (!(outf = fopen(term_str_E, "w"))) { 
                    fprintf(corexEnv->error_file,
                        "CORAL::error : unable to create temporary file\n");
                    fprintf(corexEnv->error_file,
                        "Aborting attempt to process module\n");
                    unlink(tmp_rewrite);
                    return ;
                }
                else fclose(outf);
            }
            else fclose(outf);
         }

        // perform existential query optimization rewriting on the file
        if (!corcallEQO(tmp_rewrite, tmp_rewrite)) {
            fprintf(corexEnv->error_file, "CORAL::error: unable to call eqo\n");
            fprintf(corexEnv->error_file,
                "Aborting attempt to process module\n");
            unlink(tmp_rewrite);
            return;
        }

        // remove tmp_rewrite
        unlink(tmp_rewrite);

        if (corexEnv->C_create_magic_default) {
            // concatenate tmp_e_rewrite to the magic output file
            char buf[300];
            sprintf(buf, "cat %s >> %s", tmp_e_rewrite, term_str_E);
            if (system(buf)) {
                fprintf(corexEnv->error_file,
                    "Unable to save eqo output in %s\n", term_str_E);
            }
            
            // remove tmp_rewrite
            unlink(tmp_e_rewrite);

            strcpy(magic_input_file, term_str_E);
         }
        else
            strcpy(magic_input_file, tmp_e_rewrite);
    }
    else
        strcpy(magic_input_file, tmp_rewrite);

    // Now handle magic rewriting
    if (corparseStack->cur_file_name())
        sprintf(term_str_M, "%s.M", corparseStack->cur_file_name());
    else 
        sprintf(term_str_M, "%s.M", "stdin");

    if (corexEnv->C_create_magic_default) {
        // check to see if the magic output file for the currently
        // consulted file already exists, else create it.
        if (!(outf = fopen(term_str_M, "r"))) { 
            if (!(outf = fopen(term_str_M, "w"))) { 
                fprintf(corexEnv->error_file, 
                    "CORAL::error : unable to create temporary file\n");
                fprintf(corexEnv->error_file,
                    "Aborting attempt to process module\n");
                unlink(magic_input_file);
                return ;
            }
            else fclose(outf);
        }
        else fclose(outf);
     }
    
    // perform magic rewriting on the file
    if (!corcallMagic(magic_input_file, tmp_rewrite)) {
        fprintf(corexEnv->error_file, "CORAL::error : unable to call magic\n");
        fprintf(corexEnv->error_file, "Aborting attempt to process module\n");
        unlink(magic_input_file);
        return;
    }

    unlink(magic_input_file); 

    // open temporary magic output file and put the descriptor
    // on the scanner input stack
    if (!corconsultFile(tmp_m_rewrite, tmp_m_rewrite)) {
        CORAL_error(COR_FILE_NOT_FOUND, tmp_m_rewrite,
            "perform_rewriting", __FILE__, __LINE__);
        fprintf(corexEnv->error_file, "Aborting attempt to process module\n");
    }

    if (corexEnv->C_create_magic_default) {
        // concatenate tmp_m_rewrite to the magic output file
        char buf[300];
        sprintf(buf, "cat %s >> %s", tmp_m_rewrite, term_str_M);
        if (system(buf)) {
             fprintf(corexEnv->error_file, 
                 "Unable to save magic output in %s\n", term_str_M);
        }
     }

    // delete tmp_m_rewrite
    unlink(tmp_m_rewrite);
}

/*****************************************************************************/

