//**************************************************************************
//    Solvers.C
//**************************************************************************

#include <unistd.h>
#include <string.h>
#include <math.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <limits.h>
#include <ctype.h>

#include <sys/time.h>
#include <sys/resource.h>

#include "Solvers.h"

#include "Config.h"
#include "Globals.h"
#include "CoralError.h"
#include "Interface.h"

#include "ArgList.h"
#include "FuncArg.h"
#include "BuiltinSymbols.h"
#include "TupleIterator.h"
#include "Tuple.h"
#include "Term.h"
#include "BindEnv.h"
#include "HashSimpleRelation.h"
#include "Execute.h"
#include "DerivedRelation.h"
#include "BuiltinUtil.h"
#include "GlobalBuiltinVars.h"
#include "Util.h"
#include "Database.h"
#include "IncrLoader.h"
#include "CoralBreak.h"
#include "Suspend.h"
#include "Suspension.h"
#include "ConnectArray.h"
#include "QFMaterializedModuleData.h"
#include "ClientProtocol.h"

#include "Assert.h"

//**************************************************************************

#define NilArg corNilSymbol

#define TMP_CONSULT_FILE "/tmp/__consult.%s"

//**************************************************************************

#ifdef    SAVE_INTER
static int unlink(char *path){

    static cnt = 0;
    char cmmd[100];
    fprintf(stderr,"Not unlinking %s (copying to corsave.%d)\n",path,cnt);
    sprintf(cmmd,"/bin/cp %s /tmp/corsave.%d\n", path, cnt++);
    system(cmmd);
    return(0);
}

#endif   // SAVE_INTER

//**************************************************************************
//    cpp predefines certain names, and replaces them.  In order to avoid this,
//    put any name you think might be a predefined name into this list.  Look
//    at the man page for cpp on each host machine.  This is a very partial
//    list so far.

#ifdef UNIX_HP
#define COR_CPP_OPTS   "-P -Umips -Usparc -Uhp -Uultrix -Usun -Uunix -Uhpux"
#else
#define COR_CPP_OPTS   "-P -B -Umips -Usparc -Uhp -Uultrix -Usun -Uunix -Uhpux"
#endif

//**************************************************************************

static void user_continue();
static int BP_Init = 0;            /* Indicates a set_break has been done */


//**************************************************************************
//      Handle print_module
//**************************************************************************
corBindEnv* corPrintModuleSolver(corBuiltinRelation&, corTupleIterator& iterator)

{
   int argcount = iterator.arg_list.count();

   if (argcount != 2) {
      CORAL_error(COR_EMPTY_ARGLIST, NULL, "PrintModuleSolver");
      return(iterator.bindenv);
   }

   corArgList *copy_args = new corArgList(argcount);
   corCopyArgs(copy_args, &iterator.arg_list,
      iterator.bindenv, argcount, NULL);

   char arg_char[100];  // relation name
   char ad[100];  // adornment
   int   pos = 0;
   ((*copy_args)[0]).sprint(arg_char,&pos);
   pos = 0;
   ((*copy_args)[1]).sprint(ad,&pos);

   // convert the adornment
   corBitVector adornment;
   if (corConvertAdornment(adornment, ad, strlen(ad)) == -1) {
     fprintf(corexEnv->error_file, "Bad kind of adornment.\n");
     return(iterator.bindenv);
   }

   Name rel_name = corEnterSymbol(arg_char);
   if (!corCurDB->RelationTable) {
     CORAL_error(COR_MODULE_NOT_FOUND, NULL, "PrintModuleSolver");
     return(iterator.bindenv);
   }

   Association *ptr = corSymbolLookup(corCurDB->RelationTable, rel_name);
   if (HashNone(ptr->arg)) {
     CORAL_error(COR_MODULE_NOT_FOUND, NULL, "PrintModuleSolver");
     return(iterator.bindenv);
   }

   corDerivedRelation *rel = (corDerivedRelation *)ptr->val;
   corDerivedMethod *method;
   for (method = rel->methods; method ; method = method->next) {
      if (!method->valid_method) continue;

      if (adornment==method->adornment) {
        corQFMaterializedModuleData *md = (corQFMaterializedModuleData *)method->md;
        if (!md)
          CORAL_error(COR_MODULE_NOT_FOUND, NULL, "PrintModuleSolver");
        else
          md->print(corexEnv->print_file);
          return(iterator.bindenv);

      }
   }
   CORAL_error(COR_MODULE_NOT_FOUND, NULL, "PrintModuleSolver");
   return(iterator.bindenv);
}

//**************************************************************************
// answer(n) -
//    This function changes the current user's num_answers.
//    This is the number of answers that will be returned for a query
//    before it is suspended.
//**************************************************************************
corBindEnv *corAnswerSolver(corBuiltinRelation& rel, corTupleIterator& iterator)
{
  if (iterator.arg_list.count() != 1) {
    fprintf(corexEnv->error_file, "Answers per query presently at %d\n",
        corcurUser->num_answers);
  } else {
    corTerm term1 (iterator.arg_list[0], iterator.bindenv);
    cordeep_dereference_term (term1);
    int n = make_int(term1.expr);
    if (n <= 0)
      corcurUser->num_answers = DEFLT_ANSWRS_PER_QRY;
    else
      corcurUser->num_answers = n;

    fprintf(corexEnv->print_file,"Number of answers set to: %d\n",n);
  }

  return(iterator.bindenv);
}

//**************************************************************************
// login("user_name", "passwd") -
//    This function creates a new user on this connection with the login
//    name given.  A numerical user id (valid this session only) is
//    echoed back.  There is presently no validation of the password.
//**************************************************************************
corBindEnv *corLoginSolver(corBuiltinRelation& rel, corTupleIterator& iterator)
{
  if (iterator.arg_list.count() !=2) {
    fprintf(corexEnv->error_file,
            "login(user_name, password). Two arguments needed.\n");
    iterator.set_no_match(); 
  } else {
    corTerm term1 (iterator.arg_list[0], iterator.bindenv);
    cordeep_dereference_term (term1);
    corTerm term2 (iterator.arg_list[1], iterator.bindenv);
    cordeep_dereference_term (term2);
    if ((term1.expr.kindof() != COR_SYMBOL_CONST) || 
        (term2.expr.kindof() != COR_SYMBOL_CONST)){

        fprintf(corexEnv->error_file,
            "login() - user_name and password must be strings.\n");
        iterator.set_no_match(); 
        return(iterator.bindenv);
    }

    char *u_name = corSymbolString(term1.expr);
    char *passwd = corSymbolString(term2.expr);
    int id = corcurConn->NewUser(u_name, passwd);
    if (id < 0){
        fputs("login failed.\n", corexEnv->error_file);
    } else {
        fprintf(corexEnv->print_file,"%s%d\n",CP_USER,id);
        fprintf(corexEnv->print_file,"%s%s\n",CP_LOGIN,u_name);
        fprintf(corexEnv->print_file,"%s%s\n",CP_PSWD,passwd);
        fprintf(corexEnv->print_file,"User %s logged in. Session id: %d\n",
            u_name, id ); 
    }
  }
  return(iterator.bindenv);
}

//**************************************************************************
// logout(id) -
//    This function removes a user from the connection, freeing up space
//    for another user.  A friendly client should logout all of its
//    users before closing its connection, but if it doesn't coral
//    will take care of the clean up. 
//**************************************************************************
corBindEnv *corLogoutSolver(corBuiltinRelation& rel, corTupleIterator& iterator)
{
  int id;

  if (iterator.arg_list.count() == 0){
      // Default - logout the current user
    id = corcurUser->user_id;

  } else if (iterator.arg_list.count() != 1) {

    fprintf(corexEnv->error_file,
        "Usage: logout. logout(user_id). or logout(login_name).\n");
    return(iterator.bindenv);
  } else {

    corTerm term1 (iterator.arg_list[0], iterator.bindenv);
    cordeep_dereference_term (term1);

    if (term1.expr.kindof() == COR_SYMBOL_CONST ) {

      char *nme = corSymbolString(term1.expr);

      if ((id = corcurConn->FindUser(nme)) < 0){
        fprintf(corexEnv->error_file, "User '%s' not found.\n", nme);
        return(iterator.bindenv);
      }

    } else {
      id = make_int(term1.expr);
      if ((id < 0) || (id > MAX_USRS_PER_CONN) || (!corcurConn->userArray[id])){
        fprintf(corexEnv->error_file,"Invalid id: %d\n", id);
        return(iterator.bindenv);
      }
    }
  }

  corcurConn->DeleteUser(id);
  return(iterator.bindenv);
}

//**************************************************************************
// who(). -
//    List all users logged in on all connections.
//**************************************************************************
corBindEnv *corWhoSolver(corBuiltinRelation& rel, corTupleIterator& iterator)
{
  int cnt;

  if (iterator.arg_list.count() == 0) {

    cnt = corclients->ListWho(corexEnv->print_file);
    if (!corexEnv->C_quiet_mode_default)
      fprintf(corexEnv->print_file,"Number of users: %d\n", cnt);

  } else {

    corTerm term1 (iterator.arg_list[0], iterator.bindenv);
    cordeep_dereference_term (term1);
    int conn = make_int(term1.expr);

    if ((conn < 0) || (conn > MAX_CONNECTIONS) || (!corclients->connArray[conn]))
      fprintf(corexEnv->error_file,"Invalid connection: %d\n", conn);

    else {
      cnt = corclients->ListWho(corexEnv->print_file,corclients->connArray[conn]);
      if (!corexEnv->C_quiet_mode_default)
        fprintf(corexEnv->print_file,"Number of users on connection %d: %d\n",
                conn, cnt);
    }
  }
  return(iterator.bindenv);
}

//**************************************************************************
// whoami() -
//    Tell who the current user is.
//**************************************************************************
corBindEnv *corWhoamiSolver(corBuiltinRelation& rel, corTupleIterator& iterator)
{
  if (iterator.arg_list.count() != 0) {
    fprintf(corexEnv->error_file,"whoami().: No arguments needed.\n");
  } else
    corcurUser->whoami(corexEnv->print_file);

  return(iterator.bindenv);
}

//**************************************************************************
// yield() -
//    Release the server to allow other connections a chance to use it.
//**************************************************************************
corBindEnv *corYieldSolver(corBuiltinRelation& rel, corTupleIterator& iterator)
{
  if (iterator.arg_list.count() != 0) 
    fprintf(corexEnv->error_file,"yield().: No arguments needed.\n");
  else {
    corcurUser->yield();
  }
  return(iterator.bindenv);
}

//**************************************************************************
// user(id) -
//    This function switches to the user specified, if they exist on this
//    connection.  If the user id specified is invalid, the user remains
//    unchanged.  A connection can always switch to id 0, which is user
//    "nobody".
//**************************************************************************
corBindEnv *corUserSolver(corBuiltinRelation& rel, corTupleIterator& iterator)
{
  int id = 0;

  if (iterator.arg_list.count() != 1) 
    fprintf(corexEnv->error_file,"Usage: user(user_id). or user(login_name).\n");
  else {
    corTerm term1 (iterator.arg_list[0], iterator.bindenv);
    cordeep_dereference_term (term1);

    if (term1.expr.kindof() == COR_SYMBOL_CONST ) {

      char *nme = corSymbolString(term1.expr);

      if ((id = corcurConn->FindUser(nme)) < 0){
        fprintf(corexEnv->error_file, "User '%s' not found.\n", nme);
        return(iterator.bindenv);
      }

    } else {

      id = make_int(term1.expr);
      if ((id < 0) || (id > MAX_USRS_PER_CONN) || (!corcurConn->userArray[id])){
        fprintf(corexEnv->error_file,"Invalid id: %d\n", id);
        return(iterator.bindenv);
      }
    }

    corcurConn->useUser(id);
      // Update the client
    fprintf(corexEnv->print_file,"%s%d\n",CP_USER, id);
  }
  return(iterator.bindenv);
}

//**************************************************************************
// kill() -
//    Delete a suspended query.
//**************************************************************************
corBindEnv *corKillSolver(corBuiltinRelation& rel, corTupleIterator& iterator)
{
  if (iterator.arg_list.count() != 1) {
    fprintf(corexEnv->error_file,"syntax: kill(id). One argument needed.\n");
    iterator.set_no_match(); 
  } else {
    corTerm term1 (iterator.arg_list[0], iterator.bindenv);
    cordeep_dereference_term (term1);
    int id = make_int(term1.expr);
    corsuspension.kill(id);
  }

  return(iterator.bindenv);
}

//**************************************************************************
// list_sus() -
//    List suspended queries.
//    list_sus(user_id) for certain person(user_id>=0),
//    list_sus() for current user, list_sus(all) for all users on
//    this connection. 
//**************************************************************************
corBindEnv *corListSusSolver(corBuiltinRelation& rel, corTupleIterator& iterator)
{
  int id;

  if (iterator.arg_list.count() == 0) 
    corsuspension.list(corcurUser->user_id);  // default
  else {
    corTerm term1 (iterator.arg_list[0], iterator.bindenv);
    cordeep_dereference_term (term1);
    if (term1.expr.kindof() == COR_SYMBOL_CONST ) {
      char *s = corSymbolString(term1.expr);

      if (!strcmp(s, "all")) 
        corsuspension.list(-2);    // all users

      else if ((id = corcurConn->FindUser(s)) < 0)
        fprintf(corexEnv->error_file, "User '%s' not found.\n", s);

      else 
        corsuspension.list(id);

    } else {
      int id = make_int(term1.expr);
      if ((id < 0) || (id > MAX_USRS_PER_CONN) || (!corcurConn->userArray[id]))
        fprintf(corexEnv->error_file,"Invalid id: %d\n", id);
      else
        corsuspension.list(id);
    }
  }
  return(iterator.bindenv);
}

//**************************************************************************
// resume() -
//    Resume a suspended query.
//    resume(sq_id) to resume, a user may only resume their own suspended
//    queries.
//**************************************************************************
corBindEnv *corResumeSolver(corBuiltinRelation& rel, corTupleIterator& iterator)
{
  if (iterator.arg_list.count() != 1) {
    fprintf(corexEnv->error_file,"syntax: resume(id). One argnument needed.\n");
    iterator.set_no_match(); 
    return(iterator.bindenv);
  } else {
    corTerm term1 (iterator.arg_list[0], iterator.bindenv);
    cordeep_dereference_term (term1);
    int id = make_int(term1.expr);
    // fprintf(exEnv->print_file,"going to resume %d\n", id);
    corsuspension.resume(id);
    return(iterator.bindenv);
  }
}

//**************************************************************************
// sync() -
//    Print out a sync message for the client.  Sync with a single argument
//    prints out that argument too.
//**************************************************************************
corBindEnv *corSyncSolver(corBuiltinRelation& rel, corTupleIterator& iterator)
{
  if (iterator.arg_list.count() == 1) {
    corTerm term1 (iterator.arg_list[0], iterator.bindenv);
    cordeep_dereference_term (term1);

    if (term1.expr.kindof() == COR_SYMBOL_CONST ) {
      char *s = corSymbolString(term1.expr);

      fprintf(corexEnv->print_file,"%s%s\n", CP_SYNC, s);

    } else {

      int id = make_int(term1.expr);
      fprintf(corexEnv->print_file,"%s%d\n", CP_SYNC, id);
    }
  } else {
    fprintf(corexEnv->print_file,"%s\n", CP_SYNC);
  }

  return(iterator.bindenv);
}

//**************************************************************************
//    Handle coral break
//**************************************************************************
corBindEnv* corBreakSolver(corBuiltinRelation&, corTupleIterator& iterator)
{
    int argcount = iterator.arg_list.count();

    if (!BP_Init) {
        BP_Init = 1;
        corb.all = 0;
        corb.module_flag = 1;
        corb.scc_flags     = new corBitVector(50);
        for (int j=0; j<=49; j++) {
            (corb.rule_flags[j]) = NULL;
        }
    }

    if (argcount <= 0) {
        CORAL_error(COR_EMPTY_ARGLIST, NULL, "BreakSolver");
        return(iterator.bindenv);
    }

    corArgList *copy_args = new corArgList(argcount);
    corCopyArgs(copy_args, &iterator.arg_list,
        iterator.bindenv, argcount, NULL);

    char arg_char[100];
    int    pos = 0;
    int    num;

    ((*copy_args)[0]).sprint(corb.module_name,&pos);
    strcat(corb.module_name,"0");

    for (int i=1; i <= argcount - 1; i++) {
        pos = 0;
        ((*copy_args)[i]).sprint(arg_char,&pos);
        int len = strlen(arg_char);
        if ((len > 3) && (strncmp(arg_char,"scc", 3) == 0)) {
            int index = 3;
            int done = 0;
            while ((arg_char[index] != '\0') && (!done)) {
                if (!isdigit(arg_char[index])) done = 1;
                index++;
            }

            if (done) {
                //    BUILTIN_FAIL_RETURN
                return(0);
            }
            else {
                int s_num = atoi((char*)(arg_char+3));
                corb.scc_flags->set(s_num);
                num = s_num;
            }
        }
        else if ((len > 4) && (strncmp(arg_char, "rule", 4) == 0)) {
            int index = 4;
            int done = 0;

            if (!(corb.rule_flags[num-1]))
                (corb.rule_flags[num-1]) = new corBitVector(50);

            while ((arg_char[index] != '\0') && (!done)) {
                if (!isdigit(arg_char[index])) done = 1;
                index++;
            }

            if (done) {
                //    BUILTIN_FAIL_RETURN
                return(0);
            }
            else {
                int r_num = atoi((char*)(arg_char+4));
                (corb.rule_flags[num-1])->set(r_num);
            }
        }
        else if ((len == 3) && (strcmp(arg_char, "all") == 0)) {
            corb.all = 1;
        }
        else {
            //    BUILTIN_FAIL_RETURN
            return(0);
        }
    }
    return(iterator.bindenv);
}

//**************************************************************************
#define FAIL_RETURN  {iterator.set_no_match(); return(NULL);}

corBindEnv *corClearBreakSolver(corBuiltinRelation&, corTupleIterator& iterator)
{
    int argcount = iterator.arg_list.count();

    if (!BP_Init) {
        fprintf(corexEnv->error_file, "No breakpoints are currently set.\n");
        //    BUILTIN_FAIL_RETURN
        return(0);
    }

    if (argcount <= 0) {
        CORAL_error(COR_EMPTY_ARGLIST, NULL, "ClearBreakSolver");
        return(iterator.bindenv);
    }

    corArgList *copy_args = new corArgList(argcount);
    corCopyArgs(copy_args, &iterator.arg_list,
        iterator.bindenv, argcount, NULL);

    char arg_char[100];
    int    pos = 0;
    int    num;

    ((*copy_args)[0]).sprint(corb.module_name,&pos);
    strcat(corb.module_name,"0");

        /* Only the module name is given */
    if (argcount == 1)
        corb.module_flag = 0;

    for (int i=1; i <= argcount - 1; i++) {
        pos = 0;
        ((*copy_args)[i]).sprint(arg_char,&pos);
        int len = strlen(arg_char);
        if ((len > 3) && (strncmp(arg_char,"scc", 3) == 0)) {
            int index = 3;
            int done = 0;
            while ((arg_char[index] != '\0') && (!done)) {
                if (!isdigit(arg_char[index])) done = 1;
                index++;
            }

            if (done) {
                //    BUILTIN_FAIL_RETURN
                return(0);
            }
            else {
                int s_num = atoi((char*)(arg_char+3));
                corb.scc_flags->set(s_num,0);
                num = s_num;
            }
        }
        else if ((len > 4) && (strncmp(arg_char, "rule", 4) == 0)) {
            int index = 4;
            int done = 0;

            if (!(corb.rule_flags[num-1]))
                continue;

            while ((arg_char[index] != '\0') && (!done)) {
                if (!isdigit(arg_char[index])) done = 1;
                index++;
            }

            if (done) {
                //    BUILTIN_FAIL_RETURN
                return(0);
            }
            else {
                int r_num = atoi((char*)(arg_char+4));
                (corb.rule_flags[num-1])->set(r_num,0);
            }
        }
        else if ((len == 3) && (strcmp(arg_char, "all") == 0)) {
            corb.module_flag = 0;
            if (corb.all)
                corb.all = 0;
            else {
                    // Turn off all (like really all) breakpoints
                (corb.scc_flags)->clear();
                int j;
                for (j=0; j<50; j++)
                    if (corb.rule_flags[j])
                        (corb.rule_flags[j])->clear();
            }
        }
        else {
            //    BUILTIN_FAIL_RETURN
            return(0);
        }
    }
    return(iterator.bindenv);
}

//**************************************************************************
#define FAIL_RETURN  {iterator.set_no_match(); return(NULL);}

corBindEnv* corUnivSolver(corBuiltinRelation& , corTupleIterator& iterator)
{
    if (iterator.arg_list.count() != 2) {
        fprintf(corexEnv->error_file, "Two arguments required for Univ.\n");
        FAIL_RETURN;
    }
    
    corTerm term1 (iterator.arg_list[0], iterator.bindenv);
    cordeep_dereference_term(term1);

    corTerm term2 (iterator.arg_list[1], iterator.bindenv);
    cordeep_dereference_term(term2);

    if (term1.expr.kindof() == COR_VARIABLE) {
        //if (term2.expr.kindof() != COR_FUNCTOR)
        if (!is_list(term2.expr)) {
            fprintf(corexEnv->error_file, "Second argument must be a list\n");
            FAIL_RETURN;
        }
        corTerm first_arg_term(make_car(term2.expr), iterator.bindenv);
        cordeep_dereference_term(first_arg_term);

        if (first_arg_term.expr.kindof() != COR_SYMBOL_CONST) {
            fprintf(corexEnv->error_file,
                "First element of list must be functor name\n");
            FAIL_RETURN;
        }
        corTerm rest_arg_term(make_cdr(term2.expr), iterator.bindenv);
        cordeep_dereference_term(rest_arg_term);

        if ((rest_arg_term.expr.FuncArgP())->functor() == corConsSymbol) {
            corArgList * arg_list;
            corArgList * temp_list = new corArgList(1);
            corTerm    temp_term;
            temp_term.expr = make_car(rest_arg_term.expr);
            temp_term.bindenv = iterator.bindenv;
            cordeep_dereference_term(temp_term);
            arg_list = new corArgList(1);
            (*arg_list)[0] = temp_term.expr;
            int count = 1;
            for (;;) {
                rest_arg_term.expr = make_cdr(rest_arg_term.expr);
                rest_arg_term.bindenv = iterator.bindenv;
                cordeep_dereference_term(rest_arg_term);

                if ((rest_arg_term.expr.FuncArgP())->functor() == corConsSymbol)
                {
                    temp_term.expr = make_car(rest_arg_term.expr);
                    temp_term.bindenv = iterator.bindenv;
                    cordeep_dereference_term(temp_term);
                    (* temp_list)[0] = temp_term.expr;
                    arg_list = concat_arglists(arg_list, temp_list);
                    count++;
                }
                else if (rest_arg_term.expr.other == NilArg)
                    break;
                else {
                    fprintf(corexEnv->error_file, "Error in    list argument.\n");
                    return(NULL);
                }
            }
        
            corFuncArg *fun_arg = corFuncArg::New(first_arg_term.expr.SymbolP(), count);
            int i;
        
            for (i = 0; i < count; i++)
                (fun_arg->args)[i] = (*arg_list)[i];
            
            corTerm new_term(fun_arg, iterator.bindenv);
            corStackMark stackmark;
            if (!corunify_args(term1, new_term)) {
                stackmark.pop_to();
                FAIL_RETURN;
            }
            return(iterator.bindenv);
        }
        else if (rest_arg_term.expr.other == NilArg) {
            corFuncArg * fun_arg = corFuncArg::New(first_arg_term.expr.SymbolP(), 0);

            corTerm new_term(fun_arg, iterator.bindenv);
            corStackMark stackmark;
            if (!corunify_args(term1, new_term)) {
                stackmark.pop_to();
                FAIL_RETURN;
            }
            return(iterator.bindenv);
        }
        else {
            fprintf(corexEnv->error_file, "List type needed as second argument.\n");
            FAIL_RETURN;
        }
    }
    
    if (term1.expr.kindof() != COR_FUNCTOR) {
        fprintf(corexEnv->error_file, "First argument must be a functor\n");
        FAIL_RETURN;
    }

    int num = (term1.expr.FuncArgP())->arity();
    corArg * name_arg = make_arg(functor_name(term1.expr));
    if (num == 0) {
        //make list with just the name 
        corArg * arg_list = make_cons(name_arg, corNilSymbol);

        corTerm new_term(arg_list, iterator.bindenv);
        corStackMark stackmark;
        if (!corunify_args(term2, new_term)) {
            stackmark.pop_to();
            FAIL_RETURN;
        }
        return(iterator.bindenv);
    }

    // else forming the list 

    corArg    * arg_list = make_cons((term1.expr.FuncArgP())->args[num-1], corNilSymbol);
    int i;
    for (i = (num-2); i >= 0; i--)
        arg_list = make_cons((term1.expr.FuncArgP())->args[i], arg_list);
    
    arg_list = make_cons(name_arg, arg_list);
    
    corTerm new_term(arg_list, iterator.bindenv);
    corStackMark stackmark;
    if (!corunify_args(term2, new_term)) {
        stackmark.pop_to();
        FAIL_RETURN;
    }
    return(iterator.bindenv);
}

#undef FAIL_RETURN

//**************************************************************************
corBindEnv* corCallSolver(corTupleIterator& iterator)
{
    corBindEnv *env = iterator.bindenv;

    corRelation  *rel;
    corTupleIterator *iter;

    if (iterator.ipos++ == 0) {
        iterator.stack_mark.get_mark();
        int num = iterator.arg_list.arity();
        if (num != 1) {
            fprintf(corexEnv->error_file, "Call takes only one argument.\n");
            iterator.set_no_match();
            //return(iterator.bindenv);
            return(NULL);
        }
        corTerm term(iterator.arg_list[0], env);
        cordeep_dereference_term(term);

        /* Is this necessary ?? PRAVEEN
          if ( term.expr.kindof() == COR_VARIABLE )
          term = env->lookup(((VarArg *)term.expr)->var);
        */

        if ( term.expr.kindof() != COR_FUNCTOR ) {
            fprintf(corexEnv->error_file,
                "Wrong type of argument supplied to call.\n");
            iterator.set_no_match();
            //return(iterator.bindenv);
            return(NULL);
        }
        corFuncArg * rel_name = term.expr.FuncArgP();
        rel = find_relation(rel_name->functor(), rel_name->arity());
        if (!rel) {
            fprintf(corexEnv->error_file,"Relation %s with arity %d not found.\n",
                (rel_name->functor())->string(), rel_name->arity());
            iterator.set_no_match();
            //return(iterator.bindenv);
            return(NULL);
        }
    
        // check arity as find_relation does not
        if ((rel->arity() > 0) && (rel->arity() != rel_name->arity())) {
            fprintf(corexEnv->error_file,"Relation arity does not match.\n");
            iterator.set_no_match();
            // return(iterator.bindenv);
            return(NULL);
        }

        iter = new corTupleIterator(rel, rel_name->args, env);
        iterator.answer_iterator = iter;
    }
    else {
        iterator.stack_mark.pop_to();
        iter = iterator.answer_iterator;
        rel = iter->relation;
    }

    corBindEnv * result = iter->get_next();
    if (iter->no_match()) {
        iterator.set_no_match();
        return(NULL);
    }
    else {
        iterator.reset_no_match();
        return(result);
    }
}

//**************************************************************************
corBindEnv* corMemberSolver(corTupleIterator& iterator)
{
    corBindEnv *env = iterator.bindenv;

    corRelation *rel;
    corArgList *args;
    corTupleIterator *iter;
    if (iterator.ipos++ == 0) {
        iterator.stack_mark.get_mark();
        int num = iterator.arg_list.arity() - 1;

        args = new corArgList(num);

        corTerm term(iterator.arg_list[0], env);
        cordeep_dereference_term(term);

        if ( term.expr.kindof() == COR_FUNCTOR ) {
            corFuncArg * rel_name = term.expr.FuncArgP();
            rel = find_relation(rel_name->functor(), rel_name->arity());

            if (!rel) {
                fprintf(corexEnv->error_file,
                    "Relation %s with arity %d not found.\n",
                    (rel_name->functor())->string(), rel_name->arity());
                iterator.set_no_match();
                return(iterator.bindenv);
            }
        }
        else if ( term.expr.kindof() == COR_SYMBOL_CONST ) {
            rel = find_relation(term.expr.SymbolP(), -1);

            if (!rel) {
                fprintf(corexEnv->error_file, "Relation %s not found.\n",
                    (term.expr.SymbolP())->string());
                iterator.set_no_match();
                return(iterator.bindenv);
            }
        }
        else if ( term.expr.kindof() == COR_RELATION )
            rel = term.expr.RelationP();
        else {
            fprintf(corexEnv->error_file,
            "Wrong 1st argument provided to Member: %s\n",
            corSymbolString(term.expr));
        
            iterator.set_no_match();
            return(iterator.bindenv);
        }

        if ((rel->arity() > 0) && (rel->arity() != num)) {
            fprintf(corexEnv->error_file,"Relation arity does not match.\n");
            iterator.set_no_match();
            return(iterator.bindenv);
        }

        for (int i = 1; i <= num; i++)
            (*args)[i-1] = iterator.arg_list[i];

        iter = new corTupleIterator(rel, *args, env);
        iterator.answer_iterator = iter;
    }
    else {
        iterator.stack_mark.pop_to();
        iter = iterator.answer_iterator;
        rel = iter->relation;
    }

    corBindEnv * result = iter->get_next();
    if (iter->no_match()) {
        iterator.set_no_match();
        return(NULL);
    }
    else {
        iterator.reset_no_match();
        return(result);
    }
}

//**************************************************************************
corBindEnv* corMemTupleSolver(corTupleIterator& iterator)
{
    corBindEnv *env = iterator.bindenv;

    corRelation *rel;
    corTupleIterator *iter;

    if (iterator.ipos++ == 0) {
        iterator.stack_mark.get_mark();
        //int num = iterator.arg_list.arity() - 1;

        corTerm term(iterator.arg_list[0], env);
        cordeep_dereference_term(term);
        if ( term.expr.kindof() == COR_FUNCTOR ) {
            corFuncArg * rel_name = term.expr.FuncArgP();
            rel = find_relation(rel_name->functor(), rel_name->arity());

            if (!rel) {
                fprintf(corexEnv->error_file,
                    "Relation %s with arity %d not found.\n",
                    (rel_name->functor())->string(), rel_name->arity());
                iterator.set_no_match();
                return(iterator.bindenv);
            }
        }
        else if ( term.expr.kindof() == COR_SYMBOL_CONST ) {
            rel = find_relation(term.expr.SymbolP(), -1);

            if (!rel) {
                fprintf(corexEnv->error_file,"Relation %s not found.\n",
                    (term.expr.SymbolP())->string());
                iterator.set_no_match();
                return(iterator.bindenv);
            }
        }
        else if ( term.expr.kindof() == COR_RELATION )
            rel = term.expr.RelationP();
        else {
            fprintf(corexEnv->error_file,
                "Wrong first argument provided to MemTuple.\n");
            iterator.set_no_match();
            return(iterator.bindenv);
        }
        corArgList * args = cormake_vararglist(rel->arity());
        corArrayBindEnv *env0 = new corArrayBindEnv(rel->arity());
        iter = new corTupleIterator(rel, *args, env0);
        iter->stack_mark.get_mark();
        iterator.answer_iterator = iter;
    }
    else {
        iterator.stack_mark.pop_to();
        // iter->stack_mark.pop_to();
        // delete iter->bindenv;
        iter = iterator.answer_iterator;
        rel = iter->relation;
    }

    for (;;) {
        corTuple * ret_tuple = iter->get_next_tuple();
        if (iter->no_match()) {
            iterator.set_no_match();
            return(NULL);
        }
        corFuncArg * func = (corFuncArg *)make_arg("", ret_tuple->_args);

        corTerm lterm(iterator.arg_list[1], env);
        corTerm rterm(func, iter->bindenv);
        if (corunify_args(lterm, rterm))
        {
            iterator.reset_no_match();
            return(iterator.bindenv);
        }
    }
}

//**************************************************************************
corBindEnv* corMultiUnionSolver(corTupleIterator& iterator)
{
    corBindEnv *env = iterator.bindenv;
    if (iterator.ipos > 0) {
        iterator.stack_mark.pop_to();
        iterator.set_no_match(); return(env);
    }
    iterator.stack_mark.get_mark();
    iterator.ipos++;

    if ((iterator.arg_list).count() != 3) {
        fprintf(corexEnv->error_file,
            "Error: unionmax: wrong number of arguments.\n");
        return(NULL);
    }

    corTerm terms[3];
    for (int i = 0; i < 3; i++) {
        terms[i].expr = iterator.arg_list[i];
        terms[i].bindenv = env;
        cordeep_dereference_term(terms[i]);
    }

    corRelation *rels[2];
    for (i = 0; i < 2; i++) {
      switch (terms[i].kindof()) {
        case COR_RELATION:
            rels[i] = terms[i].expr.RelationP();
            break;
        case COR_FUNCTOR:
            rels[i] = find_relation((terms[i].expr.FuncArgP())->_functor,
               (terms[i].expr.FuncArgP())->args.count());
            if (rels[i] == NULL) {
                CORAL_error(COR_REL_NOT_FOUND, NULL, "get_next");
                iterator.stack_mark.pop_to();
                return(NULL);
            }
            break;
        case COR_SYMBOL_CONST:
            rels[i] = find_relation(terms[i].expr.SymbolP(), -1);
            if (rels[i] == NULL) {
                CORAL_error(COR_REL_NOT_FOUND, NULL, "get_next");
                iterator.stack_mark.pop_to();
                return(NULL);
            }
            break;
        default:
            CORAL_error(COR_BAD_ARG_TYPE, NULL, "get_next");
            iterator.stack_mark.pop_to();
            return(NULL);
        }
    }
    corStackMark stack_mark0;

    corArrayBindEnv * env0 = new corArrayBindEnv(rels[0]->arity());
    corArgList * args0 = cormake_vararglist(rels[0]->arity());
    corArrayBindEnv * env1 = new corArrayBindEnv(rels[1]->arity());
    corArgList * args1 = cormake_vararglist(rels[1]->arity());
    corArrayBindEnv * env2 = new corArrayBindEnv(rels[1]->arity());
    corArgList * args2 = cormake_vararglist(rels[1]->arity());
    corArrayBindEnv * env3 = new corArrayBindEnv(rels[1]->arity());
    corArgList * args3 = cormake_vararglist(rels[1]->arity());

    if (rels[0]->arity() != rels[1]->arity()) {
        fprintf(corexEnv->error_file,
            "Error: Union: Both sets must have equal arity\n");
        iterator.stack_mark.pop_to();
        return(NULL);
    }

    corRelation *union_rel = new corHashSimpleRelation(rels[0]->arity());
    
    corTupleIterator *iter0 = new corTupleIterator(rels[0], *args0, env0);
    corTupleIterator *iter1 = new corTupleIterator(rels[1], *args1, env1);
    corTupleIterator *iter2 = new corTupleIterator(rels[1], *args2, env2);
    corTupleIterator *iter3 = new corTupleIterator(union_rel, *args3, env3);

    corTuple *ret_tuple;
    for(;;) {
        ret_tuple = iter0->get_next_tuple();
        if (iter0->no_match())
           break;

       union_rel->insert_new(ret_tuple);
    }
    
    // second relation
    corTuple *ret_tuple2;
    int _arity = rels[1]->arity();
    for (;;) {
        ret_tuple = iter1->get_next_tuple();
        if (iter1->no_match())
           break;

        // if there is a match find out how many copies there are
        corStackMark stack_mark2;
        iter2->reset();
        for ( int i = 0; i < _arity; i++)
            (iter2->arg_list)[i] = ret_tuple->arg(i);
        int rel2_count = 0;
        //    looping through the second relation
        for (;;) {
            ret_tuple2 = iter2->get_next_tuple();
            if (iter2->no_match()) {
                   stack_mark2.pop_to();
                  iter2->release();
                break;
            }
            rel2_count++;
        }

        // find out how many copies are there in the union
        corStackMark stack_mark3;
        iter3->reset();
        for ( i = 0; i < _arity; i++)
            (iter3->arg_list)[i] = ret_tuple->arg(i);
        int union_count = 0;
        // loop through the union
        for (;;) {
            ret_tuple2 = iter3->get_next_tuple();
            if (iter3->no_match()) {
                stack_mark3.pop_to();      
                  iter3->release();
                break;
             }
            union_count++;
        }
    
        // insert number of extra copies in union
        for (i = 0; i < (rel2_count - union_count); i++)    
        union_rel->insert_new(ret_tuple);
    }


    stack_mark0.pop_to();
    corTerm result_term;
    result_term.expr = union_rel;
    result_term.bindenv = NULL;
    if ( ! corunify_args (result_term, terms[2]))
      iterator.set_no_match();

    if (iterator.no_match())
        iterator.stack_mark.pop_to();
    return(env);
}

//**************************************************************************
corBindEnv* corUnionSolver(corTupleIterator& iterator)
{
    corBindEnv *env = iterator.bindenv;
    if (iterator.ipos > 0) {
        iterator.stack_mark.pop_to();
        iterator.set_no_match(); return(env);
    }
    iterator.stack_mark.get_mark();
    iterator.ipos++;

    if ((iterator.arg_list).count() != 3) {
        fprintf(corexEnv->error_file, "Error: unionsum: wrong number arguments.\n");
        return(NULL);
    }

    corTerm terms[3];
    for (int i = 0; i < 3; i++) {
        terms[i].expr = iterator.arg_list[i];
        terms[i].bindenv = env;
        cordeep_dereference_term(terms[i]);
    }

    corRelation *rels[2];
    for (i = 0; i < 2; i++) {
        switch (terms[i].kindof()) {
        case COR_RELATION:
            rels[i] = terms[i].expr.RelationP();
            break;
        case COR_FUNCTOR:
            rels[i] = find_relation((terms[i].expr.FuncArgP())->_functor,
               (terms[i].expr.FuncArgP())->args.count());
            if (rels[i] == NULL) {
                CORAL_error(COR_REL_NOT_FOUND, NULL, "get_next");
                iterator.stack_mark.pop_to();
                return(NULL);
            }
            break;
        case COR_SYMBOL_CONST:
            rels[i] = find_relation(terms[i].expr.SymbolP(), -1);
            if (rels[i] == NULL) {
                CORAL_error(COR_REL_NOT_FOUND, NULL, "get_next");
                iterator.stack_mark.pop_to();
                return(NULL);
            }
            break;
        default:
            CORAL_error(COR_BAD_ARG_TYPE, NULL, "get_next");
            iterator.stack_mark.pop_to();
            return(NULL);
        }
    }
    corStackMark stack_mark0;

    corArrayBindEnv * env0 = new corArrayBindEnv(rels[0]->arity());
    corArgList * args0 = cormake_vararglist(rels[0]->arity());
    corTupleIterator * iter0 = new corTupleIterator(rels[0], *args0, env0);

    corArrayBindEnv * env1 = new corArrayBindEnv(rels[1]->arity());
    corArgList * args1 = cormake_vararglist(rels[1]->arity());
    corTupleIterator * iter1 = new corTupleIterator(rels[1], *args1, env1);

    if (rels[0]->arity() != rels[1]->arity()) {
        fprintf(corexEnv->error_file,
            "Error: Union: Both sets must have equal arity\n");
        iterator.stack_mark.pop_to();
        return(NULL);
    }

    corRelation* union_rel = new corHashSimpleRelation(rels[0]->arity());

    corTuple *ret_tuple;
    for(;;) {
        ret_tuple = iter0->get_next_tuple();
        if (iter0->no_match())
            break;
        union_rel->insert_new(ret_tuple);
    }

    for (;;) {
        ret_tuple = iter1->get_next_tuple();
        if (iter1->no_match())
            break;
       union_rel->insert_new(ret_tuple);
    }

    stack_mark0.pop_to();
    corTerm result_term;
    result_term.expr = union_rel;
    result_term.bindenv = NULL;
    if ( ! corunify_args (result_term, terms[2]))
        iterator.set_no_match();

    if (iterator.no_match())
        iterator.stack_mark.pop_to();

    return(env);
}

//**************************************************************************
corBindEnv* corIntersectionSolver(corTupleIterator& iterator)
{
    corBindEnv *env = iterator.bindenv;
    if (iterator.ipos > 0) {
        iterator.stack_mark.pop_to();
        iterator.set_no_match(); return(env);
    }
    iterator.stack_mark.get_mark();
    iterator.ipos++;

    if ((iterator.arg_list).count() != 3) {
        fprintf(corexEnv->error_file, "Error: Inter: wrong number arguments.\n");
        return(NULL);
    }

    corTerm terms[3];
    for (int i = 0; i < 3; i++) {
        terms[i].expr = iterator.arg_list[i];
        terms[i].bindenv = env;
        cordeep_dereference_term(terms[i]);
    }

    corRelation *rels[2];
    for (i = 0; i < 2; i++) {
        switch (terms[i].kindof()) {
        case COR_RELATION:
            rels[i] = terms[i].expr.RelationP();
            break;
        case COR_FUNCTOR:
            rels[i] = find_relation((terms[i].expr.FuncArgP())->_functor,
               (terms[i].expr.FuncArgP())->args.count());
            if (rels[i] == NULL) {
                CORAL_error(COR_REL_NOT_FOUND, NULL, "get_next");
                iterator.stack_mark.pop_to();
                return(NULL);
            }
            break;
        case COR_SYMBOL_CONST:
            rels[i] = find_relation(terms[i].expr.SymbolP(), -1);
            if (rels[i] == NULL) {
                CORAL_error(COR_REL_NOT_FOUND, NULL, "get_next");
                iterator.stack_mark.pop_to();
                return(NULL);
            }
            break;
        default:
            CORAL_error(COR_BAD_ARG_TYPE, NULL, "get_next");
            iterator.stack_mark.pop_to();
            return(NULL);
        }
    }
    corStackMark stack_mark0;

    corArrayBindEnv *env0 = new corArrayBindEnv(rels[0]->arity());
    corArrayBindEnv *env1 = new corArrayBindEnv(rels[0]->arity());
    corArrayBindEnv *env2 = new corArrayBindEnv(rels[0]->arity());
    corArrayBindEnv *env3 = new corArrayBindEnv(rels[0]->arity());
    corArgList *args0 = cormake_vararglist(rels[0]->arity());
    corArgList *args1 = cormake_vararglist(rels[0]->arity());
    corArgList *args2 = cormake_vararglist(rels[0]->arity());
    corArgList *args3 = cormake_vararglist(rels[0]->arity());

    if (rels[0]->arity() != rels[1]->arity()) {
        fprintf(corexEnv->error_file,
            "Error: Intersection: Both sets must have equal arity\n");
        iterator.stack_mark.pop_to();
        return(NULL);
    }

    corRelation *intersection_rel = new corHashSimpleRelation(rels[0]->arity());

    corTuple *ret_tuple0;
    corTuple *ret_tuple1;
    corTupleIterator *iter0 = new corTupleIterator(rels[0], *args0, env0);
    corTupleIterator *iter1 = new corTupleIterator(rels[0], *args1, env1);
    corTupleIterator *iter2 = new corTupleIterator(rels[1], *args2, env2);
    corTupleIterator *iter3 = new corTupleIterator(intersection_rel, *args3, env3);
    int _arity = rels[1]->arity();
    // loop through outer relation
    for(;;) {
        ret_tuple0 = iter0->get_next_tuple();
        if (iter0->no_match())
            break;

        // see if tuple already in intersection
        corStackMark stack_mark3;
        iter3->reset();
        for (int i = 0; i < _arity; i++)
             (iter3->arg_list)[i] = ret_tuple0->arg(i);
        
        ret_tuple1 = iter3->get_next_tuple();
        if (!iter3->no_match()) {
            stack_mark3.pop_to();
            iter3->release();
            continue;
        }
                
        // loop through relation again to count number of copies
        corStackMark stack_mark1;
        iter1->reset();
        for (i = 0; i < _arity; i++)
             (iter1->arg_list)[i] = ret_tuple0->arg(i);
        int rel1_count = 0;
        for(;;) {
            ret_tuple1 = iter1->get_next_tuple();
            if (iter1->no_match()) {
                   stack_mark1.pop_to();
                   iter1->release();
                   break;
               }
            rel1_count++;
        }

        // loop through second relation to find number of copies
        corStackMark stack_mark2;
        iter2->reset();
        for (i = 0; i < _arity; i++)
             (iter2->arg_list)[i] = ret_tuple0->arg(i);
        int rel2_count = 0;
        for(;;) {
            ret_tuple1 = iter2->get_next_tuple();
            if (iter2->no_match()) {
                 stack_mark2.pop_to();
                 iter2->release();
                 break;
            }
            rel2_count++;
        }

        if (rel2_count == 0)
            continue;

        int min;
        if (rel2_count < rel1_count)
            min = rel2_count;
        else
            min = rel1_count;

        for ( i = 0; i < min ; i++)
            intersection_rel->insert_new(ret_tuple0);
    }

    stack_mark0.pop_to();
    corTerm result_term;
    result_term.expr = intersection_rel;
    result_term.bindenv = NULL;
    if ( ! corunify_args (result_term, terms[2]))
        iterator.set_no_match();

    if (iterator.no_match())
        iterator.stack_mark.pop_to();
    return(env);
}

//**************************************************************************
corBindEnv* corDiffSolver(corTupleIterator& iterator)
{
    corBindEnv *env = iterator.bindenv;
    if (iterator.ipos > 0) {
        iterator.stack_mark.pop_to();
        iterator.set_no_match(); return(env);
    }
    iterator.stack_mark.get_mark();
    iterator.ipos++;

    if ((iterator.arg_list).count() != 3) {
        fprintf(corexEnv->error_file,
            "Error: Difference: wrong number arguments.\n");
        return(NULL);
    }

    corTerm terms[3];
    for (int i = 0; i < 3; i++) {
        terms[i].expr = iterator.arg_list[i];
        terms[i].bindenv = env;
        cordeep_dereference_term(terms[i]);
    }

    corRelation *rels[2];
    for (i = 0; i < 2; i++) {
        switch (terms[i].kindof()) {
        case COR_RELATION:
            rels[i] = terms[i].expr.RelationP();
            break;
        case COR_FUNCTOR:
            rels[i] = find_relation((terms[i].expr.FuncArgP())->_functor,
               (terms[i].expr.FuncArgP())->args.count());
            if (rels[i] == NULL) {
                CORAL_error(COR_REL_NOT_FOUND, NULL, "get_next");
                iterator.stack_mark.pop_to();
                return(NULL);
            }
            break;
        case COR_SYMBOL_CONST:
            rels[i] = find_relation(terms[i].expr.SymbolP(), -1);
            if (rels[i] == NULL) {
                CORAL_error(COR_REL_NOT_FOUND, NULL, "get_next");
                iterator.stack_mark.pop_to();
                return(NULL);
            }
            break;
        default:
            CORAL_error(COR_BAD_ARG_TYPE, NULL, "get_next");
            iterator.stack_mark.pop_to();
            return(NULL);
        }
    }
    corStackMark stack_mark0;

    corArrayBindEnv *env0 = new corArrayBindEnv(rels[0]->arity());
    corArrayBindEnv *env1 = new corArrayBindEnv(rels[0]->arity());
    corArrayBindEnv *env2 = new corArrayBindEnv(rels[0]->arity());
    corArrayBindEnv *env3 = new corArrayBindEnv(rels[0]->arity());
    corArgList *args0 = cormake_vararglist(rels[0]->arity());
    corArgList *args1 = cormake_vararglist(rels[0]->arity());
    corArgList *args2 = cormake_vararglist(rels[0]->arity());
    corArgList *args3 = cormake_vararglist(rels[0]->arity());

    if (rels[0]->arity() != rels[1]->arity()) {
        fprintf(corexEnv->error_file,
            "Error: Intersection: Both sets must have equal arity\n");
        iterator.stack_mark.pop_to();
        return(NULL);
    }

    corRelation* diff_rel = new corHashSimpleRelation(rels[0]->arity());

    corTuple *ret_tuple0;
    corTuple *ret_tuple1;
    corTupleIterator *iter0 = new corTupleIterator(rels[0], *args0, env0);
    corTupleIterator *iter1 = new corTupleIterator(rels[0], *args1, env1);
    corTupleIterator *iter2 = new corTupleIterator(rels[1], *args2, env2);
    corTupleIterator *iter3 = new corTupleIterator(diff_rel, *args3, env3);
    int _arity = rels[1]->arity();

    // loop through outer relation
    for(;;) {
        ret_tuple0 = iter0->get_next_tuple();
        if (iter0->no_match())
            break;

        //    see if tuple already in difference
        corStackMark stack_mark3;
        iter3->reset();
        for (int i = 0; i < _arity; i++)
             (iter3->arg_list)[i] = ret_tuple0->arg(i);
        
        ret_tuple1 = iter3->get_next_tuple();
        if (!iter3->no_match()) {
            stack_mark3.pop_to();
            iter3->release();
            continue;
        }
                
        // loop through relation again to count number of copies
        corStackMark stack_mark1;
        iter1->reset();
        for (i = 0; i < _arity; i++)
            (iter1->arg_list)[i] = ret_tuple0->arg(i);
        int rel1_count = 0;
        for(;;) {
            ret_tuple1 = iter1->get_next_tuple();
            if (iter1->no_match()) {
                   stack_mark1.pop_to();
                   iter1->release();
                   break;
               }
            rel1_count++;
        }

        // loop through second relation to find number of copies
        corStackMark stack_mark2;
        iter2->reset();
        for (i = 0; i < _arity; i++)
            (iter2->arg_list)[i] = ret_tuple0->arg(i);
        int rel2_count = 0;
        for(;;) {
            ret_tuple1 = iter2->get_next_tuple();
            if (iter2->no_match()) {
                stack_mark2.pop_to();
                iter2->release();
                break;
            }
            rel2_count++;
        }

        for ( i = 0; i < (rel1_count - rel2_count) ; i++)
            diff_rel->insert_new(ret_tuple0);
    }

    stack_mark0.pop_to();
    corTerm result_term;
    result_term.expr = diff_rel;
    result_term.bindenv = NULL;
    if ( ! corunify_args (result_term, terms[2]))
        iterator.set_no_match();

    if (iterator.no_match())
        iterator.stack_mark.pop_to();

    return(env);
}

//**************************************************************************
corBindEnv* corSubsetSolver(corTupleIterator& iterator)
{
    corBindEnv *env = iterator.bindenv;
    if (iterator.ipos > 0) {
        iterator.stack_mark.pop_to();
        iterator.set_no_match(); return(env);
    }
    iterator.stack_mark.get_mark();
    iterator.ipos++;

    if ((iterator.arg_list).count() != 2) {
        fprintf(corexEnv->error_file,
            "Error: subset: wrong number of arguments.\n");
        return(NULL);
    }

    corTerm terms[2];
    for (int i = 0; i < 2; i++) {
        terms[i].expr = iterator.arg_list[i];
        terms[i].bindenv = env;
        cordeep_dereference_term(terms[i]);
    }

    corRelation *rels[2];
    for (i = 0; i < 2; i++) {
        switch (terms[i].kindof()) {
        case COR_RELATION:
            rels[i] = terms[i].expr.RelationP();
            break;
        case COR_FUNCTOR:
            rels[i] = find_relation((terms[i].expr.FuncArgP())->_functor,
               (terms[i].expr.FuncArgP())->args.count());
            if (rels[i] == NULL) {
                CORAL_error(COR_REL_NOT_FOUND, NULL, "get_next");
                iterator.stack_mark.pop_to();
                return(NULL);
            }
            break;
        case COR_SYMBOL_CONST:
            rels[i] = find_relation(terms[i].expr.SymbolP(), -1);
            if (rels[i] == NULL) {
                CORAL_error(COR_REL_NOT_FOUND, NULL, "get_next");
                iterator.stack_mark.pop_to();
                return(NULL);
            }
            break;
        default:
            CORAL_error(COR_BAD_ARG_TYPE, NULL, "get_next");
            iterator.stack_mark.pop_to();
            return(NULL);
        }
    }
    corStackMark stack_mark0;
    corArrayBindEnv *env0 = new corArrayBindEnv(rels[0]->arity());
    corArrayBindEnv *env1 = new corArrayBindEnv(rels[0]->arity());
    corArrayBindEnv *env2 = new corArrayBindEnv(rels[0]->arity());
    corArgList *args0 = cormake_vararglist(rels[0]->arity());
    corArgList *args1 = cormake_vararglist(rels[0]->arity());
    corArgList *args2 = cormake_vararglist(rels[0]->arity());

    if (rels[0]->arity() != rels[1]->arity()) {
        fprintf(corexEnv->error_file,
            "Error: Subset: Both sets must have equal arity\n");
        iterator.stack_mark.pop_to();
        return(NULL);
    }

    corRelation *subset_rel = new corHashSimpleRelation(rels[0]->arity());


    corTuple *ret_tuple0;
    corTuple *ret_tuple1;
    int no_match = 0;
    corTupleIterator *iter0 = new corTupleIterator(rels[0], *args0, env0);
    corTupleIterator *iter1 = new corTupleIterator(rels[0], *args1, env1);
    corTupleIterator *iter2 = new corTupleIterator(rels[1], *args2, env2);
    int _arity = rels[1]->arity();
    //    loop through outer relation
    for(;;) {
        ret_tuple0 = iter0->get_next_tuple();
        if (iter0->no_match())
            break;

        // loop through relation again to count number of copies
        corStackMark stack_mark1;
        iter1->reset();
        for (i = 0; i < _arity; i++)
            (iter1->arg_list)[i] = ret_tuple0->arg(i);
        int rel1_count = 0;
        for(;;) {
            ret_tuple1 = iter1->get_next_tuple();
            if (iter1->no_match()) {
                stack_mark1.pop_to();
                iter1->release();
                break;
            }
            rel1_count++;
        }

        /* loop through second relation to find number of copies */
        corStackMark stack_mark2;
        iter2->reset();
        for (i = 0; i < _arity; i++)
            (iter2->arg_list)[i] = ret_tuple0->arg(i);
        int rel2_count = 0;
        for(;;) {
            ret_tuple1 = iter2->get_next_tuple();
            if (iter2->no_match()) {
                stack_mark2.pop_to();
                iter2->release();
                break;
            }
            rel2_count++;
        }
        if (rel1_count > rel2_count) {
            no_match = 1;
            break;
        }
    }

    stack_mark0.pop_to();
    if (!no_match)
        iterator.reset_no_match();
    else {
        iterator.set_no_match();
        iterator.stack_mark.pop_to();
    }
    return(env);
}

//**************************************************************************
corBindEnv* corMakeSetSolver(corTupleIterator& iterator)
{
    corBindEnv *env = iterator.bindenv;
    if (iterator.ipos > 0) {
        iterator.stack_mark.pop_to();
        iterator.set_no_match(); return(env);
    }
    iterator.stack_mark.get_mark();
    iterator.ipos++;

    if ((iterator.arg_list).count() != 2) {
        fprintf(corexEnv->error_file, "Error: make_set: wrong number arguments.\n");
        return(NULL);
    }

    corTerm terms[2];
    for (int i = 0; i < 2; i++) {
        terms[i].expr = iterator.arg_list[i];
        terms[i].bindenv = env;
        cordeep_dereference_term(terms[i]);
    }

    corRelation *rel;
    switch (terms[0].kindof())
    {
        case COR_RELATION:
            rel = terms[0].expr.RelationP();
            break;
        case COR_FUNCTOR:
            rel = find_relation((terms[0].expr.FuncArgP())->_functor,
               (terms[0].expr.FuncArgP())->args.count());
            if (rel == NULL) {
                CORAL_error(COR_REL_NOT_FOUND, NULL, "get_next");
                iterator.stack_mark.pop_to();
                return(NULL);
            }
            break;
        case COR_SYMBOL_CONST:
            rel = find_relation(terms[0].expr.SymbolP(), -1);
            if (rel == NULL) {
                CORAL_error(COR_REL_NOT_FOUND, NULL, "get_next");
                iterator.stack_mark.pop_to();
                return(NULL);
            }
            break;
        default:
            CORAL_error(COR_BAD_ARG_TYPE, NULL, "get_next");
            iterator.stack_mark.pop_to();
            return(NULL);
    }
    corStackMark stack_mark0;

    corArrayBindEnv * env0 = new corArrayBindEnv(rel->arity());
    corArgList * args0 = cormake_vararglist(rel->arity());
    corArrayBindEnv * env1 = new corArrayBindEnv(rel->arity());
    corArgList * args1 = cormake_vararglist(rel->arity());

    corRelation  *set_rel = new corHashSimpleRelation(rel->arity());

    corTupleIterator * iter0 = new corTupleIterator(rel, *args0, env0);
    corTupleIterator * iter1 = new corTupleIterator(set_rel, *args1, env1);

    corTuple *ret_tuple0;
    corTuple *ret_tuple1;
    int _arity = rel->arity();
    for(;;) {
        ret_tuple0 = iter0->get_next_tuple();
        if (iter0->no_match())
            break;

        corStackMark stack_mark1;
        iter1->reset();
        for(int i = 0; i < _arity; i++)
            (iter1->arg_list)[i] = ret_tuple0->arg(i);

        ret_tuple1 = iter1->get_next_tuple();
        if (!iter1->no_match()) {
            stack_mark1.pop_to();
            iter1->release();
            continue;
        }

        stack_mark1.pop_to();
        iter1->release();
        set_rel->insert_new(ret_tuple0);
    }

    stack_mark0.pop_to();
    corTerm result_term;
    result_term.expr = set_rel;
    result_term.bindenv = NULL;
    if ( ! corunify_args (result_term, terms[1]))
      iterator.set_no_match();

    if (iterator.no_match())
        iterator.stack_mark.pop_to();

    return(env);
}

//**************************************************************************
corBindEnv* corUnionDSolver(corTupleIterator& iterator)
{
    corBindEnv *env = iterator.bindenv;

    if (iterator.ipos > 0) {
        iterator.stack_mark.pop_to();
        iterator.set_no_match(); return(env);
    }
    iterator.stack_mark.get_mark();
    iterator.ipos++;

    if ((iterator.arg_list).count() != 2) {
        fprintf(corexEnv->error_file,
            "Error: unionsum_d: wrong number of arguments.\n");
        return(NULL);
    }

    corTerm terms[2];
    for (int i = 0; i < 2; i++) {
        terms[i].expr = iterator.arg_list[i];
        terms[i].bindenv = env;
        cordeep_dereference_term(terms[i]);
    }

    corRelation *rels[2];
    for (i = 0; i < 2; i++) {
        switch (terms[i].kindof()) {
        case COR_RELATION:
            rels[i] = terms[i].expr.RelationP();
            break;
        case COR_FUNCTOR:
            rels[i] = find_relation((terms[i].expr.FuncArgP())->_functor,
               (terms[i].expr.FuncArgP())->args.count());
            if (rels[i] == NULL) {
                CORAL_error(COR_REL_NOT_FOUND, NULL, "get_next");
                iterator.stack_mark.pop_to();
                return(NULL);
            }
            break;
        case COR_SYMBOL_CONST:
            rels[i] = find_relation(terms[i].expr.SymbolP(), -1);
            if (rels[i] == NULL) {
                CORAL_error(COR_REL_NOT_FOUND, NULL, "get_next");
                iterator.stack_mark.pop_to();
                return(NULL);
            }
            break;
        default:
            CORAL_error(COR_BAD_ARG_TYPE, NULL, "get_next");
            iterator.stack_mark.pop_to();
            return(NULL);
        }
    }
    corStackMark stack_mark0;

    corArrayBindEnv * env0 = new corArrayBindEnv(rels[0]->arity());
    corArgList * args0 = cormake_vararglist(rels[0]->arity());
    corTupleIterator * iter0 = new corTupleIterator(rels[0], *args0, env0);

    if (rels[0]->arity() != rels[1]->arity()) {
        fprintf(corexEnv->error_file,
            "Error: union_d: Both sets must have equal arity\n");
        iterator.stack_mark.pop_to();
        return(NULL);
    }

    corTuple *ret_tuple;
    for(;;) {
        ret_tuple = iter0->get_next_tuple();
        if (iter0->no_match())
            break;

        rels[1]->insert_new(ret_tuple);
    }


    stack_mark0.pop_to();

    return(env);
}

//**************************************************************************
corBindEnv* corMultiUnionDSolver(corTupleIterator& iterator)
{
    corBindEnv *env = iterator.bindenv;

    if (iterator.ipos > 0) {
        iterator.stack_mark.pop_to();
        iterator.set_no_match(); return(env);
    }
    iterator.stack_mark.get_mark();
    iterator.ipos++;

    if ((iterator.arg_list).count() != 2) {
        fprintf(corexEnv->error_file,
            "Error: unionmax_d wrong number arguments.\n");
        return(NULL);
    }

    corTerm terms[2];
    for (int i = 0; i < 2; i++) {
        terms[i].expr = iterator.arg_list[i];
        terms[i].bindenv = env;
        cordeep_dereference_term(terms[i]);
    }

    corRelation *rels[2];
    for (i = 0; i < 2; i++) {
        switch (terms[i].kindof()) {
        case COR_RELATION:
            rels[i] = terms[i].expr.RelationP();
            break;
        case COR_FUNCTOR:
            rels[i] = find_relation((terms[i].expr.FuncArgP())->_functor,
               (terms[i].expr.FuncArgP())->args.count());
            if (rels[i] == NULL) {
                CORAL_error(COR_REL_NOT_FOUND, NULL, "get_next");
                iterator.stack_mark.pop_to();
                return(NULL);
            }
            break;
        case COR_SYMBOL_CONST:
            rels[i] = find_relation(terms[i].expr.SymbolP(), -1);
            if (rels[i] == NULL) {
                CORAL_error(COR_REL_NOT_FOUND, NULL, "get_next");
                iterator.stack_mark.pop_to();
                return(NULL);
            }
            break;
        default:
            CORAL_error(COR_BAD_ARG_TYPE, NULL, "get_next");
            iterator.stack_mark.pop_to();
            return(NULL);
        }
    }
    corStackMark stack_mark0;

    if (rels[0]->arity() != rels[1]->arity()) {
        fprintf(corexEnv->error_file,
            "Error: unionmax_d: Both sets must have equal arity\n");
        iterator.stack_mark.pop_to();
        return(NULL);
    }


    corTuple *ret_tuple0;
    corTuple *ret_tuple1;

    corArrayBindEnv * env0 = new corArrayBindEnv(rels[0]->arity());
    corArrayBindEnv * env1 = new corArrayBindEnv(rels[0]->arity());
    corArrayBindEnv * env2 = new corArrayBindEnv(rels[0]->arity());
    corArgList * args0 = cormake_vararglist(rels[0]->arity());
    corArgList * args1 = cormake_vararglist(rels[0]->arity());
    corArgList * args2 = cormake_vararglist(rels[0]->arity());

    corTupleIterator * iter0 = new corTupleIterator(rels[0], *args0, env0);
    corTupleIterator * iter1 = new corTupleIterator(rels[0], *args1, env1);
    corTupleIterator * iter2 = new corTupleIterator(rels[1], *args2, env2);

    int _arity = rels[1]->arity();

    // loop through first relation
    for(;;) {
        ret_tuple0 = iter0->get_next_tuple();
        if (iter0->no_match())
            break;

        // loop through both relations to count number of copies */

        // first relation...

        corStackMark stack_mark1;
        iter1->reset();

        for (i = 0; i < _arity; i++)
            (iter1->arg_list)[i] = ret_tuple0->arg(i);

        int rel1_count = 0;

        for(;;) {
            ret_tuple1 = iter1->get_next_tuple();
            if (iter1->no_match()) {
                stack_mark1.pop_to();
                iter1->release();
                break;
            }
              rel1_count++;
        }

        // second relation...


        corStackMark stack_mark2;
        iter2->reset();

        for (i = 0; i < _arity; i++)
            (iter2->arg_list)[i] = ret_tuple0->arg(i);

          int rel2_count = 0;

        for(;;) {
            ret_tuple1 = iter2->get_next_tuple();
            if (iter2->no_match()) {
                stack_mark2.pop_to();
                iter2->release();
                break;
            }
            rel2_count++;
        }

        int diff = rel1_count - rel2_count;
        if (diff <= 0) continue; // 2nd rel already has correct number

        // else insert diff many copies of ret_tuple0 into 2nd rel...

        while ( diff-- > 0 ) {
             rels[1]->insert_new(ret_tuple0);
        }
    }

    stack_mark0.pop_to();
    return(env);
}

//**************************************************************************
corBindEnv* corIntersectionDSolver(corTupleIterator& iterator)
{
    corBindEnv *env = iterator.bindenv;
    if (iterator.ipos > 0) {
        iterator.stack_mark.pop_to();
        iterator.set_no_match(); return(env);
    }
    iterator.stack_mark.get_mark();
    iterator.ipos++;

    if ((iterator.arg_list).count() != 2) {
        fprintf(corexEnv->error_file, "Error: inter_d wrong number arguments.\n");
        return(NULL);
    }

    corTerm terms[2];
    for (int i = 0; i < 2; i++) {
        terms[i].expr = iterator.arg_list[i];
        terms[i].bindenv = env;
        cordeep_dereference_term(terms[i]);
    }

    corRelation *rels[2];
    for (i = 0; i < 2; i++) {
        switch (terms[i].kindof()) {
        case COR_RELATION:
            rels[i] = terms[i].expr.RelationP();
            break;
        case COR_FUNCTOR:
            rels[i] = find_relation((terms[i].expr.FuncArgP())->_functor,
               (terms[i].expr.FuncArgP())->args.count());
            if (rels[i] == NULL) {
                CORAL_error(COR_REL_NOT_FOUND, NULL, "get_next");
                iterator.stack_mark.pop_to();
                return(NULL);
            }
            break;
        case COR_SYMBOL_CONST:
            rels[i] = find_relation(terms[i].expr.SymbolP(), -1);
            if (rels[i] == NULL) {
                CORAL_error(COR_REL_NOT_FOUND, NULL, "get_next");
                iterator.stack_mark.pop_to();
                return(NULL);
            }
            break;
        default:
            CORAL_error(COR_BAD_ARG_TYPE, NULL, "get_next");
            iterator.stack_mark.pop_to();
            return(NULL);
        }
    }
    corStackMark stack_mark0;

    if (rels[0]->arity() != rels[1]->arity()) {
        fprintf(corexEnv->error_file,
            "Error: inter_d: Both sets must have equal arity\n");
        iterator.stack_mark.pop_to();
        return(NULL);
    }

    corTuple *ret_tuple0;
    corTuple *ret_tuple1;

    corArrayBindEnv * env0 = new corArrayBindEnv(rels[0]->arity());
    corArrayBindEnv * env1 = new corArrayBindEnv(rels[0]->arity());
    corArrayBindEnv * env2 = new corArrayBindEnv(rels[0]->arity());
    corArgList * args0 = cormake_vararglist(rels[0]->arity());
    corArgList * args1 = cormake_vararglist(rels[0]->arity());
    corArgList * args2 = cormake_vararglist(rels[0]->arity());

    corTupleIterator * iter0 = new corTupleIterator(rels[0], *args0, env0);
    corTupleIterator * iter1 = new corTupleIterator(rels[0], *args1, env1);
    corTupleIterator * iter2 = new corTupleIterator(rels[1], *args2, env2);

    int _arity = rels[1]->arity();

    // loop through outer relation
    for(;;) {
        ret_tuple0 = iter0->get_next_tuple();
        if (iter0->no_match())
            break;

        // loop through both relations to count number of copies

        // first relation...

        corStackMark stack_mark1;
        iter1->reset();

        for (i = 0; i < _arity; i++)
            (iter1->arg_list)[i] = ret_tuple0->arg(i);

          int rel1_count = 0;

        for(;;) {
            ret_tuple1 = iter1->get_next_tuple();
            if (iter1->no_match()) {
                stack_mark1.pop_to();
                iter1->release();
                break;
            }
            rel1_count++;
        }

        // second relation...


        corStackMark stack_mark2;
        iter2->reset();

        for (i = 0; i < _arity; i++)
            (iter2->arg_list)[i] = ret_tuple0->arg(i);

          int rel2_count = 0;

        for(;;) {
            ret_tuple1 = iter2->get_next_tuple();
            if (iter2->no_match()) {
                stack_mark2.pop_to();
                iter2->release();
                break;
            }
            rel2_count++;
        }

        int diff = rel2_count - rel1_count;
        if (diff <= 0) continue; // 2nd rel already has correct number

        // else delete diff many copies of ret_tuple0 from 2nd rel...

        corStackMark stack_mark3;
        iter2->reset();

        for (i = 0; i < _arity; i++)
            (iter2->arg_list)[i] = ret_tuple0->arg(i);

        while ( diff-- > 0 ) {
            ret_tuple1 = iter2->get_next_tuple();
            ret_tuple1->do_delete();
        }

        stack_mark3.pop_to();
        iter2->release();
    }

    stack_mark0.pop_to();
    return(env);
}

//**************************************************************************
corBindEnv* corDiffDSolver(corTupleIterator& iterator)
{
    corBindEnv *env = iterator.bindenv;

    if (iterator.ipos > 0) {
        iterator.stack_mark.pop_to();
        iterator.set_no_match(); return(env);
    }
    iterator.stack_mark.get_mark();
    iterator.ipos++;

    if ((iterator.arg_list).count() != 2) {
        fprintf(corexEnv->error_file,
            "Error: difference_d: wrong number of arguments.\n");
        return(NULL);
    }

    corTerm terms[2];
    for (int i = 0; i < 2; i++) {
        terms[i].expr = iterator.arg_list[i];
        terms[i].bindenv = env;
        cordeep_dereference_term(terms[i]);
    }

    corRelation *rels[2];
    for (i = 0; i < 2; i++) {
        switch (terms[i].kindof()) {
        case COR_RELATION:
            rels[i] = terms[i].expr.RelationP();
            break;
        case COR_FUNCTOR:
            rels[i] = find_relation((terms[i].expr.FuncArgP())->_functor,
                (terms[i].expr.FuncArgP())->args.count());
            if (rels[i] == NULL) {
                CORAL_error(COR_REL_NOT_FOUND, NULL, "get_next");
                iterator.stack_mark.pop_to();
                return(NULL);
            }
            break;
        case COR_SYMBOL_CONST:
            rels[i] = find_relation(terms[i].expr.SymbolP(), -1);
            if (rels[i] == NULL) {
                CORAL_error(COR_REL_NOT_FOUND, NULL, "get_next");
                iterator.stack_mark.pop_to();
                return(NULL);
            }
            break;
        default:
            CORAL_error(COR_BAD_ARG_TYPE, NULL, "get_next");
            iterator.stack_mark.pop_to();
            return(NULL);
        }
    }
    corStackMark stack_mark0;

    corArrayBindEnv * env0 = new corArrayBindEnv(rels[0]->arity());
    corArgList * args0 = cormake_vararglist(rels[0]->arity());
    corTupleIterator * iter0 = new corTupleIterator(rels[0], *args0, env0);

    corArrayBindEnv * env1 = new corArrayBindEnv(rels[1]->arity());
    corArgList * args1 = cormake_vararglist(rels[1]->arity());
    corTupleIterator * iter1 = new corTupleIterator(rels[1], *args1, env1);

    if (rels[0]->arity() != rels[1]->arity()) {
        fprintf(corexEnv->error_file,
            "Error: difference_d: Both sets must have equal arity\n");
        iterator.stack_mark.pop_to();
        return(NULL);
    }

    corTuple *ret_tuple0;
    corTuple *ret_tuple1;
    int _arity = rels[0]->arity();

    // scan all of rels[0]
    for(;;) {
        ret_tuple0 = iter0->get_next_tuple();
        if (iter0->no_match())
            break;

        // delete 1 copy of ret_tuple0 from rels[1] if it exists.

        corStackMark stack_mark1;
        iter1->reset();
        for (int i = 0; i < _arity; i++)
            (iter1->arg_list)[i] = ret_tuple0->arg(i);

        ret_tuple1 = iter1->get_next_tuple(); 
        if (!(iter1->no_match())) ret_tuple1->do_delete();
        stack_mark1.pop_to();
        iter1->release();
    }

    stack_mark0.pop_to();
    return(env);
}

//**************************************************************************
corBindEnv* corMakeSetDSolver(corTupleIterator& iterator)
{
    corBindEnv *env = iterator.bindenv;

    if (iterator.ipos > 0) {
        iterator.stack_mark.pop_to();
        iterator.set_no_match(); return(env);
    }
    iterator.stack_mark.get_mark();
    iterator.ipos++;

    if ((iterator.arg_list).count() != 1) {
        fprintf(corexEnv->error_file,
            "Error: make_set_d: wrong number of arguments provided.\n");
        return(NULL);
    }

    corTerm terms;
    terms.expr = iterator.arg_list[0];
    terms.bindenv = env;
    cordeep_dereference_term(terms);

    corRelation *rel;
    switch (terms.kindof()) {
        case COR_RELATION:
            rel = terms.expr.RelationP();
            break;
        case COR_FUNCTOR:
            rel = find_relation((terms.expr.FuncArgP())->_functor,
                 (terms.expr.FuncArgP())->args.count());
            if (rel == NULL) {
                CORAL_error(COR_REL_NOT_FOUND, NULL, "get_next");
                iterator.stack_mark.pop_to();
                return(NULL);
            }
            break;
        case COR_SYMBOL_CONST:
            rel = find_relation(terms.expr.SymbolP(), -1);
            if (rel == NULL) {
                CORAL_error(COR_REL_NOT_FOUND, NULL, "get_next");
                iterator.stack_mark.pop_to();
                return(NULL);
            }
            break;
        default:
            CORAL_error(COR_BAD_ARG_TYPE, NULL, "get_next");
            iterator.stack_mark.pop_to();
            return(NULL);
    }
    corStackMark stack_mark0;

    corArrayBindEnv * env0 = new corArrayBindEnv(rel->arity());
    corArgList * args0 = cormake_vararglist(rel->arity());
    corArrayBindEnv * env1 = new corArrayBindEnv(rel->arity());
    corArgList * args1 = cormake_vararglist(rel->arity());

    corTupleIterator * iter0 = new corTupleIterator(rel, *args0, env0);
    corTupleIterator * iter1 = new corTupleIterator(rel, *args1, env1);

    corTuple *ret_tuple0;
    corTuple *ret_tuple1;

    int _arity = rel->arity();

    for(;;) {
        ret_tuple0 = iter0->get_next_tuple();
        if (iter0->no_match())
            break;

         corStackMark stack_mark1;
         iter1->reset();
         for(int i = 0; i < _arity; i++)
            (iter1->arg_list)[i] = ret_tuple0->arg(i);

         // want to delete all but 1 copy of ret_tuple0.    know there is at least
         // one, so set a new iterator to point at the first one, then delete
         // the rest.

         ret_tuple1 = iter1->get_next_tuple();
         // this is the first one.
     
         // now delete the rest...
         for (;;) {
            ret_tuple1 = iter1->get_next_tuple();
            if (iter1->no_match()) {
                stack_mark1.pop_to();
                iter1->release();
                break;
             }
            // delete the tuple...
             ret_tuple1->do_delete();
        }
    }

    stack_mark0.pop_to();

    return(env);
}

//**************************************************************************
corBindEnv* corCutSolver(corBuiltinRelation& , corTupleIterator& iterator)
{
    if (corCurrModuleInfo->UsePipelining && iterator.calling_pei)
        iterator.calling_pei->cut_crossed = 1;
    return(iterator.bindenv);
}

//**************************************************************************
corBindEnv* corQuitSolver(corBuiltinRelation&, corTupleIterator& iterator)
{
  corexEnv->C_interrupt_raised = 2; // quit().
  corexEnv->quit = 1;

  return(iterator.bindenv);
}

//**************************************************************************

static int _CORAL_counter;

//**************************************************************************
corBindEnv* corResetCounterSolver(corBuiltinRelation& , corTupleIterator& iterator)
{
    if (iterator.arg_list.arity() != 1) {
        fprintf(corexEnv->error_file,
            "CORAL :: Error -- bad arity for reset_counter()");
        iterator.set_no_match(); 
        return(NULL);
    }

    corTerm term1 (iterator.arg_list[0], iterator.bindenv);
    cordeep_dereference_term(term1);
    if (term1.expr.kindof() != COR_NUM_CONST ) {
        fprintf(corexEnv->error_file,
            "CORAL :: Error -- bad argument 1 to reset_counter =");
        term1.printon(corexEnv->error_file);
        fprintf(corexEnv->error_file,"\n");
        iterator.set_no_match();
        return(iterator.bindenv);
    }

    _CORAL_counter = make_int(term1.expr);

    return(iterator.bindenv);
}

//**************************************************************************
corBindEnv* corCounterSolver(corBuiltinRelation& , corTupleIterator& iterator)
{
    if (iterator.arg_list.arity() != 1) {
        fprintf(corexEnv->error_file,"CORAL :: Error -- bad arity for counter()");
        iterator.set_no_match(); 
        return(NULL);
    }

    corTerm term1 (iterator.arg_list[0], iterator.bindenv);
    cordeep_dereference_term(term1);
    if (term1.expr.kindof() != COR_VARIABLE ) {
        fprintf(corexEnv->error_file,
            "CORAL :: Error -- bad argument 1 to counter =");
        term1.printon(corexEnv->error_file);
        fprintf(corexEnv->error_file,"\n");
        iterator.set_no_match();
        return(iterator.bindenv);
    }

    _CORAL_counter++;
    corTerm term2(_CORAL_counter, iterator.bindenv);
    corunify_args(term2, term1);

    return(iterator.bindenv);
}

//**************************************************************************
corBindEnv* corEmptyPrintSolver(corBuiltinRelation& , corTupleIterator& iterator)
{
    return(iterator.bindenv);
}

//**************************************************************************
#define FAIL_RETURN  {iterator.set_no_match(); return(NULL);}

corBindEnv* corRelOptionsSolver(corBuiltinRelation& rel, corTupleIterator& iterator)
{
    if (iterator.arg_list.count() == 0) {
        if (rel.name == corExplainOnSymbol) {
            corexEnv->GlobalRelationOptions |= REL_EXPLAIN;
            if (coralConfig.server_mode && (! corcurUser->explain_sock)){
               
                  // Tell client to start explain
                fprintf(corexEnv->print_file,"%s%d %d %s %hu\n",
                    CP_EXPL, corcurConn->conn_id, corcurUser->user_id,
                    corclients->expl_sock->host, corclients->expl_sock->port);
                  // Client will block until Explain connects
                corcurUser->yield();
            }
        }
        else if (rel.name == corExplainOffSymbol) {
            corexEnv->GlobalRelationOptions &= ~REL_EXPLAIN;
        }
        else if (rel.name == corTraceSymbol) {
            corexEnv->GlobalRelationOptions |= REL_DISPLAY_INSERTIONS;
        }
        else if (rel.name == corUntraceSymbol) {
            corexEnv->GlobalRelationOptions &= ~REL_DISPLAY_INSERTIONS;
        }
        else if (rel.name == corProfileOnSymbol) {
            corexEnv->GlobalRelationOptions |= REL_PROFILE_STATS;
        }
        else if (rel.name == corProfileOffSymbol) {
            corexEnv->GlobalRelationOptions &= ~REL_PROFILE_STATS;
        }

        return(iterator.bindenv);
    }

    FOR_EACH_ARG(arg, iterator.arg_list) {

        corTerm term1 (*arg, iterator.bindenv);
        cordeep_dereference_term(term1);
        if ( term1.expr.kindof() != COR_FUNCTOR) {
            term1.expr.printon(corexEnv->error_file);
            CORAL_error(COR_BAD_ARGLIST, NULL, "RelOptionsSolver");
            FAIL_RETURN;
        }
        
        corFuncArg * fa = term1.expr.FuncArgP();
        if (fa->arity() != 1 || fa->args[0].kindof() != COR_SYMBOL_CONST) {
            term1.expr.printon(corexEnv->error_file);
            CORAL_error(COR_BAD_ARGLIST, NULL, "RelOptionsSolver");
            FAIL_RETURN;
        }

        char *adorn = corSymbolString(fa->args[0]);

        corRelation *newrel = corfind_external_relation(fa->functor(), strlen(adorn));
        if (! newrel) {
            CORAL_error(COR_BAD_REL_NAME, corSymbolString(term1.expr),
                "RelOptionsSolver");
            FAIL_RETURN;
        }
        
        if (rel.name == corTraceSymbol) {
            cororOptionToRelation(newrel, adorn, REL_DISPLAY_INSERTIONS);
        }
        else if (rel.name == corUntraceSymbol) {
            corandOptionToRelation(newrel, adorn, ~REL_DISPLAY_INSERTIONS);
        }
        else if (rel.name == corProfileOnSymbol) {
            cororOptionToRelation(newrel, adorn, REL_PROFILE_STATS);
        }
        else if (rel.name == corProfileOffSymbol) {
            corandOptionToRelation(newrel, adorn, ~REL_PROFILE_STATS);
        }
        else if (rel.name == corExplainOnSymbol) {
            cororOptionToRelation(newrel, adorn, REL_EXPLAIN);
        }
        else if (rel.name == corExplainOffSymbol) {
            corandOptionToRelation(newrel, adorn, ~REL_EXPLAIN);
        }

    } END_EACH_ARG;

    return(iterator.bindenv);
}

#undef FAIL_RETURN

//**************************************************************************
corBindEnv* corDebugSolver(corBuiltinRelation& rel, corTupleIterator& iterator)
{
    if (iterator.arg_list.count() == 0) {
        CORAL_error(COR_EMPTY_ARGLIST, NULL, "DebugSolver");
        return(iterator.bindenv);
    }

    if (rel.name == corSetSymbol) {
      //Set each global variable specified in the argument list

        FOR_EACH_ARG(arg, iterator.arg_list) {
            corTerm term1 (*arg, iterator.bindenv);
            cordeep_dereference_term(term1);
            if (term1.expr.kindof() != COR_SYMBOL_CONST) {
                CORAL_error(COR_BAD_ARG_TYPE, NULL, "DebugSolver");
                term1.printon(corexEnv->error_file);
                fprintf(corexEnv->error_file,"\n");
                iterator.set_no_match();
                return(iterator.bindenv);
            }
            corAssignValueToGlobal(term1.expr, 1);
        } END_EACH_ARG;
    }
    else if (rel.name == corClearSymbol) {
        //Clear each global variable specified in the argument list

        FOR_EACH_ARG(arg, iterator.arg_list) {
            corTerm term1 (*arg, iterator.bindenv);
            cordeep_dereference_term(term1);
            if (term1.expr.kindof() != COR_SYMBOL_CONST){
                CORAL_error(COR_BAD_ARG_TYPE, NULL, "DebugSolver");
                term1.printon(corexEnv->error_file);
                fprintf(corexEnv->error_file,"\n");
                iterator.set_no_match();
                return(iterator.bindenv);
            }
            corAssignValueToGlobal(term1.expr, 0);
        } END_EACH_ARG;
    }
    else if (rel.name == corAssignSymbol) {
        //Assign the value in the second argument to the global variable 
        //specifed in the first argument

        if (iterator.arg_list.count() != 2) {
            CORAL_error(COR_BAD_ARITY, "2 arguments needed","DebugSolver");
            return(iterator.bindenv);
        }

        corTerm term1 (iterator.arg_list[0], iterator.bindenv);
        cordeep_dereference_term(term1);
        if (term1.expr.kindof() != COR_SYMBOL_CONST){
            CORAL_error(COR_BAD_ARG_TYPE, NULL, "DebugSolver");
            term1.printon(corexEnv->error_file);
            fprintf(corexEnv->error_file,"\n");
            iterator.set_no_match();
            return(iterator.bindenv);
        }

        corTerm term2 (iterator.arg_list[1], iterator.bindenv);
        cordeep_dereference_term(term2);
        if (term2.expr.kindof() == COR_NUM_CONST) {
            switch(iterator.arg_list[1].cNum_kindof()) {
            case COR_INTEGER:
                corAssignValueToGlobal(iterator.arg_list[0],
                    iterator.arg_list[1].NumIval());
                break;
            case COR_DOUBLE:
                corAssignValueToGlobal(iterator.arg_list[0],
                    (iterator.arg_list[1].NumDval()));
                break;
            }
        }
        else if (term2.expr.kindof() == COR_SYMBOL_CONST) {
            char *file_name = corSymbolString(term1.expr);
            if (!strcmp(file_name, "trace_file")) {
                char *new_file = corSymbolString(term2.expr);
                if (!set_trace_file(new_file)) {
                    iterator.set_no_match();
                    return(iterator.bindenv);
                }
            }
            else if (!strcmp(file_name, "print_file")) {
                char *new_file = corSymbolString(term2.expr);
                if (!set_print_file(new_file)) {
                    iterator.set_no_match();
                    return(iterator.bindenv);
                }
            }
            else {
                CORAL_error(COR_BAD_GLOBAL_VAR, NULL, "AssignValueToGlobal");
                iterator.set_no_match();
                return(iterator.bindenv);
            }
        }
        else {
            CORAL_error(COR_BAD_ARG_TYPE, NULL, "DebugSolver");
            term2.printon(corexEnv->error_file);
            fprintf(corexEnv->error_file,"\n");
            iterator.set_no_match();
            return(iterator.bindenv);
        }
    }
    else {
        // Should never reach here. Safety check.
        abort();
    }

    return(iterator.bindenv);
}

//**************************************************************************
corBindEnv* corHelpSolver(corBuiltinRelation& , corTupleIterator& iterator)
{
    if (iterator.arg_list.count() == 0) {
        char name_buffer[1000];
        sprintf(name_buffer, "%s%s/general.help", corexEnv->coral_path, HELP_PATH);
        if (corDisplayFile(name_buffer))
            CORAL_error(COR_FILE_NOT_FOUND, "Sorry, no help available", NULL);
    }
    else {
        corTerm term1 (iterator.arg_list[0], iterator.bindenv);
        cordeep_dereference_term (term1);
        if (term1.expr.kindof() != COR_SYMBOL_CONST ) {
            CORAL_error(COR_BAD_ARG_TYPE, NULL, "HelpSolver");
            term1.printon(corexEnv->error_file);
            fprintf(corexEnv->error_file,"\n");
            iterator.set_no_match();
            return(iterator.bindenv);
        }

        char *help_name = corSymbolString(term1.expr);
        char name_buffer[1000];
        sprintf(name_buffer, "%s%s/%s.help", corexEnv->coral_path, HELP_PATH,
            help_name);
        if (corDisplayFile(name_buffer))
            CORAL_error(COR_FILE_NOT_FOUND, "Sorry, no help available", NULL);
    }

    return(iterator.bindenv);
}

//**************************************************************************
corBindEnv* corConsultSolver(corBuiltinRelation& , corTupleIterator& iterator)
{
    char tmp_consult[MAX_PATH_LEN];

#    ifdef DO_TRACE
        if (corexEnv->dbg_get_next) {
            fprintf(corexEnv->trace_file, "Entering %s\n", "consult");
        }
#    endif

    sprintf(tmp_consult, TMP_CONSULT_FILE, coral_pid);

    if (iterator.arg_list.count() != 1) {
        CORAL_error(COR_BAD_ARITY, "1 argument to consult needed",
            "ConsultSolver");
        iterator.set_no_match();
        return(iterator.bindenv);
    }

    corTerm term1 (iterator.arg_list[0], iterator.bindenv);
    cordeep_dereference_term(term1);

    if (term1.expr.kindof() != COR_SYMBOL_CONST ) {
        CORAL_error(COR_BAD_ARG_TYPE, NULL, "ConsultSolver");
        term1.printon(corexEnv->error_file);
        fprintf(corexEnv->error_file,"\n");
        iterator.set_no_match();
        return(iterator.bindenv);
    }

    char *file_name = corSymbolString(term1.expr);

    if (cortableFile(file_name)) {
        correadTable(file_name,iterator);
        return(iterator.bindenv);
    } 
    else if (corincrLoadFile(file_name)) {
        char input[100];

        if (corexEnv->C_interactive_mode_default) {
            fprintf(corexEnv->error_file,
                "\t... sure you want to incrementally load code ? (y/n)[n]");
            fgets(input,100,corexEnv->input_file);
            if ((input[0] != 'y') && (input[0] != 'Y')) {
                iterator.set_no_match();
                return(iterator.bindenv);
            }
        }
#        ifdef INCRLOADER
            loadIncrFile(file_name,iterator);
#        else
            fprintf(corexEnv->error_file,"No incremental loader!\n");
#        endif
        return(iterator.bindenv);
    }
    else {
        char term_str[200];
        sprintf(term_str, "/lib/cpp %s %s > %s", COR_CPP_OPTS,
                     file_name, tmp_consult);
        if (system(term_str)) {
            fprintf(corexEnv->error_file,
                "CORAL::error : Unable to execute the shell command\n");
            fprintf(corexEnv->error_file, "%s\n", term_str);
            fprintf(corexEnv->error_file, "Aborting attempt to consult file\n");
            unlink(tmp_consult);
            iterator.set_no_match();
            return(iterator.bindenv);
        }
        
        
        if (!corconsultFile(tmp_consult, file_name)) {
            CORAL_error(COR_FILE_NOT_FOUND, file_name, "ConsultSolver");
            iterator.set_no_match();
            unlink(tmp_consult);        
            return(iterator.bindenv);
        }
    
        unlink(tmp_consult);
    }
    return(iterator.bindenv);
}

//**************************************************************************
corBindEnv* corTuplePrintSolver(corBuiltinRelation& , corTupleIterator& iterator)
{
    if (!corexEnv->C_verbose_answer_default) return(iterator.bindenv);

    int count = iterator.arg_list.count();

    if (count > 0) {
        fprintf(corexEnv->print_file, " (");
        corArgList *copy_args = new corArgList(count);
        corCopyArgs(copy_args, &iterator.arg_list,
            iterator.bindenv, count, NULL);

        FOR_EACH_ARG(arg, ((corArgList)(*copy_args))) {
            arg->printon(corexEnv->print_file);
            if (--count) fprintf(corexEnv->print_file,", ");
        } END_EACH_ARG;

        delete copy_args;
        fprintf(corexEnv->print_file,")\n");
    }
    else {
        if (corexEnv->C_verbose_answer_default)
            fprintf(corexEnv->print_file, "yes.\n");
    }

    fputs("In TuplePrintSolver, about to fflush exEnv->print_file\n",
        corexEnv->error_file);
    fprintf(corexEnv->error_file, "exEnv->print_file == stdout?: %s\n",
            (corexEnv->print_file == stdout) ? "yes" : "no"  );

    fflush(corexEnv->print_file);

    if (corexEnv->C_interactive_mode_current) user_continue();

    return(iterator.bindenv);
}

//**************************************************************************
corBindEnv* corBindingsPrintSolver(corBuiltinRelation& , corTupleIterator& iterator)
{
    int count = iterator.arg_list.count();

    if (count > 0) {
        corArgList *copy_args = new corArgList(count);
        corCopyArgs(copy_args, &iterator.arg_list,
            iterator.bindenv, count, NULL);

        for (int i = 0; i < count; ) {
            iterator.arg_list[i].printon(corexEnv->print_file);
            fprintf(corexEnv->print_file, "=");
            (*copy_args)[i].printon(corexEnv->print_file);
            /***
            if ((*copy_args)[i].kindof() == COR_NUM_CONST)
            delete (*copy_args)[i];
            ****/

            if (++i < count) fputs(", ",corexEnv->print_file);
         }

        delete copy_args;
        fputs(".\n",corexEnv->print_file);
    }
    else {
        if (corexEnv->C_verbose_answer_default)
            fputs("yes.\n",corexEnv->print_file);
    }

    fflush(corexEnv->print_file);

    if (corexEnv->C_interactive_mode_current) user_continue();

    return(iterator.bindenv);
}

//**************************************************************************
corBindEnv *corPrintSolver(corBuiltinRelation& , corTupleIterator& iterator)
{
    int count = iterator.arg_list.count();

    if (count > 0) {
        corArgList *copy_args = new corArgList(count);
        corCopyArgs(copy_args, &iterator.arg_list,
            iterator.bindenv, count, NULL);

        FOR_EACH_ARG(arg, (*copy_args)) {
            arg->printon(corexEnv->print_file);
            if (--count)
                fputs(" ",corexEnv->print_file);
        } END_EACH_ARG;

        delete copy_args;
     }

    fflush(corexEnv->print_file);

    return(iterator.bindenv);
}

//**************************************************************************
corBindEnv* corPrintfSolver(corBuiltinRelation&, corTupleIterator& iterator)
{
    int argcount = iterator.arg_list.count();

    if (argcount <= 0) {
        CORAL_error(COR_EMPTY_ARGLIST, NULL, "PrintfSolver");
        return(iterator.bindenv);
    }

    corArgList *copy_args = new corArgList(argcount);
    corCopyArgs(copy_args, &iterator.arg_list,
        iterator.bindenv, argcount, NULL);

    corTerm term1(copy_args->arg(0), NULL);
    if (term1.expr.kindof() != COR_SYMBOL_CONST) {
        CORAL_error(COR_BAD_ARG_TYPE, "arg 1 isn't a format string",
            "PrintSolver");
        term1.printon(corexEnv->error_file);
        fputs("\n",corexEnv->error_file);
        iterator.set_no_match();
        return(iterator.bindenv);
    }

    char *format_str = new char[200];
    char *temp1;
    char *temp2 = corSymbolString(term1.expr);
    int count = 1;
    for (;*temp2;) {
        if (*temp2 == '\0') {
  error:    break;
        }
        else if (*temp2 == '%') {
            temp1 = format_str;
            *temp1 = '%';
            temp1++;

            temp2++;
            if (*temp2 == '-') {
                *temp1 = '-';
                temp1++;
                temp2++;
            }
            while ((*temp2 >= '0') && (*temp2 <= '9')) {
                *temp1 = *temp2;
                temp1++;
                temp2++;
            }
            if (*temp2 == '.') {
                *temp1 = '.';
                temp1++;
                temp2++;
            }
            while ((*temp2 >= '0') && (*temp2 <= '9')) {
                *temp1 = *temp2;
                temp1++;
                temp2++;
            }
            if ((*temp2 == 'h') || (*temp2 == 'l')) {
                *temp1 = *temp2;
                temp1++;
                temp2++;
            }
            switch(*temp2) {
                case 'd': case 'i': case 'u': case 's': case 'f':
                case 'e': case 'E': case 'g': case 'G':
                    *temp1 = *temp2;
                    temp1++;
                    temp2++;
                    *temp1 = '\0';
                    if (count >= argcount) {
                        fprintf(corexEnv->error_file,
                            "Not enough arguments to print.\n");
                        goto error;
                    }
                    {
                        corTerm term(copy_args->arg(count), NULL);
                        count++;
                        term.expr.print(NULL, corexEnv->print_file, format_str);
                        temp1 = format_str;
                        break;
                    }
                case '%':
                    fputc('%', corexEnv->print_file);
                    break;
            }
        }
        else {
            fputc( *temp2, corexEnv->print_file);
            temp2++;
        }
    }

    fflush(corexEnv->print_file);

    delete copy_args;
    return(iterator.bindenv);
}


static corCcoralTimer coralTimer;		// 


//**************************************************************************
corBindEnv* corStartTimeSolver(corBuiltinRelation&, corTupleIterator& iterator)
{
    extern int corbeginMem;    // main.C

#ifndef UNIX_HP
    extern struct rusage corSystemUsage;
    getrusage(RUSAGE_SELF, &corSystemUsage);
#endif
    corbeginMem = (int)sbrk(0); // save the top of the execution stack.

#    ifdef DEBUG
        corResetCounts();
#    endif

#ifdef UNIX_HP
    coralTimer.reset();
#endif

    return(iterator.bindenv);
}

#define COR_U_SEC 1000000

//**************************************************************************
#define FAIL_RETURN  {iterator.set_no_match(); return(NULL);}

corBindEnv* corEndTimeSolver(corBuiltinRelation& , corTupleIterator& iterator)
{
    FILE *fd;
    extern struct rusage corSystemUsage;
    extern struct rusage corstartRUsage;

    if (iterator.arg_list.count() == 1) {
        corTerm term1 (iterator.arg_list[0], iterator.bindenv);
        cordeep_dereference_term(term1);
        if (term1.expr.kindof() != COR_SYMBOL_CONST ) {
            fprintf(corexEnv->error_file,
                "CORAL :: Error -- bad argument to display_timer =");
            term1.printon(corexEnv->error_file);
            fprintf(corexEnv->error_file,"\n");
            iterator.set_no_match();
            return(iterator.bindenv);
        }
        char *file_name = corSymbolString(term1.expr);

        if (!strcmp(file_name, "stdout"))
            fd = stdout;
        else if (!strcmp(file_name, "stderr")) 
            fd = stderr;
        else if (!strcmp(file_name, corexEnv->print_filename))
            fd = corexEnv->print_file;
        else if (!strcmp(file_name, corexEnv->trace_filename))
            fd = corexEnv->trace_file;
        else if (!(fd = fopen(file_name, "a"))) {
            CORAL_error(COR_CANNOT_OPEN_FILE, file_name, "EndTimeSolver");
            FAIL_RETURN;
        }

    }
    else fd = corexEnv->error_file;

    long t1, t2;
    double t3, t4;
    struct rusage tmp_usage;
    //int mem_use;
    
#    ifndef UNIX_HP
    getrusage(RUSAGE_SELF, &tmp_usage);
    t1 = tmp_usage.ru_utime.tv_sec - corSystemUsage.ru_utime.tv_sec;
    t2 = tmp_usage.ru_utime.tv_usec - corSystemUsage.ru_utime.tv_usec;
    t3 = ((double) t1) + ((double) t2) / COR_U_SEC;
    fprintf(fd, "User CPU Usage = %7.2f seconds\n", t3);

    t1 = tmp_usage.ru_stime.tv_sec - corSystemUsage.ru_stime.tv_sec;
    t2 = tmp_usage.ru_stime.tv_usec - corSystemUsage.ru_stime.tv_usec;
    t4 = ((double) t1) + ((double) t2) / COR_U_SEC;
    fprintf(fd, "System CPU Usage = %7.2f seconds\n", t4);

//    t1 = tmp_usage.ru_utime.tv_sec - startRUsage.ru_utime.tv_sec;
//    t2 = tmp_usage.ru_utime.tv_usec - startRUsage.ru_utime.tv_usec;
//    t3 = ((double) t1) + ((double) t2) / COR_U_SEC;
// this was to show accumulative. changed to show total. -wei
    fprintf(fd, "Total CPU Usage = %7.2f seconds\n", t3 + t4); 
# else
    coralTimer.look();
    fprintf(fd, "User CPU Usage = %7.2f seconds\n", coralTimer.usr);
    fprintf(fd, "System CPU Usage = %7.2f seconds\n", coralTimer.sys );
    fprintf(fd, "Total CPU Usage = %7.2f seconds\n",coralTimer.usr +
	    coralTimer.sys); 
#endif

// the following stuff are quite buggy. -wei

#    ifdef DEBUG 
        //     this is an incorrect memory metric !!! -- PRAVEEN
        //mem_use = (int)sbrk(0);
        //fprintf(fd,"Memory Usage = %d bytes\n",mem_use - beginMem );
#    endif

#    ifdef DEBUG
        fprintf(fd,"Page Faults = %ld\n",
            (tmp_usage.ru_minflt + tmp_usage.ru_majflt) - 
            (corSystemUsage.ru_minflt + corSystemUsage.ru_majflt));

        fprintf(fd,"Number of Context Switches = %ld\n",
            (tmp_usage.ru_nvcsw + tmp_usage.ru_nivcsw) -
            (corSystemUsage.ru_nvcsw + corSystemUsage.ru_nivcsw));
#    endif


#    ifdef DEBUG 
        //     this is an incorrect memory metric !!! -- PRAVEEN
        //fprintf(fd, "Total Memory Usage = %d bytes\n", mem_use - startMemUsage);
#    endif

#    ifdef DEBUG
        if (!corexEnv->C_quiet_mode_default)
            corDumpCounts(fd);
#    endif

    if ((fd != corexEnv->error_file) && (fd != stdout) 
                && (fd != corexEnv->print_file))
        fclose(fd);
    
    return(iterator.bindenv);
}

#undef FAIL_RETURN

//**************************************************************************
corTuple* corCputimeSolver(corTuple *query)
{
    extern struct rusage corstartRUsage;

    if (query == NULL) {
        fprintf(corexEnv->error_file,"Null argument passed to cputime \n");
        return(NULL); 
    }

    int arity;
    if ((arity = query->arity()) != 1) {
        CORAL_error(COR_BAD_ARITY, "1 argument to cputime needed", NULL);
        return(NULL);
    }

    if (((query->args())[0]).kindof() != COR_VARIABLE) {
        fprintf(corexEnv->error_file,
            "CORAL :: Error -- first argument to cputime should be free");
        (query->args())[0].printon(corexEnv->error_file);
        fprintf(corexEnv->error_file,"\n");
        return(NULL);
    }

    long t1 , t2;
    double t3;
    struct rusage tmp_usage;

#ifndef UNIX_HP
// this time is the accumulative time spent in user mode.
    getrusage(RUSAGE_SELF, &tmp_usage);
    t1 = tmp_usage.ru_utime.tv_sec - corstartRUsage.ru_utime.tv_sec;
    t2 = tmp_usage.ru_utime.tv_usec - corstartRUsage.ru_utime.tv_usec;
    t3 = ((double) t1) + ((double) t2) / COR_U_SEC;
#else
    coralTimer.look();
    t3 = coralTimer.coral_usr;
#endif


    corArgList *alist = new corArgList(arity);
    (*alist)[0] = (t3);

    corTuple *return_tuple = new corTuple(alist);
    return(return_tuple);
}
#undef COR_U_SEC

//**************************************************************************
corBindEnv * corShellSolver(corBuiltinRelation& , corTupleIterator& iterator)
{
    corTerm term1 (iterator.arg_list[0], iterator.bindenv);
    cordeep_dereference_term(term1);
    if (term1.expr.kindof() != COR_SYMBOL_CONST ) {
        fprintf(corexEnv->error_file,"CORAL :: Error -- bad argument to shell =");
        term1.printon(corexEnv->error_file);
        fprintf(corexEnv->error_file,"\n");
        iterator.set_no_match();
        return(iterator.bindenv);
    }

    char *term_str = corSymbolString(term1.expr);
    if (system(term_str)) {
        CORAL_error(COR_FAILED_COMMAND, term_str, "ShellSolver");
        iterator.set_no_match();
    }

    return(iterator.bindenv);
}

//**************************************************************************
corTuple* corExtractSolver(corTuple *query)
{
    if (query == NULL) {
        fprintf(corexEnv->error_file,"Null argument passed to ExtractSolver\n"); 
        return(NULL);
    } 

    int arity = query->arity(); 
    corArgList *alist = new corArgList(arity);
    (*alist)[0] = (query->args())[0];
    (*alist)[1] = (query->args())[1];

    char *arg0 = (char *) make_string((query->args())[0]); 
    int arg1 = (int) make_int((query->args())[1]); 
    char* retvalue;
    if (arg1 >= strlen(arg0)) {
        delete [] arg0;
        delete alist;
        return(NULL);
    }
    retvalue = new char[2];
    retvalue[0] = arg0[arg1];
    retvalue[1] = '\0';
    
    (*alist)[2] = make_arg(retvalue);
    corTuple *return_tuple = new corTuple(alist);
    return(return_tuple);
}

//**************************************************************************
corTuple* corLengthSolver(corTuple *query)
{
    if (query == NULL) {
        fprintf(corexEnv->error_file,"Null argument passed to LengthSolver\n"); 
        return(NULL); 
    } 
    int arity = query->arity(); 
    corArgList *alist = new corArgList(arity);
    (*alist)[0] = (query->args())[0];
    char* arg0;
    arg0 = (char *)make_string((query->args())[0]); 
    int    retvalue;
    retvalue = strlen(arg0);
    (*alist)[1] = (retvalue);
    corTuple *return_tuple = new corTuple(alist);
    return(return_tuple);
}

//**************************************************************************
corTuple* corSubstringSolver(corTuple *query)
{
    if (query == NULL) {
        fprintf(corexEnv->error_file,"Null argument passed to SubstringSolver\n"); 
        return(NULL); 
    } 

    int arity = query->arity(); 
    corArgList *alist = new corArgList(arity);
    (*alist)[0] = (query->args())[0];
    (*alist)[1] = (query->args())[1];

    char* arg0;
    arg0 = (char *) make_string((query->args())[0]);
    char* arg1;
    arg1 = (char *) make_string((query->args())[1]);

    char *tmpstr = strstr(arg0, arg1);
    if (!tmpstr)
        tmpstr = "";

    (*alist)[2] = make_arg(tmpstr);
    corTuple *return_tuple = new corTuple(alist);
    delete [] arg0;
    delete [] arg1;
    return(return_tuple);
}

//**************************************************************************
#define FAIL_RETURN { \
    iterator.set_no_match();      \
    if (tmp_file) fclose(tmp_file);     \
    if (tmp_file_name) unlink(tmp_file_name); \
    return(NULL); \
}

corBindEnv* corpipeOutSolver(corBuiltinRelation& , corTupleIterator& iterator)
{
    //    Arglist should have two args.  The first a command to which the output
    //    is given, the second argument a relation name.
    FILE *tmp_file = NULL;
    char *tmp_file_name = NULL;

#    ifdef DO_TRACE
        if (corexEnv->dbg_get_next) {
            fprintf(corexEnv->trace_file, "Entering %s\n", corPipeOutSymbol->string());
        }
#    endif

    if (iterator.arg_list.arity() != 2) {
        fprintf(corexEnv->error_file, "%s: Number of arguments must be 2.\n",
            corPipeOutSymbol->string());
        FAIL_RETURN;
    }
        
    corTerm term1 (iterator.arg_list[0], iterator.bindenv);
    cordeep_dereference_term(term1);
    if (term1.expr.kindof() != COR_SYMBOL_CONST) {
        fprintf(corexEnv->error_file,"CORAL :: Error -- bad first argument to %s =",
            corPipeOutSymbol->string() );
        term1.printon(corexEnv->error_file);
        fprintf(corexEnv->error_file,"\n");
        FAIL_RETURN;
    }

    corTerm term2 (iterator.arg_list[1], iterator.bindenv);
    cordeep_dereference_term(term2);
    if ( term2.expr.kindof() != COR_FUNCTOR) {
        fprintf(corexEnv->error_file,
            "CORAL :: Error -- bad second argument to %s =",
            corPipeOutSymbol->string() );
        term2.printon(corexEnv->error_file);
        fprintf(corexEnv->error_file,"\n");
        FAIL_RETURN;
    }
    corFuncArg * fa = term2.expr.FuncArgP();

    char name[100];
    tmp_file_name = cormk_tmp_file_name(name);
    tmp_file = coropen_file(tmp_file_name);
    if (!tmp_file)
        FAIL_RETURN;

    corRelation *printrel = find_relation(fa->functor(), fa->arity());
    if (!printrel) {
        fprintf(corexEnv->error_file,
            "CORAL :: Error - %s: relation %s(arity = %d) not found\n",
                corPipeOutSymbol->string(), fa->functor()->string(), fa->arity());
        FAIL_RETURN;
    }

#    ifdef DO_TRACE
        if (corexEnv->dbg_get_next) {
            fprintf(corexEnv->trace_file,
                "%s - printing facts..\n", corPipeOutSymbol->string());
        }
#    endif

    printrel->print_facts(tmp_file);
    fclose(tmp_file);
    tmp_file = NULL;

    char *command = (term1.expr.SymbolP())->string();
    char *term_str = new char [strlen(command)+100];
    sprintf(term_str, "%s < %s", command, tmp_file_name);

#    ifdef DO_TRACE
        if (corexEnv->dbg_get_next) {
            fprintf(corexEnv->trace_file, "%s - going to execute command: %s \n",
                corPipeOutSymbol->string(), term_str);
        }
#    endif

    if (system(term_str)) {
        fprintf(corexEnv->error_file,
            "CORAL :: Error -- Command \"%s\" failed !!\n", term_str);
        delete [] term_str;
        FAIL_RETURN;
    }

    delete [] term_str;

    unlink(tmp_file_name);
    iterator.reset_no_match();

    return(iterator.bindenv);
}

#undef FAIL_RETURN

//**************************************************************************
#define FAIL_RETURN { \
    iterator.set_no_match();      \
    if (tmp_file) fclose(tmp_file);     \
    if (tmp_file_name) unlink(tmp_file_name); \
    return(NULL); \
}

corBindEnv* corwriteTableSolver(corBuiltinRelation&, corTupleIterator& iterator)
{
    // Arglist can have one or two args. The first is the relation name,
    // and the second is the output file. If the second argument is not
    // provided, the output is sent to stdout

    FILE *tmp_file = NULL;
    char *tmp_file_name = NULL;
    int sort=0;

#    ifdef DO_TRACE
        if (corexEnv->dbg_get_next) {
            fprintf(corexEnv->trace_file,
                "Entering %s\n", corWriteTableSymbol->string());
        }
#    endif

    corTerm term1 (iterator.arg_list[0], iterator.bindenv);
    cordeep_dereference_term(term1);
    if ( term1.expr.kindof() != COR_FUNCTOR) {
        fprintf(corexEnv->error_file,
                "CORAL :: Error -- bad first argument to %s =",
                corWriteTableSymbol->string() );
        term1.printon(corexEnv->error_file);
        fputs("\n",corexEnv->error_file);
        FAIL_RETURN;
    }
    corFuncArg * fa = term1.expr.FuncArgP();

    char *outfile = NULL;
    if (iterator.arg_list.arity() > 1) {
        corTerm term2 (iterator.arg_list[1], iterator.bindenv);
        cordeep_dereference_term(term2);
        outfile = (term2.expr.SymbolP())->string();
    }

    char* sortstring = NULL;    // to hold args passed in from command-line.

    if (iterator.arg_list.arity() > 2) {
        // now get the sort string. Should be arg 3.
        corTerm term3 (iterator.arg_list[2], iterator.bindenv);
        cordeep_dereference_term(term3);
        sortstring = (term3.expr.SymbolP())->string();
        sort=1;
    }

    char name[100];
    tmp_file_name = cormk_tmp_file_name(name);
    tmp_file = coropen_file(tmp_file_name);
    if (!tmp_file) {
        FAIL_RETURN;
    }

    corRelation *printrel = find_relation(fa->functor(), fa->arity());
    if (!printrel) {
        fprintf(corexEnv->error_file,
            "CORAL :: Error - %s: relation %s(arity = %d) not found\n",
                corWriteTableSymbol->string(), fa->functor()->string(),
                fa->arity());
        FAIL_RETURN;
    }

    fprintf(tmp_file, "## %s %d %s\n", fa->functor()->string(), fa->arity(),
        outfile);
    printrel->corRelation::print_facts(tmp_file);
    fclose(tmp_file);
    tmp_file = NULL;

    // Now that the facts are in the temporary file, call the awk script
    // on it, to write out the facts to outfile in a tabular form.

    char *term_str = new char [strlen(tmp_file_name)+ 256];
    char *sortoptions = new char[256]; // alloc space for sort options
    strcpy(sortoptions,"");             // init it.

    if (sort || sortstring != NULL) {
        sprintf(sortoptions,"sortoptions=%s",(sortstring == NULL ? "\"\"" :
            sortstring)); 
    }

    // set up the command to run the writetable routine.
    sprintf(term_str, "writetable.CORAL %s %s < %s", (sort ? "sort=1" : ""),
        sortoptions, tmp_file_name);
    if (system(term_str)) {
        fprintf(corexEnv->error_file,
            "CORAL :: Error -- Command \"%s\" failed !!\n", term_str);
        delete [] term_str;
        FAIL_RETURN;
    }

    delete [] term_str;

    unlink(tmp_file_name);
    iterator.reset_no_match();

    return(iterator.bindenv);
}

#undef FAIL_RETURN

//**************************************************************************
#define FAIL_RETURN { \
    iterator.set_no_match();      \
    if (tmp_file) fclose(tmp_file);     \
    if (tmp_file_name) unlink(tmp_file_name); \
    return(NULL); \
}

// this routine takes the name of the file in, processes it
// This has the components of the previous functon readTableSolver.
//
corBindEnv* correadTable(const char* infile,corTupleIterator& iterator)
{
    char name[100];
    FILE *tmp_file = NULL;
    char *tmp_file_name = NULL;

    tmp_file_name = cormk_tmp_file_name(name);
    char *term_str = new char [strlen(tmp_file_name)+ 50];

    sprintf(term_str, "readtable.CORAL < %s > %s", infile, tmp_file_name);
    if (system(term_str)) {
        fprintf(corexEnv->error_file,
            "CORAL :: Error -- Command \"%s\" failed !!\n", term_str) ; 
        delete [] term_str;
        FAIL_RETURN;
    }
    
    if (!corconsultFile(tmp_file_name, tmp_file_name)) {
        CORAL_error(COR_FILE_NOT_FOUND, tmp_file_name, "readTable");
        iterator.set_no_match();
        unlink(tmp_file_name);
        delete [] term_str;
        return(iterator.bindenv);
    }

    delete [] term_str;
    unlink(tmp_file_name);
    return(iterator.bindenv);
}

#undef FAIL_RETURN

//**************************************************************************
#define FAIL_RETURN { \
    iterator.set_no_match();      \
    if (tmp_file) fclose(tmp_file);     \
    if (tmp_file_name) unlink(tmp_file_name); \
    return(NULL); \
}

corBindEnv *corpipeInSolver(corBuiltinRelation&, corTupleIterator& iterator)
{
    //    Arglist should have one arg, which is  a command to be run and its
    //    output used as input to the system

    FILE *tmp_file = NULL;
    char *tmp_file_name = NULL;

#    ifdef DO_TRACE
        if (corexEnv->dbg_get_next) {
            fprintf(corexEnv->trace_file,"Entering %s\n", corPipeInSymbol->string());
        }
#    endif

    if (iterator.arg_list.arity() != 1) {
        fprintf(corexEnv->error_file, "%s: Number of arguments must be 1.\n",
            corPipeInSymbol->string());
        FAIL_RETURN;
    }
        
    corTerm term1 (iterator.arg_list[0], iterator.bindenv);
    cordeep_dereference_term(term1);
    if ( term1.expr.kindof() != COR_SYMBOL_CONST) {
        fprintf(corexEnv->error_file,"CORAL :: Error -- bad first argument to %s =",
            corPipeInSymbol->string());
        term1.printon(corexEnv->error_file);
        fprintf(corexEnv->error_file,"\n");
        FAIL_RETURN;
    }
    char *command = (term1.expr.SymbolP())->string();

    char *term_str = new char [strlen(command)+100];

    char name[100];
    tmp_file_name = cormk_tmp_file_name(name);

    sprintf(term_str, "%s > %s", command, tmp_file_name);

#    ifdef DO_TRACE
        if (corexEnv->dbg_get_next) {
            fprintf(corexEnv->trace_file,
                "%s: Executing command %s\n", corPipeInSymbol->string(), term_str);
        }
#    endif

    if (system(term_str)) {
        fprintf(corexEnv->error_file,
            "CORAL :: Error -- Command \"%s\" failed !!\n", term_str);
        delete [] term_str;
        unlink(tmp_file_name);
        FAIL_RETURN;
    }

#    ifdef DO_TRACE
        if (corexEnv->dbg_get_next) {
            fprintf(corexEnv->trace_file, "%s - redirecting input.\n",
                corPipeInSymbol->string());
        }
#    endif

    if (!corconsultFile(tmp_file_name, tmp_file_name)) {
        CORAL_error(COR_FILE_NOT_FOUND, tmp_file_name, "pipeInSolver");
        iterator.set_no_match();
        unlink(tmp_file_name);
        delete [] term_str;
        return(iterator.bindenv);
    }

    delete [] term_str;
    unlink(tmp_file_name);
    return(iterator.bindenv);
}

#undef FAIL_RETURN

//**************************************************************************
#define FAIL_RETURN {iterator.set_no_match(); return(iterator.bindenv);}

corBindEnv *coraddIndexSolver1(corBuiltinRelation&, corTupleIterator& iterator)
{
    //    Arglist should have one arg - a funcarg with relation name, and 
    //    single arg with the binding pattern.

#    ifdef DO_TRACE
        if (corexEnv->dbg_indexing) {
            fprintf(corexEnv->trace_file, "Entering %s\n",
                corMakeIndexSymbol->string());
        }
#    endif

    if (iterator.arg_list.arity() != 1) {
        fprintf(corexEnv->error_file, "%s: Number of arguments must be 1.\n",
            corMakeIndexSymbol->string());
        FAIL_RETURN;
    }
        
    corTerm term1 (iterator.arg_list[0], iterator.bindenv);
    cordeep_dereference_term(term1);
    if ( term1.expr.kindof() != COR_FUNCTOR) {
        fprintf(corexEnv->error_file,"CORAL :: Error -- bad argument to %s =",
            corMakeIndexSymbol->string() );
        term1.printon(corexEnv->error_file);
        fprintf(corexEnv->error_file,"\n");
        FAIL_RETURN;
    }

    corFuncArg * fa = term1.expr.FuncArgP();
    if (fa->arity() != 1 || fa->args[0].kindof() != COR_SYMBOL_CONST) {
        fprintf(corexEnv->error_file,"CORAL :: Error -- bad argument to %s =",
            corMakeIndexSymbol->string() );
        term1.printon(corexEnv->error_file);
        fprintf(corexEnv->error_file,"\n");
        FAIL_RETURN;
    }
    char *adorn = (fa->args[0].SymbolP())->string();

    corRelation *indrel = find_relation(fa->functor(), strlen(adorn));
    if (!indrel) {
        fprintf(corexEnv->error_file,
            "CORAL :: Error - %s: relation %s(arity %d) not found\n",
            corMakeIndexSymbol->string(), fa->functor()->string(),
            (int) strlen(adorn));
        FAIL_RETURN;
    }

    if (indrel->r_kind != COR_R_ANSWER) {
        fprintf(corexEnv->error_file,
            "CORAL :: Error - %s: relation %s(arity %d) not base\n",
            corMakeIndexSymbol->string(), fa->functor()->string(),
            (int) strlen(adorn));
        FAIL_RETURN;
    }
    if (((corStorageRelation *)(indrel))->add_index(adorn) == COR_I_SUCCESS)
        iterator.reset_no_match();
    else
        iterator.set_no_match();

    return(iterator.bindenv);
}

#undef FAIL_RETURN

//**************************************************************************
#define FAIL_RETURN {iterator.set_no_match(); return(iterator.bindenv);}

corBindEnv* coraddIndexSolver2(corBuiltinRelation&, corTupleIterator& iterator)
{
    //    Arglist should have two args - a funcarg with relation name (with a 
    //    single arg with the binding pattern) and a list of bound vars.

#    ifdef DO_TRACE
        if (corexEnv->dbg_indexing) {
            fprintf(corexEnv->trace_file,
                "Entering %s\n", corMakeIndexSymbol->string());
        }
#    endif

    if (iterator.arg_list.arity() != 2) {
        fprintf(corexEnv->error_file, "%s: Number of arguments must be 1 or 2.\n",
            corMakeIndexSymbol->string());
        FAIL_RETURN;
    }
        
    corTerm term1 (iterator.arg_list[0], iterator.bindenv);
    cordeep_dereference_term(term1);
    if (term1.expr.kindof() != COR_FUNCTOR) {
        fprintf(corexEnv->error_file,"CORAL :: Error -- bad argument to %s =",
            corMakeIndexSymbol->string() );
        term1.printon(corexEnv->error_file);
        fprintf(corexEnv->error_file,"\n");
        FAIL_RETURN;
    }
    corFuncArg * fa = term1.expr.FuncArgP();

    corTerm term2 (iterator.arg_list[1], iterator.bindenv);
    cordeep_dereference_term(term2);
  
    corRelation *indrel = find_relation(fa->functor(), fa->arity() );
    if (!indrel) {
        fprintf(corexEnv->error_file,
            "CORAL :: Error - %s: relation %s(arity %d) not found\n",
            corMakeIndexSymbol->string(), fa->functor()->string(), fa->arity());
        FAIL_RETURN;
    }

    fprintf(corexEnv->error_file, "CORAL:: add_index - pattern form index addition not fully supported yet\n");
    /**********
    if (indrel->add_index(term1.expr->argsterm2.expr) == COR_I_SUCCESS)
        iterator.reset_no_match();
    else iterator.set_no_match();
    **********/

    return(iterator.bindenv);
}

#undef FAIL_RETURN

//**************************************************************************
corBindEnv* ArgTypeSolver(corBuiltinRelation& rel, corTupleIterator& iterator)
{
    if (iterator.arg_list.count() != 1) {
        CORAL_error(COR_BAD_ARITY, "1 argument needed", NULL);
        iterator.set_no_match();
        return(iterator.bindenv);
    }
    
    corTerm term1 (iterator.arg_list[0], iterator.bindenv);
    cordeep_dereference_term (term1);

    int (*type_function)(corCArg);
    
    if (rel.name == corIsConstantSymbol)
        type_function = is_constant;
    else if (rel.name == corIsVarSymbol)
        type_function = is_var;
    else if (rel.name == corIsListSymbol)
        type_function = is_list;
    else if (rel.name == corIsFunctorSymbol)
        type_function = is_functor;
    else if (rel.name == corIsStringSymbol)
        type_function = is_string;
    else if (rel.name == corIsNumSymbol)
        type_function = is_num;
    else if (rel.name == corIsIntSymbol)
        type_function = is_int;
    else if (rel.name == corIsLongSymbol)
        type_function = is_long;
    else if (rel.name == corIsShortSymbol)
        type_function = is_short;
    else if (rel.name == corIsFloatSymbol)
        type_function = is_float;
    else if (rel.name == corIsDoubleSymbol)
        type_function = is_double;
    else {
        iterator.set_no_match();
        return(iterator.bindenv);
    }

    // actually, here a simplify() should be called
    /*
    if ( term1.expr.kindof() == COR_VARIABLE )
        term1 = iterator.bindenv->lookup(((VarArg *)term1.expr)->var);
    */
    
    corTermLink *rv = NULL;
    if ((*type_function)(term1.expr.simplify(iterator.bindenv, rv,
        iterator.bindenv, NULL)))
    {
        iterator.reset_no_match();
    }
    else {
        iterator.set_no_match();
    }

    return(iterator.bindenv);
}

//**************************************************************************
corBindEnv* corHistorySolver(corBuiltinRelation&, corTupleIterator& iterator)
{
    if (iterator.arg_list.count() == 0) {
        (corparseEnv->history)->display(corexEnv->error_file);
        return(iterator.bindenv);
    }

    if (iterator.arg_list.count() != 1) {
        CORAL_error(COR_BAD_ARITY, "1 argument to history needed", NULL);
        iterator.set_no_match();
        return(iterator.bindenv);
    }
    
    corTerm term1 (iterator.arg_list[0], iterator.bindenv);
    cordeep_dereference_term (term1);
    if (term1.expr.kindof() != COR_NUM_CONST ) {
        CORAL_error(COR_BAD_ARG_TYPE, NULL, "HistorySolver");
        term1.printon(corexEnv->error_file);
        fputs("\n",corexEnv->error_file);
        iterator.set_no_match();
        return(iterator.bindenv);
    }

    char *tmp = (corparseEnv->history)->getCommand(make_int(term1.expr));

    if (!tmp) {
        fprintf(corexEnv->error_file, "CORAL::Error: out of history range!\n");
        iterator.set_no_match();
        return(iterator.bindenv);
    }

    corparseStack->handle_new_buff(tmp,"buffer");
    corparseStack->parse(corexEnv->error_file,1,
            "syntax error -- skipping to next period\n");

    // Now modify the history stack
    (corparseEnv->history)->addCommand(tmp, corparseEnv->command_count++);
    corparseEnv->is_history_command = 1;

    return(iterator.bindenv);
}

//**************************************************************************
corBindEnv* corDisplayDefaultsSolver(corBuiltinRelation&, corTupleIterator& iterator)
{
    if (iterator.arg_list.count() != 0) {
        CORAL_error(COR_BAD_ARITY, "no arguments to display_defaults needed",
            NULL);
    }

    char name[100];
    char *tmp_file_name = cormk_tmp_file_name(name);
    FILE *tmp_file = coropen_file(tmp_file_name);
    if (!tmp_file) {
        iterator.set_no_match();
        return(iterator.bindenv);
    }
    
    corexEnv->display(tmp_file);
    fclose(tmp_file);
    corDisplayFile(tmp_file_name);

    unlink(tmp_file_name);
    return(iterator.bindenv);
}

//**************************************************************************
corBindEnv *corResetDefaultsSolver(corBuiltinRelation&, corTupleIterator& iterator)
{
    if (iterator.arg_list.count() != 0) {
        CORAL_error(COR_BAD_ARITY,
            "no arguments to reset_defaults needed",NULL);
    }

    if (coralConfig.server_mode)
        corexEnv->init((FILE *) NULL, (FILE *) NULL, (FILE *) NULL);
    else
        corexEnv->init();
    return(iterator.bindenv);
}

//**************************************************************************
#define FAIL_RETURN {iterator.set_no_match(); return(iterator.bindenv);}

corBindEnv* corInsertFactSolver(corBuiltinRelation&,corTupleIterator& iterator)
{
#    ifdef DO_TRACE
        if (corexEnv->dbg_get_next) {
            fprintf(corexEnv->trace_file,"Entering %s\n", "insert");
        }
#    endif

    if (iterator.arg_list.count() > 0) {
        FOR_EACH_ARG(arg, iterator.arg_list) {
            corTerm term1 (*arg, iterator.bindenv);
            cordeep_dereference_term(term1);
            if ( term1.expr.kindof() != COR_FUNCTOR) {
                CORAL_error(COR_BAD_ARGLIST, "Arguments must be functors",
                    "InsertFactSolver");
                term1.printon(corexEnv->error_file);
                fprintf(corexEnv->error_file,"\n");
                FAIL_RETURN;
            }
            corRelation *rel = corfind_local_or_external_relation(
                (term1.expr.FuncArgP())->functor(),
                (term1.expr.FuncArgP())->arity());
            rel->insert_new((term1.expr.FuncArgP())->args, iterator.bindenv);
            // wX-ATTN: delete term1.expr;
        } END_EACH_ARG
    }
    return(iterator.bindenv);
}

#undef FAIL_RETURN

//**************************************************************************
#define FAIL_RETURN {iterator.set_no_match(); return(iterator.bindenv);}

corBindEnv* corDeleteFactSolver(corBuiltinRelation&,corTupleIterator& iterator)
{
#    ifdef DO_TRACE
        if (corexEnv->dbg_get_next) {
            fprintf(corexEnv->trace_file,"Entering %s\n", "delete");
        }
#    endif

    if (iterator.arg_list.count() > 0) {
        FOR_EACH_ARG(arg, iterator.arg_list) {
            corTerm term1 (*arg, iterator.bindenv);
            cordeep_dereference_term(term1);
            if ( term1.expr.kindof() != COR_FUNCTOR) {
                CORAL_error(COR_BAD_ARGLIST, "Arguments must be functors",
                    "DeleteFactSolver");
                term1.printon(corexEnv->error_file);
                fprintf(corexEnv->error_file,"\n");
                FAIL_RETURN;
            }

            corRelation *rel = corfind_local_or_external_relation(
                (term1.expr.FuncArgP())->functor(),
                (term1.expr.FuncArgP())->arity());

            corTuple *tuple1 = new corTuple((term1.expr.FuncArgP())->args);
            tuple1->env_size = corparserStruct.rule_var_names.length;
            tuple1->bindenv = iterator.bindenv;
            rel->tuple_delete(tuple1);
            delete tuple1;

            // wX-ATTN: delete term1.expr;
        } END_EACH_ARG
    }
    return(iterator.bindenv);
}

#undef FAIL_RETURN

//**************************************************************************
// This function takes in 3 arg. 1st is base, 2nd is exp, 3rd is answer to the
// query. 
//**************************************************************************

corTuple* corPowSolver(corTuple* in)
{
    corArgList* alist = new corArgList(3);
    corCArg  args[3];
    double nums[3],base,exp,ans;
    corCArg Ans;            // place for result.
    char type=0;
    int numargs=0;
    int numvars=0;
    const int ROOT =0, LOG =1, POW =2;
    if(in->arity() != 3) {
        CORAL_error(COR_BAD_ARITY,
            "Format of pow is: pow(Base,Exponent,Answer)","PowSolver");
        return(NULL);
    }

    for (int i=0;i<3;i++) {
        args[i] = (in->args())[i];
        if (is_num(args[i])) {
            numargs++;
            if(is_int(args[i])) {
                if (i == ROOT) 
                    base = (args[i].NumDval());
                else if (i == LOG) 
                    exp = (args[i].NumDval());
                else if (i == POW) 
                    ans = (args[i].NumDval());
            }
            else if (is_double(args[i])) {
                if (i == ROOT) 
                    base = (args[i].NumDval());
                else if (i == LOG) 
                    exp = (args[i].NumDval());
                else if (i == POW) 
                    ans = (args[i].NumDval());
            }
            else {
                CORAL_error(COR_BAD_ARG_TYPE,
                    "Arg must be of type DOUBLE or INT","PowSolver");
                return(NULL);
            }
        } else if (is_var(args[i])) {
            numvars++;
            type=i;
        } else {
            CORAL_error(COR_BAD_ARG_TYPE, "Arg must be a Variable or a Number",
                "PowSolver");
            return(NULL);
        }
    }

    if (numargs == 2 && numvars == 1) {
        switch(type) {
        case POW:
            Ans = (pow(base,exp)); // call the function.
            (*alist)[0] = args[0];
            (*alist)[1] = args[1];
            (*alist)[2] = Ans;
            break;
        case LOG:
            Ans = ((log(ans)/log(base)));
            (*alist)[0] = args[0];
            (*alist)[1] = Ans;
            (*alist)[2] = args[2];
            break;
        case ROOT:
            Ans = (pow(ans,1/exp)); // call the function.
            (*alist)[0] = Ans;
            (*alist)[1] = args[1];
            (*alist)[2] = args[2];
            break;
        default:
            // error
            break;
        }
    } 
    else if (numargs == 3 && numvars == 0) { // still fine
        if (ans == pow(base,exp)) {
            (*alist)[0] = args[0];
            (*alist)[1] = args[1];
            (*alist)[2] = args[2];
        }
    }
    else {
        CORAL_error(COR_BAD_ARG_TYPE,
            "Format of pow is: pow(Base,Exponent,Answer)", "PowSolver");
        return(NULL);
    }
    corTuple* Result = new corTuple(alist); // create the tuple.
    
    return(Result);
}

//**************************************************************************
corTuple *corCreateSetSolver(corTuple *query)
{
    if (query == NULL) {
        fprintf(corexEnv->error_file,"Null argument passed to %s \n",
            corSymbolString(corCreateSetSymbol)); 
        return(NULL); 
    }

    int arity;
    if ((arity = query->arity()) != 1) {
        CORAL_error(COR_BAD_ARITY, "1 argument to create_set needed", NULL);
        return(NULL);
    }

    if (((query->args())[0]).kindof() != COR_VARIABLE) {
        fprintf(corexEnv->error_file,
            "CORAL :: Error -- first argument to %s should be free",
            corSymbolString(corCreateSetSymbol));
        (query->args())[0].printon(corexEnv->error_file);
        fprintf(corexEnv->error_file,"\n");
        return(NULL);
    }
    corArgList *alist = new corArgList(arity);
    (*alist)[0] = corcreate_nested_multiset();

    corTuple *return_tuple = new corTuple(alist);
    return(return_tuple);
}

//**************************************************************************
#define FAIL_RETURN {iterator.set_no_match(); return(iterator.bindenv);}

corBindEnv* corAliasSolver(corBuiltinRelation&, corTupleIterator& iterator)
{
    if (iterator.arg_list.count() != 2) {
        CORAL_error(COR_BAD_ARITY, "2 arguments to alias needed", NULL);
        FAIL_RETURN;
    }

    corTerm term1 (iterator.arg_list[0], iterator.bindenv);
    cordeep_dereference_term(term1);
    if ( term1.expr.kindof() != COR_SYMBOL_CONST) {
      fprintf(corexEnv->error_file,"CORAL :: Error -- bad first argument to %s =",
            corAliasSymbol->string() );
      term1.printon(corexEnv->error_file);
      fprintf(corexEnv->error_file,"\n");
      FAIL_RETURN;
    }

    corTerm term2 (iterator.arg_list[1], iterator.bindenv);
    cordeep_dereference_term(term2);
    if ( term2.expr.kindof() != COR_SYMBOL_CONST) {
        fprintf(corexEnv->error_file,
            "CORAL :: Error -- bad second argument to %s =",
            corAliasSymbol->string() );
        term2.printon(corexEnv->error_file);
        fprintf(corexEnv->error_file,"\n");
        FAIL_RETURN;
    }

    Association *ptr = 
            corSymbolLookup(corBuiltinDB.RelationTable,term2.expr.SymbolP());
    if (HashNone(ptr->arg)) {
        fprintf(corexEnv->error_file, 
            "Specified builtin relation %s does not exist !\n",
            corSymbolString(term2.expr));
        FAIL_RETURN;
    }

    corBuiltinRelation *tmp = (corBuiltinRelation *)(ptr->val);
    corBuiltinRelation *dummy = new corBuiltinRelation(tmp->arity(),
        term1.expr.SymbolP(), tmp->solver);
    
    return(iterator.bindenv);
}

#undef FAIL_RETURN

//**************************************************************************
corTuple *corAddElemDSolver(corTuple *query)
{
    if (query == NULL) {
        fprintf(corexEnv->error_file,"Null argument passed to %s \n",
            corSymbolString(corAddElemSymbol)); 
        return((corTuple *) NULL);
    }

    int arity;
    if ((arity = query->arity()) != 2) {
        CORAL_error(COR_BAD_ARITY, "2 arguments to add_elem_d needed", NULL);
        return((corTuple *) NULL);
    }

    if (!((query->args())[0]).isConstant()) {
        fprintf(corexEnv->error_file,
            "CORAL :: Error -- First argument to %s is free",
            corSymbolString(corAddElemDSymbol));
        (query->args())[0].printon(corexEnv->error_file);
        fprintf(corexEnv->error_file,"\n");
        return((corTuple *) NULL);
    }

    if (((query->args())[1]).kindof() != COR_RELATION) {
        fprintf(corexEnv->error_file,
            "CORAL :: Error -- second argument to %s is not a set: ",
            corSymbolString(corAddElemDSymbol));
        (query->args())[1].printon(corexEnv->error_file);
        fprintf(corexEnv->error_file,"\n");
        return((corTuple *) NULL);
    }

    corArgList *alist = new corArgList(arity);
    (*alist)[0] = (query->args())[0];
    (*alist)[1] = (query->args())[1];

    ((*alist)[1]).RelationP()->
        insert_tuple(make_tuple(make_arglist1((*alist)[0])));

    corTuple *return_tuple = new corTuple(alist);
    return(return_tuple);
}

//**************************************************************************
corTuple* corAddElemSolver(corTuple *query)
{
    if (query == NULL) {
        fprintf(corexEnv->error_file,"Null argument passed to %s \n",
            corSymbolString(corAddElemSymbol)); 
        return((corTuple *) NULL);
    }

    int arity;
    if ((arity = query->arity()) != 3) {
        CORAL_error(COR_BAD_ARITY, "3 arguments to add_elem needed", NULL);
        return((corTuple *) NULL);
    }

    if (!((query->args())[0]).isConstant()) {
        fprintf(corexEnv->error_file,
            "CORAL :: Error -- first argument to %s is free",
            corSymbolString(corAddElemSymbol));
        (query->args())[0].printon(corexEnv->error_file);
        fprintf(corexEnv->error_file,"\n");
        return((corTuple *) NULL);
    }

    if (((query->args())[1]).kindof() != COR_RELATION) {
        fprintf(corexEnv->error_file,
            "CORAL :: Error -- second argument to %s is not a set: ",
            corSymbolString(corAddElemSymbol));
        (query->args())[1].printon(corexEnv->error_file);
        fprintf(corexEnv->error_file,"\n");
        return((corTuple *) NULL);
    }

    corArgList *alist = new corArgList(arity);
    (*alist)[0] = (query->args())[0];
    (*alist)[1] = (query->args())[1];

    ((*alist)[2]) = corcreate_nested_multiset();
    ((*alist)[1]).StorageRelP()->
        add_all_tuples(((*alist)[2]).StorageRelP());
    ((*alist)[2]).RelationP()->
        insert_tuple(make_tuple(make_arglist1((*alist)[0])));

    if (((query->args())[2]).isConstant()) {
        if (((query->args())[2]).kindof() != COR_RELATION) {
            fprintf(corexEnv->error_file,
                "CORAL :: Error -- third argument to %s is not a set: ",
                corSymbolString(corAddElemSymbol));
            (query->args())[2].printon(corexEnv->error_file);
            fprintf(corexEnv->error_file,"\n");
            return((corTuple *) NULL);
        }
        else if (!((query->args())[2]).equals((*alist)[2]))
            return((corTuple *) NULL);
    }

    corTuple *return_tuple = new corTuple(alist);
    return(return_tuple);
}

//**************************************************************************
corTuple* corAbsSolver(corTuple *query)
{
    if (query == NULL) {
        fprintf(corexEnv->error_file,"Null argument passed to %s \n",
            corSymbolString(corAbsOpSymbol)); 
        return(NULL); 
    }

    int arity;
    if ((arity = query->arity()) != 2) {
        CORAL_error(COR_BAD_ARITY, "2 arguments to abs needed", NULL);
        return(NULL);
    }

    if (((query->args())[0]).kindof() != COR_NUM_CONST) {
        fprintf(corexEnv->error_file,
            "CORAL :: Error -- first argument to %s is not a number: ",
            corSymbolString(corAbsOpSymbol));
        (query->args())[0].printon(corexEnv->error_file);
        fprintf(corexEnv->error_file,"\n");
        return(NULL);
    }

    if ((((query->args())[1]).kindof() != COR_NUM_CONST) &&
        (((query->args())[1]).kindof() != COR_VARIABLE))
    {
        fprintf(corexEnv->error_file,
            "CORAL :: Error -- second argument to %s is neither free nor \
            a number: ", corSymbolString(corAbsOpSymbol));
        (query->args())[1].printon(corexEnv->error_file);
        fprintf(corexEnv->error_file,"\n");
        return(NULL);
    }

    corArgList *alist = new corArgList(arity);

    double arg0, arg1;
    int arg1_is_var = 0;

    (*alist)[0] = (query->args())[0];
    if (is_double((*alist)[0]))
        arg0 = ((*alist)[0]).NumDval();
    else if (is_int((*alist)[0]))
        arg0 = make_int((*alist)[0]);
    else
        return(NULL);

    if (is_double((query->args())[1]))
        arg1 = make_double((query->args())[1]);
    else if (is_int((query->args())[1]))
        arg1 = make_int((query->args())[1]);
    else
        arg1_is_var = 1;

    if (arg1_is_var) {
        (*alist)[1] = fabs(arg0);        
    }
    else {
        if (fabs(arg0) == arg1) 
            (*alist)[1] = (query->args())[1];
        else
            return(NULL);
    }

    corTuple *return_tuple = new corTuple(alist);
    return(return_tuple);
}

//**************************************************************************
corTuple* corCeilSolver(corTuple *query)
{
    if (query == NULL) {
        fprintf(corexEnv->error_file,"Null argument passed to %s \n",
            corSymbolString(corCeilOpSymbol)); 
        return(NULL); 
    }

    int arity;
    if ((arity = query->arity()) != 2) {
        CORAL_error(COR_BAD_ARITY, "2 arguments to ceiling needed", NULL);
        return(NULL);
    }

    if (((query->args())[0]).kindof() != COR_NUM_CONST) {
        fprintf(corexEnv->error_file,
            "CORAL :: Error -- first argument to %s is not a number: ",
            corSymbolString(corCeilOpSymbol));
        (query->args())[0].printon(corexEnv->error_file);
        fprintf(corexEnv->error_file,"\n");
        return(NULL);
    }

    if ((((query->args())[1]).kindof() != COR_NUM_CONST) &&
        (((query->args())[1]).kindof() != COR_VARIABLE))
    {
        fprintf(corexEnv->error_file,
            "CORAL :: Error -- second argument to %s is neither free nor \
            a number: ", corSymbolString(corCeilOpSymbol));
        (query->args())[1].printon(corexEnv->error_file);
        fprintf(corexEnv->error_file,"\n");
        return(NULL);
    }

    corArgList *alist = new corArgList(arity);

    double arg0, arg1;
    int arg1_is_var = 0;

    (*alist)[0] = (query->args())[0];
    if (is_double((*alist)[0]))
        arg0 = make_double((*alist)[0]);
    else if (is_int((*alist)[0]))
        arg0 = make_int((*alist)[0]);
    else
        return(NULL);

    if (is_double((query->args())[1]))
        arg1 = make_double((query->args())[1]);
    else if (is_int((query->args())[1]))
        arg1 = make_int((query->args())[1]);
    else
        arg1_is_var = 1;

    if (arg1_is_var) {
        (*alist)[1] = (ceil(arg0));        
    }
    else {
        if (ceil(arg0) == arg1) 
            (*alist)[1] = (query->args())[1];
        else
            return(NULL);
    }

    corTuple *return_tuple = new corTuple(alist);
    return(return_tuple);
}

//**************************************************************************
corTuple* corFloorSolver(corTuple *query)
{
    if (query == NULL) {
        fprintf(corexEnv->error_file,"Null argument passed to %s \n",
            corSymbolString(corFloorOpSymbol)); 
        return(NULL); 
    }

    int arity;
    if ((arity = query->arity()) != 2) {
        CORAL_error(COR_BAD_ARITY, "2 arguments to floor needed", NULL);
        return(NULL);
    }

    if (((query->args())[0]).kindof() != COR_NUM_CONST) {
        fprintf(corexEnv->error_file,
            "CORAL :: Error -- first argument to %s is not a number: ",
            corSymbolString(corFloorOpSymbol));
        (query->args())[0].printon(corexEnv->error_file);
        fprintf(corexEnv->error_file,"\n");
        return(NULL);
    }

    if ((((query->args())[1]).kindof() != COR_NUM_CONST) &&
        (((query->args())[1]).kindof() != COR_VARIABLE))
    {
        fprintf(corexEnv->error_file,
            "CORAL :: Error -- second argument to %s is neither free nor \
            a number: ", corSymbolString(corFloorOpSymbol));
        (query->args())[1].printon(corexEnv->error_file);
        fprintf(corexEnv->error_file,"\n");
        return(NULL);
    }

    corArgList *alist = new corArgList(arity);

    double arg0, arg1;
    int arg1_is_var = 0;

    (*alist)[0] = (query->args())[0];
    if (is_double((*alist)[0]))
        arg0 = make_double((*alist)[0]);
    else if (is_int((*alist)[0]))
        arg0 = make_int((*alist)[0]);
    else
        return(NULL);

    if (is_double((query->args())[1]))
        arg1 = make_double((query->args())[1]);
    else if (is_int((query->args())[1]))
        arg1 = make_int((query->args())[1]);
    else
        arg1_is_var = 1;

    if (arg1_is_var) {
        (*alist)[1] = floor(arg0);        
    }
    else {
        if (floor(arg0) == arg1) 
            (*alist)[1] = (query->args())[1];
        else
            return(NULL);
    }

    corTuple *return_tuple = new corTuple(alist);
    return(return_tuple);
}

//**************************************************************************
static void user_continue()
{
  char input[100];

  if (corsuspension.set >0) { //previous suspension command.
    if (corsuspension.set >1 ) 
      corsuspension.set--;// generate the number of answers required.
    return;
  }

  if (corsuspension.set ==0)    // suspension according to page
    return;

  if (coralConfig.server_mode){//never gets here!
    //fprintf(exEnv->print_file, "Next Answer ? (y/n/all/9)[y]\n");
    //parseEnv->print_prompt();
    fputs("SEVERE WARNING: INFORM CORAL PROGRAMMER!!!\n",corexEnv->error_file);
    fprintf(corexEnv->error_file,"user_continue: server, suspension.set = %d\n",
            corsuspension.set);
    return;
  } else
    fprintf(corexEnv->print_file, "\t... next answer ? (y/n/all/9)[y]");

  fgets(input,100,corexEnv->input_file);

  // Hack to get rid of excess \n
  if (input[0] == '\n')
      fgets(input,100,corexEnv->input_file);
  
  if ((input[0] == 'n') || (input[0] == 'N')) {
    // prevent any further computation
    corexEnv->C_interrupt_raised = -1;
  }
  else if ((input[0] == 'a') || (input[0] == 'A')) {
    // prevent further answer prompting
    corexEnv->C_interactive_mode_current = 0;
  } else if (isdigit(input[0]))
    corsuspension.set=1+atoi(input); //suspended!
}

//**************************************************************************
