// **************************************************************************
//    Interface.h
// **************************************************************************

#ifndef _Interface_H_
#define _Interface_H_

#include "Arg.h"
#include "ArgList.h"
#include "BuiltinSymbols.h"
#include "BuiltinRelation.h"
//#include "HashSimpleRelation.h"
#include "TupleIterator.h"
#include "CArg.h"

#include <stdarg.h>

// **************************************************************************

class corParserStruct;
class corHashSimpleRelation;

class corRelation;
typedef corRelation* (*UserRelProc)(corTuple  *query);

// **************************************************************************
/*
    These structures are used in the C++ code created by the
    CORAL/C++ translator
*/

enum parameter_kind {
  COR_NULL,
  COR_INT,
  COR_SHORT,
  COR_LONG,
  COR_FLOAT,
  COR_DOUB,
  COR_STR
};

// represents an arg and the corresponding parameter.
class corParameterStruct {
 public:
  parameter_kind  kind;
  void *value_ptr;
  corParameterStruct() { kind = COR_NULL; value_ptr = NULL;}
};

class corExecutionList {
 public:
  corParserStruct **exec_terms;
  corParameterStruct *parameters;
  int no_of_params;

  int max_no_of_terms;
  int actual_no_of_terms;
  int step_size;
  int initialized;

  void add(corParserStruct *pstruct) ;

  void initialize(int i, int num_params, corParameterStruct *params) {
    exec_terms = new corParserStruct*[i];
    parameters = params;
    no_of_params = num_params;
    max_no_of_terms = i;
    step_size = i;
    actual_no_of_terms = 0;
    initialized = 1;
  }
  int execute();
  int execute(int index, int& param_index);
};

// **************************************************************************
/*
 *    For embedded coral programs:
 *
 *    call init_coral() at the beginning of main.
 *    call exit_coral() at the end of main.
 */
// **************************************************************************
/* 
 * These structures are used to provide C++ users with the facility
 * of scans over CORAL relations
 */
class C_ScanDesc {
  corTupleIterator *it;
  corArrayBindEnv *env;
  corArgList *arg_list;
  corStackMark *stackmark;
  int tuple_fetched;

  public :

  C_ScanDesc(corRelation *rel);
  C_ScanDesc(corRelation *rel, corTuple *match_tuple);
  ~C_ScanDesc();
  corTuple *next_tuple();
  int no_match() { return it->no_match();}
};

// **************************************************************************
/* These extern function declarations are used by the translated C++
 * program, with the embedded CORAL code removed
 */

extern int  corconsultFile(char *, char *);
extern int  corconsultFile(char *, char *, int);
extern int  corconsultBuff(char *, char *);
extern void corprocessInputStack();
extern void corcopy_structs(corParserStruct*, corParserStruct&);


// **************************************************************************
/* The following functions are defined in Interface.C */

/* -------------------------------------------------------------------- */
/* Interface functions for creating arguments of tuples:        */

extern corArg *make_arg  (char *name);
extern corArg *make_arg  (Name name);

extern corArg *make_var  (int var_num);
extern corArg *make_var  (char *print_name, int var_num);
extern corArg *make_var  (Name print_name, int var_num);

extern corArg *make_arg  (char *func_name, corArgList *args);
extern corArg *make_cons (corArg *a, corArg *corb);
extern corArg *make_cons (corCArg a, corCArg corb);
extern corArg *make_nil  ();
extern corArg *make_list (int num_arg, ...);  // Not an arg_list, a real list
extern int  corlist_length( corArg *list );

extern corArgList *make_arglist  (int len, ...);
extern corArgList *make_arglist1 (corArg *a1);
extern corArgList *make_arglist1 (corCArg a1);
extern corArgList *make_arglist2 (corArg *a1, corArg *a2);
extern corArgList *make_arglist2 (corCArg a1, corCArg a2);
extern corArgList *make_arglist3 (corArg *a1, corArg *a2, corArg *a3);
extern corArgList *make_arglist3 (corCArg a1, corCArg a2, corCArg a3);
extern corArgList *make_arglist4 (corArg *a1, corArg *a2, corArg *a3, corArg *a4);
extern corArgList *concat_arglists (corArgList *al1, corArgList *al2);
extern corArgList *cormake_vararglist (int n);

/* -------------------------------------------------------------------- */
/* Interface functions for testing and getting C++ data structures from 
        CORAL data:                         */

/** int Tuple::arity ();   and 
    Arg *Tuple::arg (int arg_pos);   **/

/** int ArgList::arity ();  and 
    Arg *ArgList::arg (int arg_pos);  **/


extern int is_int      (corArg *a);
extern int is_long     (corArg *a);
extern int is_short    (corArg *a);
extern int is_float    (corArg *a);
extern int is_double   (corArg *a);
inline int is_int (corCArg c) { return c.is_int();};
inline int is_long (corCArg c) { return c.is_long();};
inline int is_short (corCArg c) { return c.is_short();};
inline int is_float (corCArg c) { return c.is_float();};
inline int is_double (corCArg c) { return c.is_double();};

extern int is_string   (corArg *a);
extern int is_num      (corArg *a);
extern int is_constant (corArg *a);

extern int is_string   (corCArg c);
extern int is_num      (corCArg c);
extern int is_constant (corCArg c);

extern int is_var      (corCArg c);
extern int is_functor  (corCArg c);
extern int is_list     (corCArg c);
extern int is_nil      (corCArg c);

extern int    make_int     (corCArg c);
extern short  make_short   (corCArg c);
extern long   make_long    (corCArg c);
extern float  make_float   (corCArg c);
extern double make_double  (corCArg c);
extern char  *make_string  (corCArg c);

extern char *functor_name    (corArg *a);
extern char *functor_name    (corCArg c);
extern corArgList *functor_args (corArg *a);
extern corArgList *functor_args (corCArg c);
extern corArg *make_car (corArg *a);
extern corArg *make_car (corCArg c);
extern corArg *make_cdr (corArg *a);
extern corArg *make_cdr (corCArg c);

// **************************************************************************
//    Interface functions for creating relations:

typedef corHashSimpleRelation IndexedRelation;

// **************************************************************************
//    Interface functions for finding a database relation:

extern corRelation *find_relation (char *db_rel_name, int arity);

// **************************************************************************
//    Interface functions for calling the declarative module:

extern int call_coral(char *exp_pred_name, corTuple *query, corRelation *&result);
    //
    // The query is a tuple for the relation, and any
    //    facts in the relation that unify with the query tuple are 
    //    answers to the query.
    // For example, (X,Y) is a query that returns all tuples in a binary
    //    relation.
    // Similarly, (1,Y) returns all tuples that have 1 in the first 
    //    column.
    // Adds answer tuples to the relation result
    //     If the result parameter is NULL, a new relation is
    //     allocated and result points to it.
    // Returns:  Number of answer tuples.  Result < 0 implies error.

extern int call_coral(char *exp_pred_name, corRelation *queries,
              corRelation *&result);
    //
    // Relation *queries must contain queries on the relation.
    // Adds answer tuples to the relation result.
        // If the result parameter is NULL, a new relation is
        // allocated and result points to it.
    // Returns:  Number of answer tuples.  Result < 0 implies error.

// **************************************************************************
//    Interface functions for stepping through and modifying a relation:

/** FOR_EACH_TUPLE(Tuple *tuple, Relation *rel) **/
#define FOR_EACH_TUPLE(tuple,rel) {\
    corStackMark CONCAT_ID2(tuple,_stack_mark);\
    corArrayBindEnv *CONCAT_ID2(tuple,_env) = \
        new corArrayBindEnv((rel)->arity()); \
    corArgList *alist0 = cormake_vararglist((rel)->arity()); \
    corTupleIterator CONCAT_ID2(ITER,tuple) (rel, *alist0, \
        CONCAT_ID2(tuple,_env)); \
    for(;;) {    \
        * ((corTuple **) &tuple) = CONCAT_ID2(ITER,tuple).get_next_tuple(); \
        if (CONCAT_ID2(ITER,tuple).no_match()) break;    

#define END_EACH_TUPLE(tuple)  } \
    CONCAT_ID2(tuple,_stack_mark).pop_to();\
    delete CONCAT_ID2(tuple,_env); }

/** FOR_EACH_MATCHING_TUPLE(Tuple *tuple, Relation *rel, Tuple *match_tuple) **/
#define FOR_EACH_MATCHING_TUPLE(tuple,rel,match_tuple) {\
    corStackMark CONCAT_ID2(tuple,_stack_mark);\
    corArrayBindEnv *CONCAT_ID2(tuple,_env) = \
        new corArrayBindEnv((rel)->arity()); \
    corTupleIterator CONCAT_ID2(ITER,tuple) \
        (rel, (match_tuple)->args(), CONCAT_ID2(tuple,_env)); \
    for(;;) {       \
        * ((corTuple **) & tuple) = CONCAT_ID2(ITER,tuple).get_next_tuple(); \
        if (CONCAT_ID2(ITER,tuple).no_match()) break;       

extern corTuple *make_tuple(int arity);
extern corTuple *make_tuple(corArgList *arglist);

extern void update_tuple (corTuple *tuple, int arg_pos, corArg *new_val);
extern void update_tuple (corRelation *rel, corTuple *old_tuple, corTuple *new_tuple);
extern void delete_id_tuple (corRelation *rel, corTuple *tuple);
extern void delete_tuple (corRelation *rel, corTuple *tuple);

extern corRelation *make_rel(char *rel_name, int arity);

/** void Relation::insert_tuple (Tuple *tuple) */



// **************************************************************************
// Interface functions for printing

/** Arg::printon(File *file);  and 
    ArgList::printon(File *file);

    Tuple::printon(File *file);
    Relation::printon(File *file);

**/

// **************************************************************************
//    Interface functions for creating indices for relations:

// **************************************************************************
//    Interface functions for calling the imperative module:

/** The user needs to write a routine with signature
    Relation *pred_solver (Tuple *query);    or
    Tuple *pred_solver (Tuple *query);
    for the predicate `pred', and add a declaration matching the form:
    int add_builtin (char *pred, int arity, UserRelProc pred_solver);  or
    int add_builtin (char *pred, int arity, UserTupleProc pred_solver);
    
    where `arity' is the arity of the relation named by `pred'.
**/

extern int add_builtin (char *pred, int arity, UserRelProc pred_solver);
extern int add_builtin (char *pred, int arity, UserTupleProc pred_solver);
#define BUILTIN_FAIL_RETURN {__CORAL_BUILTIN_FAIL=1;return 0;}
#define VOID_BUILTIN_FAIL_RETURN {__CORAL_BUILTIN_FAIL=1;return;}

// **************************************************************************

#endif _Interface_H_

