//    ParseUtils.C


#include <stdio.h>
#include <string.h>

#include "ParseUtils.h"
#include "ParserStruct.h"

#include "DontCareArg.h"
#include "GroupingArg.h"
#include "ArgList.h"
#include "Arg.h"
#include "Globals.h"
#include "HashSimpleRelation.h"
#include "BuiltinSymbols.h"
#include "Array.h"
#include "Literal.h"
#include "ExportInfo.h"
#include "Rule.h"
#include "Util.h"


//    Declared in Gram.y 

extern corArg* corTheParameterArg;

static int C__BoundArgs12 = 1;

static int check_builtin_op(Name name);
static int is_arith_pred(Name pred_name);
static int is_convertible_pred(Name pred_name);
static void copy_arrays(struct corarray * a, struct corarray * corb, int no_arith);
static void check_for_arg_bindings(corLiteral *pred, corArg**arg1, corArg**arg2);
static void insert_predicate(struct corarray *a, corLiteral *pred);

static void simple_reorder_arith_preds(corParserStruct& corparserStruct);

// Start a new argument list.

void corpush_arg_chain(corParserStruct& corparserStruct)
{
    Arg_Chain *chain = new Arg_Chain;
    chain->pushed = corparserStruct.cur_arg_chain;
    corparserStruct.cur_arg_chain = chain;
    chain->count = 0;
    chain->last = NULL;
}


// Add an argument to the current argument list.
void corappend_arg(corParserStruct& corparserStruct, corCArg* a)
{
    Arg_Link *link = new Arg_Link;
    link->prev = corparserStruct.cur_arg_chain->last;
    link->arg = *a;
    corparserStruct.cur_arg_chain->last = link;
    corparserStruct.cur_arg_chain->count++;
}


void corappend_arg(corParserStruct& corparserStruct, corArg *a)
{
    Arg_Link *link = new Arg_Link;
    link->prev = corparserStruct.cur_arg_chain->last;
    link->arg = a;
    corparserStruct.cur_arg_chain->last = link;
    corparserStruct.cur_arg_chain->count++;
}


// Add an argument to the current argument list.
// add a 3rd argument ; wX
void corappend_parameter(corParserStruct& corparserStruct, corCArg *a,
              Parameter_Link*& listStart)
{
    Parameter_Link *link = new Parameter_Link;
    // CHNG::wX
    //link->prev = parserStruct.parameters->last;
    link->prev = listStart;
    link->parameter_slot = a;
    // CHNG::wX
    //parserStruct.parameters->last = link;
    listStart = link;
    corparserStruct.parameters->count++;
}

// Finish off and pop the currently-worked-on argument list.

void corpop_arg_chain(corParserStruct& corparserStruct, corArgList& args)
{

    // CHNG::wX
    Parameter_Link *listEnd;
    if (listEnd = corparserStruct.parameters->last)
      while (listEnd->prev) listEnd = listEnd->prev;

    register Arg_Chain *cur_chain = corparserStruct.cur_arg_chain;

    if (!cur_chain) return;

    int count = cur_chain->count;
    args.set_count(count);
    for (int a = count-1; a >= 0; a--) {
        Arg_Link *link = cur_chain->last;
        cur_chain->last = link->prev;
        args[a] = link->arg;
        if (link->arg == corTheParameterArg) {
            corappend_parameter(corparserStruct, &args[a],
                (listEnd?listEnd->prev:corparserStruct.parameters->last));
            args[a] = &corTheDontCareArg;
        }
        delete link;
    }
    corparserStruct.cur_arg_chain = cur_chain->pushed;
    delete cur_chain;
}

//    Push the currently parsed set on a stack, and allocate a new slot
//    for a new nested set.

void corpush_cur_set(corParserStruct& corparserStruct, ArgPtr set)
{
    corset_stack *link = new corset_stack;
    link->info = set;
    link->prev = corparserStruct.cur_set;
    corparserStruct.cur_set = link;
}

//    Pop the set parse stack, and de-allocate the popped node.

void corpop_cur_set(corParserStruct& corparserStruct)
{
    ASSERT(corparserStruct.cur_set != NULL); // if NULL, it is a bug;
    corset_stack *link = corparserStruct.cur_set;
    corparserStruct.cur_set = link->prev;
    delete link;
}

//    Allocate a new relation with subsumption turned off to store a nested
//    multiset.

ArgPtr corcreate_nested_multiset()
{
//    Relation *rel = AllocRelation(NestedSymbol,nested_set_arity);

    corStorageRelation *rel = new corHashSimpleRelation(nested_set_arity, 0,
        COR_DEFAULT_SET_BUCKETS);

    rel->set_subsumption(0);
    rel->name = corNestedSymbol;
    return( (ArgPtr)rel);
}


//    Insert an element into a nested set.

void corinsert_cur_set(corParserStruct& corparserStruct, corCArg* element)
{
    corRelation *rel = (corRelation *)corparserStruct.cur_set->info; 
    corArgList *copy_args = new corArgList(nested_set_arity); 

    (*copy_args)[nested_set_arity-1] = *element;
    corTuple *newtuple = new corTuple(copy_args);
    rel->insert_tuple(newtuple);
}

// **************************************************************************
// kept for some legacy const arguments in Gram.y

corCArg* corcreate_list(corParserStruct& corparserStruct, ArgPtr cdr)
{
    register Arg_Chain *cur_chain = corparserStruct.cur_arg_chain;

    while (cur_chain->last != NULL) {
        Arg_Link *link = cur_chain->last;
        corFuncArg *cons = corFuncArg::New(corConsSymbol, 2);
        cons->args[0] = link->arg;
        cons->args[1] = cdr;
        cur_chain->last = link->prev;
        delete link;
        cdr = cons;
    }
    corparserStruct.cur_arg_chain = cur_chain->pushed;
    delete cur_chain;
    return(new corCArg(cdr));
}

corCArg* corcreate_list(corParserStruct& corparserStruct, corCArg* cdr)
{
    register Arg_Chain *cur_chain = corparserStruct.cur_arg_chain;
    corFuncArg* p = NULL;

    while (cur_chain->last != NULL) {
        Arg_Link *link = cur_chain->last;
        corFuncArg *cons = corFuncArg::New(corConsSymbol, 2);
        cons->args[0] = link->arg;
        if (!p)
	  cons->args[1] = *cdr;
	else
	  cons->args[1] = p;
        cur_chain->last = link->prev;
        delete link;
        p = cons;
    }
    corparserStruct.cur_arg_chain = cur_chain->pushed;
    delete cur_chain;
    if (p) {
      delete cdr;
      return(new corCArg(p));
    } else 
      return(cdr);
}

// *************************************************************************

int corrule_var_enter(corParserStruct& corparserStruct, Name s)
{
    register int i;

    //    NOTE: this is a kludge! Unification with dont care terms is tricky,
    //    so instead we force the DontCareSymbol to be treated each time as a
    //    distinct variable.  The parser NEVER generates TheDontCareArg !!
    if (s && s->equals(corDontCareSymbol)) {
        coradd_to_array(&corparserStruct.rule_var_names, (int)s);
        return corparserStruct.rule_var_names.length - 1;
    }

    for (i=0; i < corparserStruct.rule_var_names.length; ++i) {
        if (s == (Name)corparserStruct.rule_var_names.data[i])
            return (i);
    }
    coradd_to_array(&corparserStruct.rule_var_names, (int)s);
    return corparserStruct.rule_var_names.length - 1;
}


int corrule_var_lookup(corParserStruct& corparserStruct, corCArg *c)
{
    register int i;
    if (!c) 
      return (0);
    Name s = c->SymbolP();

    for (i=0; i < corparserStruct.rule_var_names.length; ++i) {
        if ( s == (Name)corparserStruct.rule_var_names.data[i])
            return (i);
    }
    fprintf(corexEnv->error_file, "error line %d: %s not in rule\n",
        corparseEnv->cur_line_num, s->string());
    return(0);
}


int corsip_var_lookup(corParserStruct& corparserStruct, corCArg *c)
{
    register int i;
    if (!c)
      return (0);
    Name s = c->SymbolP();

    for (i=0; i < corparserStruct.sip_var_names.length; ++i) {
        if (s == (Name)corparserStruct.sip_var_names.data[i])
            return (i);
    }
    fprintf(corexEnv->error_file, "error line %d: %s not in plist\n",
        corparseEnv->cur_line_num, s->string());
    return(0);
}


// *************************************************************************
//    Makes array b a copy of array a.  If no_arith is set, arithmetic
//    predicates are not copied

static void copy_arrays(struct corarray * a, struct corarray * corb, int no_arith) 
{
    if (a->data_size == 0) {
        a->data_size = corb->data_size;
        a->data = new int[a->data_size];
    }
    else {
        int *tmp_data = a->data;
        a->data_size = corb->data_size;
        //a->data = (int*)realloc(a->data, a->data_size * sizeof(int));
        a->data = new int[a->data_size];
        if (no_arith){
            for (int i=0; i < a->length; i++)
                a->data[i] = tmp_data[i];
        }
        delete [] tmp_data;
    }

    // Copy data from b to a
    if (!no_arith) {
        memcpy(a->data, corb->data, corb->length*sizeof(int));
        a->length = corb->length;
    }
    else {
        a->length = 0;
        for (int i = 0; i < corb->length; i++) {
            if ( ! is_arith_pred(((corLiteral *)(corb->data[i]))->name()))
                a->data[a->length++] = corb->data[i];
        }
    }
}

// *************************************************************************
//    This routine checks to see if the arguments of the predicate are
//    the same as either arg1 or arg2. If so, the corresponding Arg *
//    is set to Arg *NULL to show that the argument has been bound

static void check_for_arg_bindings(corLiteral *pred, corCArg*arg1, corCArg*arg2)
{
    corVarLink *list = NULL ;

    // Get the list of variables bound in the predicate
    FOR_EACH_ARG(pred_arg, pred->args) {
        list = pred_arg->var_list(list);
    } END_EACH_ARG

    while (list != NULL) {
        if (!(*arg1).IsEmpty() && ((*arg1).VarArgP())->equals(list->arg))
             arg1->init();

        if (!(*arg2).IsEmpty() && ((*arg2).VarArgP())->equals(list->arg))
             (*arg2).init();

        list = list->next;
    }
}

// *************************************************************************
//    This routine inserts the pred at position pos in the array.

void corinsert_array_element(struct corarray *a, int pos, corLiteral * pred)
{
    // Defensive programming :-)
    if (pos > a->length) return;

    for (int i = a->length ; i > pos ; i--)
        a->data[i] = a->data[i-1] ;
    a->data[pos] = (int)pred ;
    (a->length)++ ;
}

// *************************************************************************
//    This routine inserts an arithmetic predicate in a position where its
//    first two arguments are bound.

static void insert_predicate(struct corarray *a, corLiteral *pred)
{
    corCArg arg1, arg2 ;
    corArgList temp ;
    int i ;

    //    Get the two unbound arguments of the predicate
    //    NOTE: default C__BoundArgs12 is true.
    if (C__BoundArgs12) {
        temp = pred->args;
        arg1 = temp[0];
        arg2 = temp[1];
    }
    if ((arg1.IsEmpty() ==0) && (arg1.kindof() != COR_VARIABLE)) arg1.init() ;
    if ((arg2.IsEmpty() ==0) && (arg2.kindof() != COR_VARIABLE)) arg2.init() ;

    for (i = 0; i < a->length ; i++) {
        // If both args are bound, get out of the loop
        if (arg1.IsEmpty() && arg2.IsEmpty()) break;
        check_for_arg_bindings((corLiteral *)(a->data[i]), &arg1, &arg2);
    }

    corinsert_array_element(a, i, pred);
}

// *************************************************************************
static void simple_reorder_arith_preds(corParserStruct& corparserStruct)
{
    struct corarray tempArray;
    int    curPred1 = 0;     // Current predicate in parserStruct.rhspreds

    if (corparseEnv->C_num_head_arith_lits==0) 
        return;        // No simple reordering to be done 

    tempArray.length = tempArray.data_size = 0;
    tempArray.data = NULL;

    for (curPred1 = corparseEnv->C_num_head_arith_lits; 
            curPred1 < corparserStruct.rhspreds.length; curPred1++)
    {
        coradd_to_array(&tempArray,
            (int)(corparserStruct.rhspreds.data[curPred1]));
    }

    for (curPred1 = 0; curPred1 < corparseEnv->C_num_head_arith_lits;
                            curPred1++) {
        coradd_to_array(&tempArray,
            (int)(corparserStruct.rhspreds.data[curPred1]));
    }

    //  NOTE:   The memory allocated for one of the arrays must be freed
    //          This is not done as of now!!

    corparserStruct.rhspreds = tempArray ;

        // To keep destructor of tempArray form destorying array 
    tempArray.data = NULL;
}

// *************************************************************************
// this stupid function returns Grouping* too! 'nest 
corArg* corcreate_functor(corParserStruct& corparserStruct, corCArg *c)
{
 Name op;
 if (!c)
   op = NULL;
 else op = c->SymbolP();

 int builtin_index = check_builtin_op(op);

    // Check if something like: sum<ARG>.
    if (builtin_index >= 0 && builtin_index < (int)NoAggregation &&
        corparserStruct.cur_arg_chain->count==1)
    {
        corCArg arg = corparserStruct.cur_arg_chain->last->arg;
        if (arg.kindof() == COR_GROUPING) {
            corGrouping *nested = arg.GroupingP();
            corGrouping *nest = new corGrouping(nested->nestedArgs, nested->index,
                     (enum AggregateKind)builtin_index);

           // if (nest->kind == SetAggregation) nest->check_subsum = 1;
            delete nested;
            delete corparserStruct.cur_arg_chain->last;
            Arg_Chain *old_cur_chain = corparserStruct.cur_arg_chain;
            corparserStruct.cur_arg_chain = corparserStruct.cur_arg_chain->pushed;
            delete old_cur_chain;
            return nest;
        }
    }

    if (corparserStruct.CurModule.ConvertFunctions && is_convertible_pred(op) ) {
        /* Rewrite: P(x op y) => op(x, y, TMP$i), P(TMP$i) */
        char tmp_buf[20]; corLiteral *pred;
        sprintf(tmp_buf, "TMP_%d_", corparserStruct.rhspreds.length);
        Name tmp_sym = corEnterSymbol(tmp_buf);;
        corVarArg *tmp_var = new corVarArg(tmp_sym,
            corrule_var_enter(corparserStruct, tmp_sym));
        pred = new corLiteral(corparserStruct, op,
            corparserStruct.cur_arg_chain->count + 1);
        corpop_arg_chain(corparserStruct, pred->args);
        pred->args.set_count(pred->args.count() + 1);
        pred->args[pred->args.count() - 1] = tmp_var;
        coradd_to_array(&corparserStruct.rhspreds, (int)pred);
        return tmp_var;
    }
    corFuncArg *arg = corFuncArg::New(op, corparserStruct.cur_arg_chain->count);
    corpop_arg_chain(corparserStruct, arg->args);
    return arg;
}

// *************************************************************************
corVarArg* corpull_out_arith_expr(corParserStruct& corparserStruct)
{
    char tmp_buf[20]; 
    corLiteral *pred;

    sprintf(tmp_buf, "TMP_%d_", corparserStruct.rhspreds.length);
    Name tmp_sym = corEnterSymbol(tmp_buf);
    corVarArg *tmp_var = new corVarArg(tmp_sym,
         corrule_var_enter(corparserStruct, tmp_sym));
    pred = new corLiteral(corparserStruct, corEquOpSymbol, 2);
    corpop_arg_chain(corparserStruct, pred->args);
    pred->args.set_count(2);
    pred->args[1] = tmp_var;
    coradd_to_array(&corparserStruct.rhspreds, (int)pred);
    return tmp_var;
}
 
// *************************************************************************
ArgPtr corcreate_grouping(corParserStruct& corparserStruct)
{
    if (corexEnv->dbg_aggregates) {
      fprintf(corexEnv->error_file,"in create relation count %d\n",
          corparserStruct.cur_arg_chain->count);
    }
    corGrouping *arg = corGrouping::New(corparserStruct.cur_arg_chain->count);
    arg->set_subsumption( (short)(1));
    arg->index = corparserStruct.nested_index++;
    corpop_arg_chain(corparserStruct, arg->nestedArgs->args);
    return arg;
}

// *************************************************************************
ArgPtr corcreate_multi_grouping(corParserStruct& corparserStruct)
{
    if (corexEnv->dbg_aggregates) {
        fprintf(corexEnv->error_file,"in create relation count %d\n",
            corparserStruct.cur_arg_chain->count);
    }
    corGrouping *arg = corGrouping::New(corparserStruct.cur_arg_chain->count);
    arg->set_subsumption( (short)(0));
    arg->index = corparserStruct.nested_index++;
    corpop_arg_chain(corparserStruct, arg->nestedArgs->args);
    return arg;
}
void corAddExportInfo(corParserStruct& corparserStruct, corCArg* carg)
{
 if (carg)
   corAddExportInfo(corparserStruct, carg->FuncArgP());
 else  {
   fprintf(corexEnv->error_file,"export must be followed by a functor.");
   return;
  }

}

// *************************************************************************
void corAddExportInfo(corParserStruct& corparserStruct, corFuncArg* arg)
{

    if (arg->kindof() != COR_FUNCTOR) {
        fprintf(corexEnv->error_file,"export must be followed by a functor.");
        return;
    }

    for (int i = 0; ; ) { // Do at least once
        corExportInfo *new_info = new corExportInfo;
        new_info->next = corparserStruct.CurModule.ExportedPreds;
        new_info->pred_name = arg->functor();
        new_info->adorn_symbol = NULL; /* Arity == 0. */
        new_info->sequence = corparserStruct.CurModule.ExportedCount++;
        if (arg->args.count() == 0)
            ;
        else if (arg->args[i].kindof() != COR_SYMBOL_CONST) {
            fprintf(corexEnv->error_file, "Bad kind of adornment for export %s.\n",
                arg->functor()->string());
        }
        else {
            new_info->adorn_symbol = arg->args[i].SymbolP();
            if (corConvertAdornment(new_info->adornment,
                     new_info->adorn_symbol->string(),
                     (int) new_info->adorn_symbol->length()))
            {
                fprintf(corexEnv->error_file,
                    "Bad character in adornment string '%s'\n",
                    new_info->adorn_symbol->string());
            }
        }
        new_info->predicate = NULL;
        corparserStruct.CurModule.ExportedPreds = new_info;
        if (++i >= arg->args.count())
            break;
    }
				// arg may be a leak point!
}


// *************************************************************************
void COR_set_preprocessing(corModuleInfo& Module, int val)
{
  Module.PreProcessing = val;
}

// *************************************************************************
void COR_set_no_magic(corModuleInfo& Module)
{
    Module.UseSupplementaryMagic = 0;
    Module.UseMagic = 0;
    Module.SupMagicIndexing = 0;
}

// *************************************************************************
void COR_set_use_factormagic(corModuleInfo& Module, int val)
{
    Module.UseFactorMagic = val;
}

// *************************************************************************
void COR_set_use_factoring(corModuleInfo& Module, int val)
{
    Module.UseFactoring = val;
}

// *************************************************************************
void COR_set_supplementary_magic(corModuleInfo& Module, int val)
{
    Module.UseSupplementaryMagic = val;
    if (val) Module.UseMagic = 0;
}

// *************************************************************************
void COR_set_ordinary_magic(corModuleInfo& Module, int val)
{
    Module.UseMagic = val;
    if (val) Module.UseSupplementaryMagic = 0;
}


// *************************************************************************
void COR_set_exist_opt(corModuleInfo& Module, int val)
{
    Module.UseExistOpt = val;
}

// *************************************************************************
void COR_set_rewriting(corModuleInfo& Module, int val)
{
    Module.Rewriting = val;
/*
    if (!val) {
        COR_set_no_magic(Module);
        COR_set_use_factoring(Module, 0);
        COR_set_use_factormagic(Module, 0);
        COR_set_exist_opt(Module, 0);
    }
*/
}


// *************************************************************************
void COR_set_sup_magic_indexing(corModuleInfo& Module, int val)
{
    Module.SupMagicIndexing = val;
    if (val) COR_set_supplementary_magic(Module, val);
}

// *************************************************************************
void COR_set_save_module(corModuleInfo& Module, int val)
{
    Module.SaveModule = val;
}

// *************************************************************************
void COR_set_use_ordsearch(corModuleInfo& Module, int val)
{
    Module.UseOrdSearch = val;
}

// *************************************************************************
void COR_set_eager_eval(corModuleInfo& Module, int val)
{
    Module.EagerEval = val;
}

// *************************************************************************
void COR_set_monotonic(corModuleInfo& Module, int val)
{
    Module.Monotonic = val;
}


// *************************************************************************
void COR_set_pipelining(corModuleInfo& Module, int val)
{
    Module.UsePipelining = val;

    if (val) {
        COR_set_rewriting(Module, 0);
    }
}

// *************************************************************************
void COR_set_predicate_seminaive(corModuleInfo& Module, int val)
{
    Module.UsePredicateSN = val ;
}

// *************************************************************************
void COR_set_check_subsum(corModuleInfo& Module, int val)
{
    Module.CheckSubsum = val;
}

// *************************************************************************
void COR_set_return_unify(corModuleInfo& Module, int val)
{
    Module.UseReturnUnify = val;
}

// *************************************************************************
void COR_set_multisets(corModuleInfo& Module, int val)
{
    Module.MultiSet = val;
    if (val) Module.CheckSubsum = 0;
}

// *************************************************************************
void COR_set_non_ground_facts(corModuleInfo& Module, int val)
{
    Module.NonGroundFacts = val;
}

// *************************************************************************
void COR_set_disk_resident(corModuleInfo& Module, int val)
{
    Module.DiskRel = val;
}

// *************************************************************************
void COR_set_index_deltas(corModuleInfo& Module, int val)
{
    Module.IndexDeltas = val;
}

// *************************************************************************
void COR_set_single_scc(corModuleInfo& Module, int val)
{
    Module.SingleScc = val;
}

// *************************************************************************
void COR_set_single_answer(corModuleInfo& Module, int val)
{
    Module.SingleAnswer = val;
}

// *************************************************************************
void COR_set_convert_functions(corModuleInfo& Module, int val)
{
    Module.ConvertFunctions = val;
}

// *************************************************************************
void COR_set_nested_scc_eval(corModuleInfo& Module, int val)
{
    Module.NestedSccEval = val;
}


// *************************************************************************
//    Called from the parser : takes the information collected in the 
//    parserStruct and fills the cur_rule structure with it.

void corcomplete_rule(corClause *cur_rule, corParserStruct& corparserStruct)
{
    cur_rule->num_var_names =  corparserStruct.rule_var_names.length;
    if (corparserStruct.rule_var_names.length == 0) {
        cur_rule->var_names = NULL;
    }
    else {
        // Keep the following as malloc so that we can realloc as required.
        cur_rule->var_names = (Name*)malloc(corparserStruct.rule_var_names.length
            * sizeof(Name));
        for (int i=0; i< corparserStruct.rule_var_names.length; ++i) {
            cur_rule->var_names[i] = (Name)corparserStruct.rule_var_names.data[i];
        }
    }
    coralloc_preds(cur_rule, corparserStruct.rhspreds.length);
    
    simple_reorder_arith_preds(corparserStruct);
    
    //    Copy literals from parserStruct.rhspreds to new cur_rule
    for (int i= 0; i < corparserStruct.rhspreds.length; ++i)
        cur_rule->preds[i] = (corLiteral*)corparserStruct.rhspreds.data[i];
    
    coradd_rule(cur_rule, corparserStruct.rules);
    
    corAddClauseToPredicate(cur_rule->head->predicate,cur_rule);
    corparserStruct.rule_list = (corClause*)corparserStruct.rules->chain;
}

// *************************************************************************
static int check_builtin_op(Name name)
{
    for (int i = 0; corbuiltin_ops[i] != 0; i++)
        if ((*corbuiltin_ops[i]) == name) return i;
    return -1;
}

// *************************************************************************
//    This routine checks if the predicate is an arithmetic predicate.
//    Builtin_Arith_Start depends on the builtin_ops array defined in Gram.y.

static int is_arith_pred(Name pred_name) 
{
    if (check_builtin_op(pred_name) >= corBuiltin_Arith_Start) return(1) ;
    return(0) ;
}

// *************************************************************************
static int is_convertible_pred(Name pred_name) 
{
    if (check_builtin_op(pred_name) >= corBuiltin_Convertible_Start) 
        return(1) ;
    return(0) ;
}

// *************************************************************************

