/*****************************************************************************/
//    DerivedMethod.C
/*****************************************************************************/

#include "DerivedMethod.h"

#include "Config.h"
#include "DerivedRelation.h"
#include "SymbolArg.h"
#include "Database.h"
#include "Unparse.h"
#include "Predicate.h"
#include "PipelinedDerivedRelation.h"
#include "QFModuleData.h"

/*****************************************************************************/
corDerivedMethod::corDerivedMethod(Name name, corBitVector& adorn,
    InterpDerivedSolver solvr, corQFModuleData *module)
    : adornment(adorn)
{
    corDerivedRelation *rel;

    valid_method = 1;
    md = module;
    if (corCurDB->RelationTable == NULL) {
        corCurDB->RelationTable = corAllocTab(COR_RELATION_TABLE_INCR, NULL);
    }

    char buf[1000];

    strcpy(buf, corSymbolString(name));
    sprint_adornment(buf+strlen(buf), &adorn);
    Name check_name = corEnterSymbol(buf);
    
    if (coris_magic(check_name) || coris_supp(check_name) || coris_mp_done(check_name))
    {
        fprintf(corexEnv->error_file,
            "Warning::relation name %s conflicts with system reserved names !\n"
            ,buf);
        fprintf(corexEnv->error_file, "New definition not retained !\n");
        return;
    }
    else {
        sprintf(buf, "%s", corSymbolString(name));
    }
    
    Name new_name = corEnterSymbol(buf);

    Association *ptr = corSymbolLookup(corCurDB->RelationTable, new_name);
    if (HashNone(ptr->arg)) {
        ptr->arg = new_name;
        if (module->kindof() == COR_PIPELINED) {
            ptr->val = new corPipelinedDerivedRelation((corQFPipelinedModuleData *)md,
                new_name, adorn.len());
            md = NULL;
            //    the md field is not used by pipelined relations. instead the
            //    ModuleData is stored along with the relation
        }
        else {
            ptr->val = new corDerivedRelation(new_name, adorn.len());
        }
        ((corRelation *)(ptr->val))->single_success =
            module->module_info.SingleAnswer;
        corTabInserted(corCurDB->RelationTable, ptr);
    }
    else {
        corRelation *rel1 = (corRelation *)(ptr->val);

        // search methods already defined to see if there is a conflict
        if (rel1->r_kind != COR_R_DERIVED) {
            fprintf(corexEnv->error_file, "Base relation %s already exists !\n",
                corSymbolString(new_name));
            fprintf(corexEnv->error_file, "Newer definition is NOT retained !\n");
            return;
        }

        rel = (corDerivedRelation*)(ptr->val);
        corDerivedMethod *meth = rel->methods;

        /*
            Three cases: 
            1)  The new definition is a pipelined relation.  In this case, we
                have to first delete the original relation, and insert the
                new one
            2)    The new definition is a Derived relation, and the old
                definition is a pipelined relation.  In this case too, we have
                to first delete the original relation, and insert the new one.
            3)    Both definitions are Derived relations.  Only the ModuleData
                need be replaced
        */
        if (!meth->md || (module->kindof() == COR_PIPELINED)) {
            fprintf(corexEnv->error_file, "Duplicate definition of relation %s\n",
                corSymbolString(new_name));
            fprintf(corexEnv->error_file, "Retaining newer definition\n");

            corDeleteInTab(corCurDB->RelationTable, rel->name);
            // delete the old relation, and all its methods
            delete rel;

            ptr->arg = new_name;
            if (module->kindof() == COR_PIPELINED) {
                ptr->val = new corPipelinedDerivedRelation(
                    (corQFPipelinedModuleData*)md, new_name, adorn.len());
                md = NULL;
                //    This field is not used by pipelined relations.  Instead
                //    the ModuleData is stored along with the relation 
            }
            else {
                ptr->val = new corDerivedRelation(new_name, adorn.len());
            }
            ((corRelation *)(ptr->val))->single_success =
                module->module_info.SingleAnswer;
            corTabInserted(corCurDB->RelationTable, ptr);
            rel = (corDerivedRelation*)(ptr->val);
            next = rel->methods;
            rel->methods = this;
            interp_solver = solvr;
            return;
        }

        while (meth) {
            if (meth->valid_method && (meth->adornment==adorn)) {
                fprintf(corexEnv->error_file,
                    "Duplicate definition with identical adornment\n");
                fprintf(corexEnv->error_file, "Retaining newer definition of %s",
                    corSymbolString(new_name));
                corprint_adornment(adorn, corexEnv->error_file);
                fprintf(corexEnv->error_file, "\n");

                meth->interp_solver = solvr;
                if (meth->md) delete meth->md;
                meth->md = md;
                return;
            }
            meth = meth->next;
        }
    }

    rel = (corDerivedRelation*)(ptr->val);
    next = rel->methods;
    rel->methods = this;
    interp_solver = solvr;
}

/*****************************************************************************/
corDerivedMethod::~corDerivedMethod()
{
    if (md) delete md;
}

/*****************************************************************************/

