/*****************************************************************************/
//    DerivedRelation.C
/*****************************************************************************/

#include <stdlib.h>
#include <string.h>

#include "DerivedRelation.h"

#include "TupleIterator.h"
#include "Globals.h"
#include "Unparse.h"
#include "QFMaterializedModuleData.h"
#include "Execute.h"
#include "CoralError.h"
#include "GenericIndex.h"
#include "HashSimpleRelation.h"
#include "ParseActions.h"
#include "ParseUtils.h"
#include "ParserEnv.h"
#include "ParserStruct.h"
#include "CoralBreak.h"

#include <fcntl.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/param.h>

/*****************************************************************************/

extern corArgList *corMagicArgs;
extern corBindEnv *corMagicBindEnv;
extern corStorageRelation *corMagicRelation;

/*****************************************************************************/
corDerivedRelation::corDerivedRelation(Name _name, int new_arity)
{
    name = _name;
    _arity = new_arity;
    methods = NULL;
    has_release = 1;
    r_kind = COR_R_DERIVED;
}

/*****************************************************************************/
corDerivedRelation::~corDerivedRelation()
{
    corDerivedMethod *meth = methods, *next_meth;
    while (meth) {
        next_meth = meth->next;
        delete meth;
        meth = next_meth;
    }
}

/*****************************************************************************/
corBindEnv* corDerivedRelation::get_next(corTupleIterator& iterator)
{
    this->get_next_tuple(iterator);
    if (iterator.no_match()) return NULL;
    return(iterator.bindenv);
}

/*****************************************************************************/
corTuple* corDerivedRelation::get_next_tuple(corTupleIterator& iterator)
{
    char fifo_path[MAXPATHLEN];

#    ifdef DO_PROFILE
        if (corT_Stack.count > 0)
            (corT_Stack.cur_md()->module_info.corGlobalProfileInfo.getnexts)++;
        else
            (corGlobalProfileInfo.getnexts)++ ;
#    endif
    
    int i;
    corStorageRelation* tuples;
    corTupleIterator* tuples_iterator;

    if (iterator.ipos++ == 0) {
        corBindEnv *env = iterator.bindenv;
        iterator.tuple_env = NULL;
    
        int arity = iterator.arg_list.count();
        // Find adornment of query.
        corBitVector adornment(arity);

        for (i = 0; i < arity; i++) {
            // new stuff
            corTerm term1 (iterator.arg_list[i], iterator.bindenv);
            FULL_DEREFERENCE_TERM(term1);
            corTermLink *rv = NULL;
            if ((term1.expr.simplify(iterator.bindenv, rv, iterator.bindenv,
                NULL)).isConstant())
            {
                adornment.set(i, 1);
            }
            else {
                adornment.set(i, 0);
            }
        }

        //    Find a method that matches the query's adornment.

        corDerivedMethod *method;

        // Best method with insufficent 'b' (where query is 'f').
        corDerivedMethod *best_underbound_method = NULL;

        // Count of 'b's short in best_underbound_method.
        int best_underbound_shortfall = arity + 1;

        // Best method with excess 'b' (where query is 'f').
        corDerivedMethod *best_overbound_method = NULL;

        int least_overbound_excess = arity + 1;
        
        for (method = methods; ; method = method->next) {
            int adorn_delta[3];
            if (method == NULL) {
                if ((method = best_underbound_method) != NULL) {
                    if (best_underbound_shortfall &&
                        !corexEnv->C_quiet_mode_default)
                    {
                        fprintf(corexEnv->error_file,
                            "CORAL::Warning : Using underbound method ");
                        (method->adornment).printon(corexEnv->error_file);
                        fprintf(corexEnv->error_file, " for %s ",
                            corSymbolString(name));
                        adornment.printon(corexEnv->error_file);
                        fprintf(corexEnv->error_file, "!\n");        
                    }
                    break;
                }
                if ((method = best_overbound_method) != NULL) {
                    if (least_overbound_excess && !corexEnv->C_quiet_mode_default) {
                        fprintf(corexEnv->error_file,
                            "CORAL::Warning : Using overbound method ");
                        (method->adornment).printon(corexEnv->error_file);
                        fprintf(corexEnv->error_file, " for %s ",
                            corSymbolString(name));
                        adornment.printon(corexEnv->error_file);
                        fprintf(corexEnv->error_file, "!\n");        
                        if (method->md->module_info.UseOrdSearch) {
                            fprintf(corexEnv->error_file,"CORAL:: Cannot use Ordered Search with overbound method!");
                            fprintf(corexEnv->error_file, " Sorry !\n");
                            iterator.set_no_match();
                            return((corTuple *) NULL);
                        }
                    }
                    break;
                }
                fprintf(corexEnv->error_file,
                    "No rule matches the query %s (with adornment ",
                    corSymbolString(name));
                corprint_adornment(adornment, corexEnv->error_file);
                fprintf(corexEnv->error_file, ")!\n");
                iterator.set_no_match();
                return((corTuple *) NULL);
            }
            if (!method->valid_method) continue;

            int i = adornment.len();
            if (i != method->adornment.len())
                continue;

            // Use adorn_delta to summarizes the difference between
            // adornment (from the query) and method->adornment.
            // adorn_delta[0] counts bits where adornment is 'f' (0)
            // and method.adornment is 'b' (1).
            // adorn_delta[1] counts bits where the adornments are equal.
            // adorn_delta[2] counts bits where adornment is 'b'
            //   and method.adornment is 'f'.
            adorn_delta[0] = 0;
            adorn_delta[1] = 0;
            adorn_delta[2] = 0;
            while (--i >= 0) {
                adorn_delta[1+adornment.test(i)-method->adornment.test(i)]++;
            }
            if (adorn_delta[0] + adorn_delta[2] == 0) // Exact match.
                break;
            if (adorn_delta[0] == 0 &&
                best_underbound_shortfall > adorn_delta[2])
            {
                // adornment.implies(method->adornment)
                best_underbound_method = method;
                best_underbound_shortfall = adorn_delta[2];
            }
            else if (least_overbound_excess > adorn_delta[0]) {
                // adornment.implies(method->adornment)
                best_overbound_method = method;
                least_overbound_excess = adorn_delta[0];
            }
        }

        cormy_str = corSymbolString(method->md->export_pred->name);
        coropt = (method->md->rel_options | corexEnv->GlobalRelationOptions) &
            REL_EXPLAIN;

        
#        ifdef DO_EXPLAIN
        if ((method->md->rel_options | corexEnv->GlobalRelationOptions) &
                REL_EXPLAIN)
        {
            int childpid;

            if ((corexEnv->explain_fp == NULL) && (! coralConfig.server_mode)){

                sprintf(fifo_path,"/tmp/explain_fifo");
                    // Remove the old FIFO if nessecary
                unlink(fifo_path);

                if ( mkfifo(fifo_path, S_IRUSR | S_IWUSR ) < 0)
                    fprintf(corexEnv->error_file,
                            "Can't create FIFO for explain.\n");


                if ((childpid = fork()) < 0) {
                    //    fork another process to run explain
                     fprintf(corexEnv->error_file,"can't fork\n");
                }
                else if (childpid > 0) {

                    //    parent process
                    cormy_str =corSymbolString(method->md->export_pred->name);

                    if (!(corexEnv->explain_fp = fopen(fifo_path,"w"))) {
                        CORAL_error(COR_CANNOT_OPEN_FILE,
                            "file_name" , "explain");
                        return((corTuple *) NULL);
                    }
                    setbuf(corexEnv->explain_fp,NULL);
                    fputs("coral.\n",corexEnv->explain_fp);
                    fprintf(corexEnv->explain_fp,
                        "module_name(\"%s\",X).\n",cormy_str);
                    fflush(corexEnv->explain_fp);

                } else {
                    int fd;
                    char *path = new char[MAXPATHLEN];
                    char *path2 = new char[MAXPATHLEN];
                    if ((fd = open(fifo_path, O_RDONLY)) < 0)
                        fprintf(corexEnv->error_file,
                            "Cannot open %s\n", fifo_path); 

                    dup2(fd,0); //duplicate file descriptor
                    sprintf(path,"%s/bin/coralsh",corexEnv->coral_path);
                    sprintf(path2,"%s/bin/Explain",corexEnv->coral_path);
                    execl(path, "coralsh", "-f", path2, (char*) 0);

                        /* NOTREACHED */
                    delete path;
                    delete path2;
                    exit(0);
                }

            } else {

                if (corexEnv->explain_fp == NULL){
                    fputs("Explain not connected - turning explain off.\n",
                            corexEnv->error_file);
                    corexEnv->GlobalRelationOptions &= ~REL_EXPLAIN;
                } else {

                      // Already have a FIFO (Explain is running) 

                    cormy_str = corSymbolString(method->md->export_pred->name);
                    fprintf(corexEnv->explain_fp,"module_name(\"%s\",X).\n",
                            cormy_str);
                    fflush(corexEnv->explain_fp);
                }
            }
        }

#        endif    // DO_EXPLAIN

        iterator.moduledata = method->md;

        corStackMark stack_mark;
        
        // Insert the magic tuple.
        int magic_arity = method->adornment.count();
        corArgList* magic_args = new corArgList(magic_arity);
        int arg_bit = -1;
        for (i = 0; i < magic_arity; i++) {
            arg_bit = method->adornment.next(arg_bit);
            (*magic_args)[i] = iterator.arg_list[arg_bit];
        }
        
        corStorageRelation *magic_rel;
        int old_magic_rel = 0;
        
        {
            //    WARNING: Dangerous in case of an overflow!!
            char magic_name[200];
            corsprint_magic_name(magic_name, corSymbolString(name),
                &(method->adornment));
            magic_rel = cormake_local_relation(corEnterSymbol(magic_name),
                magic_arity, method->md->module_info.IndexDeltas);
            magic_rel->local_options = method->md->rel_options ;
        }
        

        if (((corQFMaterializedModuleData *)(method->md))->use_psn) {
            corMagicArgs = magic_args;
            corMagicBindEnv = iterator.bindenv;
        }
        else
            magic_rel->insert_new(*magic_args, iterator.bindenv);
        
        // Solve the query.
        tuples = (*method->interp_solver)(method->md, magic_rel, old_magic_rel,
            &iterator);
        
        stack_mark.pop_to();
        
        corRMark *endmark = (corRMark *) NULL;
        if (iterator.moduledata->module_info.SaveModule) {
            ASSERT(tuples->headMark.prev != & (tuples->headMark));
            endmark = tuples->headMark.prev;

            // This mark was grabbed by clear_rule_pointers after it 
            // completed evaluation of the query
            // with SaveModules, the relation should only be scanned
            // until endmark, since more facts could be added to it
            // by subsequent executions of this module, even though 
            // we haven't finished scanning this Derived relation
        }

        tuples_iterator =
            new corTupleIterator(tuples, iterator.arg_list, env, NULL, endmark);

        iterator.answer_iterator = tuples_iterator;
        tuples_iterator->state = iterator.state;
        
    }
    else {
        tuples_iterator = iterator.answer_iterator;
        tuples = (corStorageRelation *)(tuples_iterator->relation);
    }
    
    corTuple *tuple = tuples->get_next_tuple(*tuples_iterator);
    
    // Careful here - we are returning a different tuple iterator, and
    // must make sure all necessary arguments are copied over as required.

    while (tuples_iterator->no_match() && iterator.state) {
        iterator.reset_no_match();
        (*(iterator.state)->solver)((iterator.state)->md,
            (iterator.state)->magic, 0, &iterator);
        tuple = tuples->get_next_tuple(*tuples_iterator);
    }

    if (tuples_iterator->no_match()) {
        iterator.set_no_match();
        tuple = (corTuple *) NULL;
    }
    else {
        iterator.reset_no_match();
        iterator.tuple_env = tuples_iterator->tuple_env;
        tuples_iterator->tuple_env = NULL;
    }
    
    return(tuple);
}


/*****************************************************************************/
/*
 * WARNING:: This function may be called multiple times on an iterator.
 * It must make sure things are not deleted multiple times.
 */

void corDerivedRelation::release(corTupleIterator& iterator)
{
    if (iterator.answer_iterator)  {
        corTupleIterator *tuples_iterator = iterator.answer_iterator;
        corStorageRelation *tuples = (corStorageRelation *)tuples_iterator->relation;

        if (iterator.state) {
            // in the case of lazy evaluation, it is possible that
            // there may still be some saved state. this happens when
            // a cut is used to suddenly terminate a computation

            if (iterator.state->rm_Arr) delete iterator.state->rm_Arr;
            corcleanup_do_query(iterator.state->md, iterator.state->magic,
                iterator.state->table);
            delete iterator.state;
            iterator.state = NULL;
        }
        

        if (!iterator.moduledata->module_info.SaveModule) {
                //tuples->empty_relation(/*deleteTuples=*/ 1);
                correclaim_relation(tuples);
                tuples_iterator->relation = NULL;
        }
        else {
            // NOTE: Several calls can be setup on the same module 
            // within the evaluation of a single rule due to the 
            // occurrence of several body literals with the same 
            // predicate name.  Because of the nested-loops join
            // strategy, marks are setup and should be released
            // in stack-order, i.e., the last mark on the relation
            // should be released

            ASSERT(tuples->headMark.prev != & (tuples->headMark));
            // should free the mark used by tuples_iterator and get a 
            // mark corresponding to the new end of relation
            corRMark *endmark = tuples_iterator->end_mark;
            ((corHashSimpleRelation *)tuples)->freeMark(endmark);
            // grab a new last mark for the answer relation
            tuples->getMark();
        }
        delete tuples_iterator;
        iterator.answer_iterator = NULL;
        if (corexEnv->explain_fp && (corexEnv->GlobalRelationOptions & REL_EXPLAIN)){
            // Send a special breakpoint (and leave the pipe open)
            fputs("explain_break(2,X).\n", corexEnv->explain_fp);
            fflush(corexEnv->explain_fp);
        }
    }
}


/*****************************************************************************/
void corDerivedRelation::print_name(FILE *outf)
{
    corDerivedMethod *method;
    char buf[500];
    corVarLink *rest;

    for (method = methods; method ; method = method->next) {
        if (!method->valid_method) continue;

        sprintf(buf, "%s", corSymbolString(name));
        sprint_adornment(buf+strlen(buf), &(method->adornment));
        fprintf(outf, "%15s/%d    : (derived) ", buf, arity());
        fprintf(outf, ": indexes ");
        corPredAnnotations *cur;
        for (cur= method->md->index_annotations; cur; cur=cur->next()) {
            if ((cur->pred == corEnterSymbol(buf)) || (cur->pred == name)) {
                if (!cur->arglist2)
                    cur->arglist1->printon(outf);
                else {
                    rest = cur->arglist1->var_list(NULL);
                    rest = cur->arglist2->var_list(rest);
                    char *argindex = corsimplify_index(cur->arglist1,
                        corcount_var_list(rest),cur->arglist2);
                    if (argindex) 
                        fprintf(outf, "%s", argindex);
                    else {
                        fprintf(outf, "(");
                        cur->arglist1->printon(outf);
                        fprintf(outf, ")(");
                        cur->arglist2->printon(outf);            
                        fprintf(outf, ")");
                    }
                }
                fprintf(outf, " ");
            }
        }
        fprintf(outf, "\n");
    }
}

/*****************************************************************************/
void corDerivedRelation::print_facts(FILE *outf, corArgList *) const
{
    corDerivedMethod *meth = methods;
    while (meth) {
        if (meth->valid_method) {
            fprintf(outf, "module %s.\n", corSymbolString(name));
            fprintf(outf, "export %s[", corSymbolString(name));
            corprint_adornment_brack(meth->adornment, outf);
            fprintf(outf, "].\n");
            corprint_Annotations(meth->md->module_info, outf,
                meth->md->index_annotations, meth->md->agg_sel_annotations,
                NULL, meth->md->multiset_annotations,
                meth->md->prioritize_annotations,
                meth->md->diskrel_annotations);
            meth->md->ruleArray->printon(outf);
            fprintf(outf, "end_module.\n");
        }
        meth = meth->next ;
    }
}


/*****************************************************************************/
int corDerivedRelation::insert_new(corArgList& , corBindEnv*, corBindEnv*, corTuple* , int)
{
    CORAL_error(COR_BAD_TUPLE_INSERT, name->string(),"insert_new");
    return -1;
}

/*****************************************************************************/
int cororOptionToRelation(corRelation *rel, char *adorn_form, RelationOptions coropt)
{
    if ((rel->r_kind != COR_R_DERIVED) && (rel->r_kind != COR_R_ANSWER)) {
        fprintf(corexEnv->error_file,
            "CORAL::Error: Relation is not a derived relation\n");
        return 0;
    }

    corBitVector* bv = new corBitVector(rel->arity());
    
    for (int i = 0; i < rel->arity(); i++, adorn_form++) {
        if (*adorn_form == 0) {
            fprintf(corexEnv->error_file, "CORAL::Error: adornment too short");
            fprintf(corexEnv->error_file, "    --- arity of %s=%d, adorn len=%d)\n",
                rel->name->string(), rel->arity(), i);
            delete bv;
            return 0;
        }
        if (*adorn_form == 'b' || *adorn_form == '1')
            bv->set(i);
    }

    if (*adorn_form != 0) {
        fprintf(corexEnv->error_file, "CORAL::Error: adornment too long");
        fprintf(corexEnv->error_file, "    --- arity of %s=%d, adorn len=%d)\n",
            rel->name->string(), rel->arity(), i); 
        delete bv;
        return 0;
    }

    corDerivedMethod *meth = ((corDerivedRelation *)rel)->methods;
    while (meth) {
        if (meth->valid_method && (meth->adornment==*bv)) {
            meth->md->rel_options |= coropt;
            delete bv;
            return 1;
        }
        meth = meth->next ;
    }
    delete bv;
    fprintf(corexEnv->error_file,
        "CORAL::Error: cannot find the adorned relation\n");
    return 1;
}

/*****************************************************************************/
int corandOptionToRelation(corRelation *rel, char *adorn_form, RelationOptions coropt)
{
    if ((rel->r_kind != COR_R_DERIVED) && (rel->r_kind != COR_R_ANSWER)) {
        fprintf(corexEnv->error_file,
            "CORAL::Error: Relation is not a derived relation\n");
        return 0;
    }

    corBitVector* bv = new corBitVector(rel->arity());
    
    for(int i = 0; i < rel->arity(); i++, adorn_form++) {
        if(*adorn_form == 0) {
            fprintf(corexEnv->error_file, "CORAL::Error: adornment too short");
            fprintf(corexEnv->error_file, "    --- arity of %s=%d, adorn len=%d)\n",
                rel->name->string(), rel->arity(), i);
            delete bv;
            return 0;
        }
        if (*adorn_form == 'b' || *adorn_form == '1')
            bv->set(i);
    }

    if (*adorn_form != 0) {
        fprintf(corexEnv->error_file, "CORAL::Error: adornment too long");
        fprintf(corexEnv->error_file, "    --- arity of %s=%d, adorn len=%d)\n",
            rel->name->string(), rel->arity(), i); 
        delete bv;
        return 0;
    }

    corDerivedMethod *meth = ((corDerivedRelation *)rel)->methods;
    while (meth) {
        if (meth->valid_method && (meth->adornment==*bv)) {
            meth->md->rel_options &= coropt;
            delete bv;
            return 1;
        }
        meth = meth->next ;
    }
    delete bv;
    fprintf(corexEnv->error_file,
        "CORAL::Error: cannot find the adorned relation\n");
    return 1;
}

/*****************************************************************************/


