/*****************************************************************************/
//    Relation.C
/*****************************************************************************/

#include "Relation.h"

#include "Globals.h"
#include "Interface.h"
#include "RelationMisc.h"
#include "StackMark.h"
#include "TupleIterator.h"
#include "PrioritizeInfo.h"
#include "BuiltinSymbols.h"
#include "HashSimpleRelation.h"
#include "VersionedBindEnv.h"
#include "CoralError.h"
#include "Unify.h"
#include "Execute.h"

/*****************************************************************************/

int corC_cont_id = 1;                //    Used for ReturnUnification

/*****************************************************************************/
int corRelation::isConstant()
{
    return 1;
}

/*****************************************************************************/
corRelation::corRelation() :corArg()
{
    char namep[100];
    sprintf(namep,"Rel_%d", corRelCount++);
    name = corEnterSymbol(namep);
    _arity = 0;
    _kindof = COR_RELATION;

    local_options = 0;
    count = 0;
    scan_count = 0;
    ref_count = 1 ;
    single_success = 0;
    has_release = 1;

    check_subsum = corexEnv->C_check_subsum_default;

    r_kind = COR_R_ANSWER;  // default

    delta_indexed = corexEnv->C_index_deltas_default ;
}

/*****************************************************************************/
corRelation::corRelation(int new_arity, int delta)
{
    char namep[100];
    sprintf(namep,"Rel_%d", corRelCount++);
    name = corEnterSymbol(namep);
    _arity = new_arity;
    _kindof = COR_RELATION;

    local_options = 0;
    count = 0;
    scan_count = 0;
    ref_count = 1 ;
    single_success = 0;
    check_subsum = corexEnv->C_check_subsum_default;
    has_release = 1;

    r_kind = COR_R_ANSWER;    // default

    if (delta == -1)
        delta_indexed = corexEnv->C_index_deltas_default ;
    else delta_indexed = delta;


    agg_sel_info = NULL;
    priority_info = NULL;

    _hash = UnknownHashValue;
    _eq_id = UnknownHashValue;

    ordsearch = 0;
    for_ordsearch = NULL;
    mp_done_reln = NULL;
}

/*****************************************************************************/
void corRelation::print(corBindEnv* env, FILE *file, char *) const
{
    this->print(env, file);
}

/*****************************************************************************/
void corRelation::print(corBindEnv*, FILE *file) const
{
    fprintf(file, "{");
    printon(file);
    fprintf(file, "}");
}

/*****************************************************************************/
void corRelation::sprint(char* str, int* pos, corBindEnv *env) const
{
    sprintf(str    + *pos, "{");
    (*pos)++    ;

    corStackMark    stack_mark0;
    corArrayBindEnv    env0(arity());
    corArgList    *args = cormake_vararglist(arity());
    corTupleIterator    iter((corRelation*)this, *args, &env0);
    corStackMark    stack_mark1;
    corTerm    term1;

    corBindEnv    *env1 = iter.get_next();
    if    (!(iter.no_match())) {
        for (int iarg = 0; iarg < arity(); iarg++) {
            if (iarg > 0) { sprintf(str+ *pos, ", ");(*pos) += 2;}
            term1 = env1->lookup(iarg);
            if (term1.expr.IsEmpty()==0) term1.expr.sprint(str, pos, env); 
        }
        for (;;) {
            corBindEnv *env1 = iter.get_next();
            if (iter.no_match()) break;
            sprintf(str + *pos, ",");
            (*pos)++ ;
            for (int iarg = 0; iarg < arity(); iarg++) {
                if (iarg > 0) { sprintf(str+ *pos, ", ");(*pos) += 2;}
                term1 = env1->lookup(iarg);
                if (term1.expr.IsEmpty()==0) term1.expr.sprint(str, pos, env); 
            }
            stack_mark1.pop_to();
        }
    }
    stack_mark0.pop_to();
    delete args;

    sprintf(str    + *pos, "}");
    (*pos)++;
}

/*****************************************************************************/
void corRelation::print_name( FILE *outf)
{
    fprintf(outf, "%15s/%d\t : (generic)\n", corSymbolString(name), arity()) ;
}

/*****************************************************************************/
int corRelation::add_hidden_relation(corPrioritizeInfo *) 
{
    fprintf(corexEnv->trace_file,
        "System error: add_hidden_relation not implemented for some %s\n",
        "relation types\n");
    return(0);
}

/*****************************************************************************/
int corRelation::pop_from_hidden_relation() 
{
    fprintf(corexEnv->trace_file,
        "System error: pop_from_hidden_relation not implemented for some %s\n",
        "relation types\n");
    return(0);
}

/*****************************************************************************/
int corRelation::pop_all_from_hidden_relation() 
{
    fprintf(corexEnv->trace_file, "System error: pop_all_from_hidden_relation not implemented for some %s\n","relation types\n");
    return(0);
}

/*****************************************************************************/
int corRelation::insert_in_hidden_relation(corTuple *) 
{
    fprintf(corexEnv->trace_file,
        "System error: insert_in_hidden_relation not implemented for some %s\n",
        "relation types\n");
    return(0);
}

/*****************************************************************************/
void corRelation::add_prioritize_info(corPredAnnotations *anno) 
{
    corPrioritizeInfo *newinfo = new corPrioritizeInfo();

    newinfo->next = NULL;

    if (anno->op_arg.kindof() != COR_FUNCTOR ||
        ((anno->op_arg.FuncArgP())->arity() != 1))
    {
        corannotation_error(anno, "Bad aggregation type");
        return;
    }

    corFuncArg *fop_arg =  anno->op_arg.FuncArgP();

    if (fop_arg->functor() == corMaxOpSymbol)
        newinfo->order = 1;
    else if (fop_arg->functor() == corMinOpSymbol)
        newinfo->order = -1;
    else { // possibly extend to irrelevant fact annotation
        corannotation_error(anno,
            "Prioritize: bad ordering - only min and max allowed ");
        return;
    }

    if (fop_arg->args[0].kindof() != COR_VARIABLE) {
        corannotation_error(anno, "Prioritize argument should be a variable");
        return;
    }

    corVarArg *order_var = fop_arg->args[0].VarArgP();

    newinfo->argnum = -1;

    for (int i=0; i < anno->arglist1->arity(); i++) {
        if ((*anno->arglist1)[i].kindof() != COR_VARIABLE) {
            corannotation_error(anno,
                "Arglist of prioritize should contain only variables");
            return;
        }
        corVarArg *listvar = (*anno->arglist1)[i].VarArgP();
        if (listvar->var == order_var->var) {
            if (newinfo->argnum == -1) {
                newinfo->argnum = i;
            }
            else {
                corannotation_error(anno,
                    "Prioritize: repeated ordering variable in argument list");
            }
        }
    }
    newinfo->next = priority_info;
    priority_info = newinfo;
    add_hidden_relation(newinfo);
}

/*****************************************************************************/
/*
    Inserts a tuple into the relation.
    Checks for subsumption depending on this->check_subsum.

    If dont_rename is set, it does not perform renaming,
    else it renames (args, env);  variables in dont_rename_env
    are not renamed.
    The (renamed) (args,env) is stored in (copy_args, copy_env)

    If OrderedSearch is used, special things are done, and 
    (copy_args, copy_env) (with additional fields) is inserted into context.
    Otherwise (copy_args, copy_env) is inserted into the relation.

    Note: Variables in dont_rename_env are not renamed
*/

int corStorageRelation::insert_new(corArgList& args, corBindEnv* tupenv,
    corBindEnv *dont_rename_env, corTuple *parent_goal, int dont_rename)
{
    corStackMark stack_mark0;
    corBindEnv *env;
    
    if (corexEnv->C_interrupt_raised) return 0;    

#    ifdef DO_TRACE
        if (DISPLAY_INSERTIONS) {
            fprintf(corexEnv->trace_file,
                "[Starting to insert %s(", corSymbolString(name));
            args.print(tupenv, corexEnv->trace_file);
            fprintf(corexEnv->trace_file, ")");
        }
#    endif

    if (args.arity() != arity()) {
        fprintf(corexEnv->trace_file,
            "Relation::insert_new - attempt to insert tuple of arity %d \n",
            args.arity());
        fprintf(corexEnv->trace_file,
            "                                 - in relation %s of arity %d \n",
            name->string(), arity() );
        return 0;
    }

    if (corhas_grouping_args(args)) {
        return insert_grouping(args,tupenv,dont_rename_env, parent_goal,
            dont_rename);
    }

    if (tupenv==NULL) {
        int env_size = args.max_vars();

        if (env_size != 0) {
            //    WARNING - this space is never reclaimed.  Make sure insert_new
            //    is not called with a null bindenv
            env = new corArrayBindEnv(args.max_vars());
        }
        else {
            env = NULL;
        }
    }
    else {
        env = tupenv;
    }

    // first some subsumption checks
    if (check_subsum==1) {
        corTuple *newtuple;

        // If non-ground facts are present, a new tuple needs to be created
        // before performing subsumption checks, otherwise the old tuple
        // itself can be used.
        newtuple = new corTuple(&args, env);

        if (!ordsearch) {
            // ordinary case
            if (is_subsumed(newtuple)) {
#                ifdef DO_TRACE
                    if (DISPLAY_INSERTIONS) {
                        fprintf(corexEnv->trace_file, " - subsumed]\n");
                    }
#                endif
                delete newtuple;
                return 0;
            }
        }
        else {
            //    OrdSearch: these are just preliminary checks
            //    subsumption against Context done later
            
            if (r_kind == COR_R_MAGIC) {
                // subsumption checking is first done against the mp_done_reln
                if (mp_done_reln && mp_done_reln->is_subsumed(newtuple)) {
#                    ifdef DO_TRACE
                        if (DISPLAY_INSERTIONS) {
                            fprintf(corexEnv->trace_file, " - subsumed]\n");
                            fprintf(corexEnv->trace_file,
                                "        OrdSearch: using mp_done\n");
                        }
#                    endif
                    delete newtuple;
                    return 0;
                }
            }
            else if (r_kind != COR_R_SUPPLEMENTARY) {
                // don't want any subsumption checking done if 
                // (ordsearch && (r_kind == COR_R_SUPPLEMENTARY))
                    
                if (is_subsumed(newtuple)) {
#                    ifdef DO_TRACE
                        if (DISPLAY_INSERTIONS) {
                            fprintf(corexEnv->trace_file, " - subsumed]\n");
                        }
#                    endif
                    delete newtuple;
                    return 0;
                }
            }
        }
        delete newtuple;
    }
             
    // now add the tuple to the appropriate relation 


    corGoalNode *g_node;
    int negated_g_node = 0;
    corTable * F;
    struct corClauseName *c_names;

    if (ordsearch && ((r_kind == COR_R_MAGIC) ||
        (r_kind == COR_R_SUPPLEMENTARY)) &&
        corT_Stack.count > 0)
    {
        F = corT_Stack.top();
        if (F && F->moduledata && F->moduledata->curr_rInfo &&
            F->moduledata->curr_rInfo->rule)
        {
            c_names = F->moduledata->curr_rInfo->rule->clause_names;
            for (; c_names; c_names = c_names->next) {
                negated_g_node = c_names->negated;
                if (negated_g_node) break;
            }
        }
    }

    //    For Ordered Search, the m_p and sup_ facts in Context are maintained
    //    as auxiliary relations "for_ordsearch" associated with the regular
    //    m_p and sup_ relations.    the "for_ordsearch" relation is used as 
    //    an index into context.
    if (ordsearch && ((r_kind == COR_R_MAGIC) ||
        (r_kind == COR_R_SUPPLEMENTARY)) && (!for_ordsearch))
    {
        for_ordsearch = new corHashSimpleRelation(arity());
        //    What other indices besides the default indices need to be
        //    created?
        //    WARNING: if linked relations are to be used, this may have to
        //    be modified
    }

    /*
        divesh: added stuff for optimizing insertion of facts by
        insert_new.    when there is a for_supp_index relation, then
        this is a magic relation, and a previous call to goal_id
        should have inserted this same fact in the for_supp_index 
        relation.    hence, instead of CopyArgs and deleting the 
        fact there, we just "move" the fact from the for_supp_index
        relation to the main relation.
    */
    //    WARNING::    ReturnUnify optimizations needed here!!!
    //    Check if this works correctly with non-ground magic facts.

    corArgList* copy_args;
    corBindEnv* copy_env = NULL;
    int env_size;

    if (for_supp_index) {
        // there is an auxiliary relation and the fact can be got from there
        corTupleIterator supp_iter(for_supp_index, args, env);
        corStackMark supp_stack_mark1;
        
        for (;;) {
            corStackMark supp_stack_mark2;
            corTuple *supp_tuple1 = supp_iter.get_next_tuple();
            supp_stack_mark2.pop_to();
            // Undo any bindings done by get_next_tuple()
            // WARNING: will work ONLY if tuples don't have bindenvs!!
            if (supp_tuple1 == NULL) {
                // should not happen
                fprintf(corexEnv->trace_file,
                    "insert_new: no matching fact in ");
                fprintf(corexEnv->trace_file, "for_supp_index relation.\n");
                stack_mark0.pop_to();
                return 0;
            }

            corArrayBindEnv supp_tuple_env(supp_tuple1->env_size);

            if (corsubsumes_arg_lists(supp_tuple1->args(), args,
                &supp_tuple_env, env))
            {
                // undo bindings due to subsumes_arg_lists
                supp_stack_mark1.pop_to(); 
                if (corsubsumes_arg_lists(args, supp_tuple1->args(), env,
                    NULL))
                { 
                    supp_stack_mark1.pop_to(); // undo bindings

                    // supp_tuple1 is the appropriate tuple that needs to
                    // be inserted, and no CopyArgs need be done

#                    ifdef DO_TRACE
                        if (DISPLAY_INSERTIONS | DISPLAY_NEW_INSERTIONS) {
                            if (!DISPLAY_INSERTIONS) {
                                fprintf(corexEnv->trace_file,
                                    "[Inserting %s(", corSymbolString(name));
                                supp_tuple1->
                                    args().print(NULL,corexEnv->trace_file);
                                fprintf(corexEnv->trace_file, ")");
                            }
                            if (supp_tuple1->env_size == 0) {
                                fprintf(corexEnv->trace_file,
                                    " - done (ground)]\n");
                            }
                            else {
                                fprintf(corexEnv->trace_file,
                                    " - done (env-size:%d)]\n",
                                    supp_tuple1->env_size);
                            }
                        }
#                    endif
                    if (ordsearch && ((r_kind == COR_R_MAGIC) ||
                        (r_kind == COR_R_SUPPLEMENTARY)))
                    {
                        // create a goal node with the appropriate fields
                        g_node = new corGoalNode(supp_tuple1->_args,
                            supp_tuple1->bindenv, this, negated_g_node);
                        g_node->env_size = supp_tuple1->env_size;

                        // NOTE: Now, the GoalNode has to be inserted into
                        // the for_ordsearch relation and also into
                        // context, and subsumption checks have to be 
                        // done there.
                    }
                    else {
                        // count++;
                        insert_tuple(supp_tuple1);
                        // now need to delete the tuple from for_supp_index
                        // delete_id_tuple(for_supp_index, supp_tuple1);
                        // NOTE: check if this will work for deletion!!
                        return 1;
                    }
                }
            }
            // not found the same tuple, so continue searching
            supp_stack_mark1.pop_to();
        }
    }
    else {
        // there is no auxiliary relation and CopyArgs needs to be called

        if (dont_rename) {
            copy_args = &args;
            copy_env = env;
            env_size = copy_env->max_var_number();
        }
        else {
            copy_args = new corArgList(arity());
            // CopyArgs also renames variables so that they are distinct,
            // and the bindenv need not be used to distinguish variables.
            env_size = corCopyArgs(copy_args, &args, env,
                arity(), dont_rename_env);
            if (dont_rename_env && dont_rename_env != &corUnusedEnv)
                copy_env = dont_rename_env->copy_shell();
            else if (env_size)
                copy_env = new corVersionedBindEnv(env_size);
            else
                copy_env = NULL;
            // delete deref_args;        
            // can reclaim space since deref_args isn't used after this
        }

#        ifdef DO_TRACE
            if (DISPLAY_NEW_INSERTIONS && !DISPLAY_INSERTIONS ) {
                fprintf(corexEnv->trace_file,
                    "[Inserting %s(", corSymbolString(name));
                copy_args->print(copy_env, corexEnv->trace_file);
                fprintf(corexEnv->trace_file, ")");
            }
#        endif

        if (ordsearch && ((r_kind == COR_R_MAGIC) ||
            (r_kind == COR_R_SUPPLEMENTARY)))
        {
            g_node = new corGoalNode(copy_args,copy_env,this,negated_g_node);
            g_node->env_size = env_size;
        }
        else {
            // do the regular stuff
            corTuple *tuple1;
            tuple1 = new corTuple(copy_args, copy_env);
            tuple1->env_size = env_size;

            // WARNING:: Will need to eventually create a version of the env
            // and bump up its size as required.

            tuple1->contid = corC_cont_id++; // See below.
            if (USING_RETURN_UNIFY) {
                // RETURN UNIFICATION
                //    We need to set the contid field if the relation is 
                //    or an interface magic relation.  (We do it for all
                //    relations, always, as a kind of backup, to ensure that 
                //    interface relations have a contid set up, even if
                //    set up from a module that does not implement return
                //    unification.

                if (r_kind == COR_R_MAGIC) {
                    //    parent_goal is a supplementary fact
                    //    WARNING::ReturnUnify optimizations needed here!!!
                    //    The following will have to change for 
                    //    tail-recursive literals.    They shouls inherit 
                    //    the parid, and not the cont-id.    Figure out
                    //    a way of finding if a goal is tail-recursive.
                    tuple1->parid = parent_goal->contid;
                }
                else {
                    tuple1->parid = parent_goal->parid;
                }
            }

            //count++;
            insert_tuple(tuple1);
#            ifdef DO_TRACE
                if (DISPLAY_INSERTIONS | DISPLAY_NEW_INSERTIONS) {
                    if (env_size == 0)
                        fprintf(corexEnv->trace_file, " - done (ground)]\n");
                    else
                        fprintf(corexEnv->trace_file,
                            " - done (env-size:%d)]\n", env_size);
                }
#            endif
            return 1;
        }
    }

    // finally, for ordered search relations a special kind of 
    // subsumption needs to be done on the basis of the position of
    // insertion of g_node in context

    // for the ordinary case, the tuple has been inserted previously
    // the program control reaches this point only with OrdSearch

    ASSERT((ordsearch && ((r_kind == COR_R_MAGIC) ||
        (r_kind == COR_R_SUPPLEMENTARY))));

    //    For ReturnUnification; see comments above
    g_node->contid = corC_cont_id++;

    if (USING_RETURN_UNIFY) {
        if (r_kind == COR_R_MAGIC)
            g_node->parid = parent_goal->contid;
        else
            g_node->parid = parent_goal->parid;
    }

    int stat = insert_in_context(g_node, (corGoalNode *) parent_goal);
#    ifdef DO_TRACE
        if (DISPLAY_INSERTIONS | DISPLAY_NEW_INSERTIONS) {
            if (env_size == 0)
                fprintf(corexEnv->trace_file, " - done (ground)]\n");
            else
                fprintf(corexEnv->trace_file,
                    " - done (env-size:%d)]\n", env_size);
        }
#    endif
    return stat;
}

/*****************************************************************************/
void corRelation::printon(FILE *file, char *) const
{
    this->printon(file);
}

/*****************************************************************************/
void corRelation::printon(FILE *file) const
{
    corStackMark stack_mark0;
    corArrayBindEnv env0(arity());
    corArgList *args = cormake_vararglist(arity());
    corTupleIterator iter((corRelation*)this, *args, &env0);
    corStackMark stack_mark1;

    corBindEnv *env1 = iter.get_next();
    if (!(iter.no_match())) {
        for (int iarg = 0; iarg < arity(); iarg++) {
            if (iarg > 0) fputs(", ", file);
            env1->lookup(iarg).printon(file);
        }
        for (;;) {
            corBindEnv *env1 = iter.get_next();
            if (iter.no_match()) break;
            fprintf(file, ",");
            for (int iarg = 0; iarg < arity(); iarg++) {
                if (iarg > 0) fputs(", ", file);
                env1->lookup(iarg).printon(file);
            }
            stack_mark1.pop_to();
        }
    }
    stack_mark0.pop_to();
    delete args;
}

/*****************************************************************************/
void corRelation::print_facts(FILE *file, corArgList *arglist) const
{
    corStackMark stack_mark0;
    corArgList *args;

    if (!arglist)
        args = cormake_vararglist(arity());
    else {
        if (arglist->arity() != arity()) {
            CORAL_error(COR_BAD_ARGLIST, NULL, "print_facts");
            fprintf(corexEnv->trace_file, "Bad arity %d - %d expected\n",
                arglist->arity(), arity());
            return;
        }
        args = arglist;
    }

    corArrayBindEnv env0(arity());
    corTupleIterator iter((corRelation*)this, *args, &env0);
    corStackMark stack_mark1;
    corTerm term;
    for (;;) {
        corBindEnv *env1 = iter.get_next();
        if (iter.no_match()) break;
        fprintf(file, "%s(", name->string());
        for (int iarg = 0; iarg < arity(); iarg++) {
            if (iarg > 0) fputs(",", file);
            term.expr = (*args)[iarg];
            term.bindenv = &env0;
            term = term.dereference();
            term.printon(file);
        }
        fprintf(file, ").\n");
        stack_mark1.pop_to();
    }
    stack_mark0.pop_to();
    if (!arglist)
        delete args;
}

/*****************************************************************************/
/*
    Hashing for nested sets
    
    We use two fields for comparing sets:
    HashVal Relation::_hash, and 
    HashVal Relation::_eq_id
    
    (1) _hash is computed using HashVal Relation::hash(BindEnv *)
        which computes a hash value for a relation based on the tuples
        in the relation.  It uses a commutative, associative hashing
        function (multiplication), as a result of which the hash value
        of a relation does not depend on the order in which tuples are 
        accessed for computing the hash value.
        This is used as a sufficient condition for INEQUALITY, i.e.,
        if two hash values are distinct, so are the relations.
        However, as with all hash values, two relations could be distinct
        and have the same hash value.
    
    (2) _eq_id is an "id" field which is used as a sufficient
        condition for equality, i.e. if two _eq_id fields are the
        same, the relations are the same. However, it may be the case that
        _eq_id of one relation is different from the _eq_id field of
        another relation, while the two relations are indeed equal.
        In which case, we physically compare the two relations, and
        if equal, set the _eq_id field of each to the lower of the two.
*/

/*****************************************************************************/
//    compute the hash value of a relation argument. 

HashVal corRelation::hash(corBindEnv * /* env is ignored */)
{
    HashVal h_accum = 1;
    HashVal h_temp;
    corTuple *T;

    if (_hash != UnknownHashValue) {
        // then set has been hashed, so
        return (_hash);
    }

    FOR_EACH_TUPLE(T, this) {
        h_temp = T->hash();     //  hash the element T

        if (h_temp == VarHashValue) {
            h_accum = VarHashValue;
            break;
        }
        else {
            h_accum = h_accum * h_temp;    
        }
    } END_EACH_TUPLE(T)

    _hash = h_accum;
    return(_hash);
}

/*****************************************************************************/
/*
    Cardinality of a tuple in a relation.
    INPUT: a pointer to a tuple, T
    OUTPUT: the number of occurrences of (*T) in (*this)
    NOTE:  We assume that T is ground, and all tuples of *this are ground.
    NOTE:  this could be made more efficient, by iterating only through the
        hash bucket for T, instead of all of the rel, should large sets
        be commonly used and too slow.  
*/

unsigned int corRelation::cardinality(corTuple* T)
{
#    ifdef DO_TRACE
        if (corexEnv->dbg_sets) {
            fprintf(corexEnv->trace_file,
                "Relation::cardinality ----- cardinality of:  ");
            T->printon(corexEnv->trace_file);
            fprintf(corexEnv->trace_file, "\n");
        }
#    endif

    HashVal h_T = T->hash();
    ASSERT((corisGroundHashValue(h_T)));

    corTuple *tuple;
    HashVal h_tuple;
    unsigned int count = 0;
  
    FOR_EACH_TUPLE(tuple, this) {
        h_tuple = tuple->hash();
        ASSERT((corisGroundHashValue(h_tuple)));

        if (corequal_tuples(tuple, T, /*display=*/ 0))
            count++;
    } END_EACH_TUPLE(tuple)

#    ifdef DO_TRACE
        if (corexEnv->dbg_sets) {
            fprintf(corexEnv->trace_file,
                "            cardinality is: %d.\n", count);
        }
#    endif

    return(count);
}

/*****************************************************************************/
//    return true iff (*this) is a subset of rel

int corRelation::contained_in(corRelation *rel)
{
#    ifdef DO_TRACE
        if (corexEnv->dbg_sets)
        fprintf(corexEnv->trace_file,
            "Relation::contained_in ----- check set containment.\n");
#    endif

    corTuple *tuple;
    int contain = 1;

    FOR_EACH_TUPLE(tuple,this) {
        if (this->cardinality(tuple) > rel->cardinality(tuple)) {
            contain = 0;
            break;
        }
    } END_EACH_TUPLE(tuple)
  
#    ifdef DO_TRACE
        if (corexEnv->dbg_sets) {
            if (contain) {
                fprintf(corexEnv->trace_file,
                    "       ----- Relation::contained_in SUCCEEDED.\n");
            }
            else {
                fprintf(corexEnv->trace_file, 
                    "       ----- Relation::contained_in FAILED.\n");
            }
        }
#    endif

    return(contain);
}

/*****************************************************************************/
/*
    Return true iff *rel = *this.
    Inputs can be assumed to be ground, since this is only called with
    nested relations for which this condition is satisfied.
*/

int corRelation::raw_equals(corRelation *rel)
{
#    ifdef DO_TRACE
        if (corexEnv->dbg_sets) {
            fprintf(corexEnv->trace_file,
                "Relation::raw_equals ----- executing set equality.\n");
        }
#    endif

    corTuple *tuple = NULL;
    int equals;

    // first check that the two relations have the same cardinality
    // or else do not proceed.
    if (this->count != rel->count)
        equals = 0;
    else {
        equals = 1;
        FOR_EACH_TUPLE(tuple,this) {
            if (this->cardinality(tuple) != rel->cardinality(tuple)) {
                equals = 0;
                break;
            }
        } END_EACH_TUPLE(tuple)
    }

#    ifdef DO_TRACE
        if (corexEnv->dbg_sets) {
            if (equals) {
                fprintf(corexEnv->trace_file, 
                    "        ----- Relation::raw_equals SUCCEEDED.\n");
            }
            else {
                fprintf(corexEnv->trace_file, 
                    "        ----- Relation::raw_equals FAILED.\n");
            }
        }
#    endif

    return equals;
}

/*****************************************************************************/

// Make sure this is not equal to UnknownHashValue and VarHashValue
static int C_eq_id_count = 1;

int corRelation::equals(corArg *arg2)
{
#    ifdef DO_TRACE
        if (corexEnv->dbg_sets) {
            fprintf(corexEnv->trace_file,
                "Relation::equals ------ executing set equality.\n");
        }
#    endif

    if (arg2->kindof() != COR_RELATION) return(0);

    corRelation *that = (corRelation*) arg2;

    HashVal h1 = this->hash();
    HashVal h2 = that->hash();

#    ifdef DO_TRACE
        if (corexEnv->dbg_sets) {
            fprintf(corexEnv->trace_file,"Set-1 = ");
            ((corArg *)this)->printon(corexEnv->trace_file);
            fprintf(corexEnv->trace_file,". Hash = %X.\n",h1);
            fprintf(corexEnv->trace_file,"Set-2 = ");
            arg2->printon(corexEnv->trace_file);
            fprintf(corexEnv->trace_file,". Hash = %X.\n\n",h2);
        }
#    endif

    if ( !corisGroundHashValue(h1) || !corisGroundHashValue(h2) ) {
        CORAL_error(COR_NESTED_SET_NOT_GROUND, NULL, "Relation::equals");
        return(-1);
    }

    if ( h1 != h2 ) {
#        ifdef DO_TRACE
            if (corexEnv->dbg_sets)
            fprintf(corexEnv->trace_file,"          ------- set equality FAILED.\n");
#        endif
        return 0;
    }

    if ( corisGroundHashValue(_eq_id) && (_eq_id == that->_eq_id) ) {
#        ifdef DO_TRACE
            if (corexEnv->dbg_sets) {
                fprintf(corexEnv->trace_file, "          ground id values exist.\n");
                fprintf(corexEnv->trace_file, 
                    "          ------- set equality SUCCEEDED.\n");
            }
#        endif
        return 1;
    }

    int rel_equals = /*this->*/ raw_equals(that); 
    
    if ( !rel_equals) {
#        ifdef DO_TRACE
            if (corexEnv->dbg_sets) {
                fprintf(corexEnv->trace_file, 
                    "          ------- set equality FAILED.\n");
            }
#        endif
        return 0;
    }

    //    The sets are equal, so assign them the same id

    if ( corisGroundHashValue(_eq_id) && corisGroundHashValue(that->_eq_id) ) {
        HashVal min_val = (that->_eq_id < _eq_id)? that->_eq_id : _eq_id;
        _eq_id = that->_eq_id = min_val;
    }
    else if ( ( ! corisGroundHashValue(_eq_id)) 
           && ( ! corisGroundHashValue(that->_eq_id) ) )
    {
        _eq_id = that->_eq_id = C_eq_id_count++;
    }
    else {
        // exactly one of them has been previously assigned 
        int id_val = corisGroundHashValue(_eq_id) ? _eq_id : that->_eq_id;
        _eq_id = that->_eq_id = id_val;
    }

#    ifdef DO_TRACE
        if (corexEnv->dbg_sets) {
            fprintf(corexEnv->trace_file, 
                "           assigning equal id values: %d.\n", _eq_id);
            fprintf(corexEnv->trace_file, 
                "           ------- set equality SUCCEEDED.\n");
        }
#    endif

    return 1;
}

/*****************************************************************************/

