/*****************************************************************************/
//    Predicate.h
/*****************************************************************************/

#ifndef _Predicate_H_
#define _Predicate_H_

#include "SymbolArg.h"
#include "Scc.h"
#include "SccWork.h"
#include "Table.h"
#include "Compile.h"

/*****************************************************************************/

struct corSccWork;
struct corScc;

class Rule;

/*****************************************************************************/

corPredicate *corAllocPredicate(corParserStruct& corparserStruct, Name name, int arity);
void corAddClauseToPredicate(corPredicate *predicate, corClause *clause);

int coris_magic(Name name);
int coris_mp_done(Name name);
int coris_supp(Name name);

/*****************************************************************************/
// A Predicate is identified by its name and arity. The other fields
// are used by various pieces of code that manipulate Predicates.

struct corPredicate {
    Name name;
    int arity;
    corPredicate *next;            // link to next predicate in the same Scc 
    corPredicate *symtabnext;        // link to next predicate of same name 
                                // in the symbol table
    corPredicate *allprednext;        // link to next predicate in AllPredicates list
    corClause *clauses;
    corClause **clause_tail;
    struct corScc *scc;

    struct corSccWork *work;        // temporaries used for Scc generation 
    char mark;                    //  ''

    int frame_offset;            // Set and used during evaluation. 
    int ext_magic;   // Added by Ken:true if clauses point to an ext magic rule
    inline int coris_magic()    { return ::coris_magic(name); }
    inline int coris_mp_done()    { return ::coris_mp_done(name); }
        inline int coris_supp() { return ::coris_supp(name); }

};

/*****************************************************************************/

#ifndef OFFSET_HACK
    inline int corget_offset(corTable *F, corPredicate *pred, int print_error = 1)
    {
        return corget_offset(F, pred->name, pred->arity, print_error) ;
    }
#else
    //    WARNING: This code should only be called after predicates have been
    //    inserted into the frame table.

    inline int corget_offset(corTable *F, corPredicate *pred, int print_error = 1)
    {
        ASSERT((pred->frame_offset >= 0) && (pred->frame_offset < F->numRels));
        return pred->frame_offset;
    }
#endif

/*****************************************************************************/

#endif _Predicate_H_

