/*****************************************************************************/
//    SemiNaive.h
/*****************************************************************************/

#ifndef _SemiNaive_H_
#define _SemiNaive_H_

//#include "SccInfo.h"
#include "RuleInfo.h"
#include "Rule.h"

/*****************************************************************************/

enum ModuleKind { COR_MATERIALIZED, COR_PIPELINED, COR_ORDSEARCH,};

/*****************************************************************************/
//    Codependent classes

class corSccInfo;
class corTable;
class corRMarkArray;
class corSuspendContext;

/*****************************************************************************/
// Structure that represents each seminaive rule.
// Each seminaive rule contains the rule from which it is derived
// and also an implicit mechanism of specifying 'old' and 'new'
// relations for literals on the right hand side. This mechanism
// is via the markOffsets array which contains for each rhs literal
// a pair of offsets into an array of RMarks created at runtime.
// Pointers to NULL RMarks (offset of -1) signify that the predicate 
// is base w.r.t the seminaive rule, while for predicates that are recursive
// w.r.t. the seminaive rule, the markOffsets array entries point to RMarks
// that specify the portion of the concerned relation (old or new)
// that is to be considered. The successBacktrack and failureBacktrack
// arrays are used for intelligent backtracking.
/*****************************************************************************/

struct corSemiNaive {
public:
    corSccInfo* scc;            // scc to which it belongs
    ModuleKind module_kind;    // Kind of module to which it belongs
                            // Either COR_ORDSEARCH or COR_MATERIALIZED
    corRuleInfo* rInfo;
    int* negated;            // This array specifies whether rInfo->relations[i]
                            // is negated or not.
                            // The size of the array is rInfo->num_literals

    int first;                // This field is generalised by predOrders
                            // and should be removed later.

    int* markOffsets;        // Size of this array is rInfo->num_literals * 2
                            // This array contains offsets into an array of
                            // RMarks that is declared for each Scc at runtime.
                            // A value of -1 implies a NULL RMark ;

    int* predOrders;        // Size of this array is rInfo->num_literals
    int* indices;            // Specifies for each literal the offset of the
                            // index that will be tried first for the literal.
                            //  Size of this array is rInfo->num_literals

    int successBacktrack;    // Used for intelligent backtracking and choice.
    int* firstFailureBacktrack;
                            //  Size of this array is rInfo->num_literals
                            //  Used for get-first intelligent backtracking.
    int* nextFailureBacktrack;
                            //  Size of this array is rInfo->num_literals
                            //  Used for get-first intelligent backtracking.

    int* btDecision[2];        // used to make backtracking decision

    int recursive;
    int ignore_this_rule;
                            // if this flag is set, (when the execution
                            // mode is supplementary magic with indexing)
                            // then this rule is ignored for evaluation.
                            // It is however used for scc analysis, etc.
                            // This flag is set only for rules of the
                            // form m_p() :- sup_i().

    int rule_id; // number of this rule in the sn_rules array of the Scc
    char* label; // label derived from the original rule: added by Ken
    
    corSemiNaive() ;
   ~corSemiNaive();
    void Add_Rule_Info(corRuleInfo *, corClause *, int);
    void clean_cache();
    int evaluate(corTable *F, corRMarkArray *rm_array = NULL, 
             corSuspendContext *s = NULL, int allow_sus =0);
    void print(FILE *fd);
    void Add_label();
};

/*****************************************************************************/

#endif _SemiNaive_H_

