// SemiNaiveEval.C
// evaluating a rule, seperated from original RuleInfo.C -8/2/94

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "RuleInfo.h"

#include "ArgList.h"
#include "StackMark.h"
#include "Config.h"
#include "Rule.h"
#include "Table.h"
#include "RMarkArray.h"
#include "TupleIterator.h"
#include "RelationList.h"
#include "SccInfo.h"
#include "VersionedBindEnv.h"
#include "Globals.h"
#include "Literal.h"
#include "Execute.h"
#include "CoralBreak.h"
#include "Suspend.h"
#include "Suspension.h"
 
#include "DatabaseStruct.h"
#include "Interface.h"

extern corDatabaseStruct *corCurDB ;

#ifdef DO_EXPLAIN
#ifdef DO_PROFILE
#define HANDLE_SUCCESS { \
    if ((rel->local_options & REL_EXPLAIN) && (corT_Stack.count > 0))\
       cordo_dump(F, rInfo, env, &(rInfo->arg_list[index])); \
        (pinfo->i_newfacts)++ ; \
    }
#else
#define HANDLE_SUCCESS  { \
    if ((rel->local_options & REL_EXPLAIN) && (corT_Stack.count > 0)) \
       cordo_dump(F, rInfo, env, &(rInfo->arg_list[index])); \
  }
#endif
#else
#ifdef DO_PROFILE
#define HANDLE_SUCCESS  { (pinfo->i_newfacts)++ ; }
#else
#define HANDLE_SUCCESS  { }
#endif
#endif

/*****************************************************************************/
// the INSERT_NG_FACT macro is actually a blob of code that
// is moved here as a macro to enhance readablity of the rest of the code.
// While this admittedly yucky, it makes it possible to understand the rest
// of the code : Praveen

#define INSERT_NG_FACT  do { \
      corTermLink *renamed_vars = NULL; \
      corArgList  new_arglist = rInfo->arg_list[index]; \
      corArgList *new_arglist1 = NULL; /* hack! */ \
      corBindEnv *new_env = env; \
      int env_size; \
      int dont_rename = 0; \
      corBindEnv *backup_env = NULL; \
      int single_fact_env=1; \
      corStackMark *backup_env_mark; \
      corBindEnv *dont_rename_env = &corUnusedEnv; \
              \
      backup_env_mark = new corStackMark; \
          \
      for (int i=0; i<rInfo_num_literals; i++) { \
        if ((s->tuples)[i] && (s->tuple_its)[i]->tuple_env  \
        && (s->tuple_its)[i]->tuple_env->is_versioned()) { \
          dont_rename_env = (s->tuple_its)[i]->tuple_env; \
          break; \
        } \
      } \
      \
      if (dont_rename_env != &corUnusedEnv) { \
        backup_env = dont_rename_env->make_version_and_swap(); \
      }\
      \
      for (int j=0; j < rInfo_num_literals; j++) {\
        if( (! (s->tuples)[j]) || (s->tuple_its)[j]->tuple_env == dont_rename_env \
           || (s->tuple_its)[j]->tuple_env == NULL) \
          continue; \
        else single_fact_env = 0; \
      } \
      \
      if (single_fact_env && dont_rename_env != &corUnusedEnv && \
         !corhas_grouping_args(rInfo->arg_list[index])) { \
          \
          if( corbackpatch_bindenvs((s->stackmark0), dont_rename_env, \
                     env, renamed_vars) && (rel->count >= 0)) { \
            \
        dont_rename = 1; \
        new_arglist1 = new corArgList(new_arglist.arity());\
        env_size = corSimpleCopyArgs( new_arglist1, \
                      &rInfo->arg_list[index], \
                      env, dont_rename_env, renamed_vars); \
        if (dont_rename_env != &corUnusedEnv) { \
          new_env = dont_rename_env->copy_shell(); \
        } \
        else new_env = (corBindEnv*) new corVersionedBindEnv(env_size); \
          }\
      } \
      if ( rel->insert_new(new_arglist1 ? *new_arglist1 : new_arglist,new_env,\
                   dont_rename_env, (s->tuples)[0], dont_rename)){ \
        HANDLE_SUCCESS;\
        s->answer_count++; \
      } \
      backup_env_mark->pop_to(); \
      delete backup_env_mark; \
      \
      if (backup_env) { \
        dont_rename_env->restore(backup_env); \
        ::delete backup_env; \
      } \
    } while(0)

#define MARKLIST(X) ((markOffsets[X]<0) ? NULL : \
            rm_array->markList[markOffsets[X]])


#ifdef    DO_PROFILE
static long profile_rule = 0;
#endif

extern  void cordo_dump(corTable * F, corRuleInfo * rInfo, corBindEnv * new_env,
    corArgList* new_arglist);
extern int corbackpatch_bindenvs(corStackMark start, corBindEnv *dont_rename_env,
        corBindEnv *rule_env, corTermLink *&renamed_vars);

extern int corSimpleCopyArgs(corArgList *new_arglist, corArgList *old_arglist,
        corBindEnv *rule_env, corBindEnv *dont_rename_env,
        corTermLink *&renamed_vars);

//*************************************************************************
/*
    Evaluates a nested loop join of the seminaive rule. The Table structure
    F is defined in interp.h. It is a structure containing pointers to the
    various relations that the seminaive rule will have to access. It is
    initialized at query execution time with pointers to current relations
    and passed to the evaluate method as an argument. The evaluate function
    accesses all relations via the level of indirection provided by the Table.
    rm_array is an array of RMarks that specify the current extents of the
    'old' and 'new' relations used for seminaive evaluation. The evaluate
    function accesses the RMarks via offsets into the rm_array specified in
    the markOffsets array structure (class variable). Look at the code to 
    evaluate an Scc in interp.C to understand how RMarks are manipulated.

    The index 'current' counts how far we are along in the join.  It
    represents the position of the current rhs literal being processed.

    The function returns the number of new tuples generated.
*/
//*************************************************************************

int corSemiNaive::evaluate(corTable *F, corRMarkArray *rm_array,
            corSuspendContext *s, int allow_sus )
                // if called by scc, allow_sus= 0;
                // if called by exec_single_rule, =1;
{
  //temp variables;

  int index;
  int offset;
  corRelation *rel;
  corArrayBindEnv *env;
  corTupleIterator *tuple_its_index;
  corBindEnv **it_bindenvs = NULL;
#ifdef DO_PROFILE
  corProfileInfo *pinfo;
#endif
  int delete_s = 0;

  // for efficiency, make the following assignments
  int rInfo_num_literals = rInfo->num_literals;
  //int rInfo_num_vars = rInfo->num_vars;
  int *rInfo_relations = rInfo->relations;
  int UsingRetUnify = USING_RETURN_UNIFY;

  if (!s) {
    s = new corSuspendContext;
    delete_s = 1;
#ifdef KEN_DEBUG
    fprintf(corexEnv->error_file, " start sn evaluation\n");
#endif

    //initialization

    s->current = 0;
    s->answer_count = 0; // Count of number of new tuples generated.
    s->backtrack_index = 0; // used to specify next backtrack point 
 
    // Quick check for empty first relation. A relation is empty if
    // both the RMarks for the relation point to the same place.
    // The RMarks specify the extent of the relation to be scanned
    // and if they both are equal, the extent between them is empty.

    if (recursive && (rInfo->num_literals > 0)) {
      int marklist_index = 2*first;
      int tmp_mark_offset;
      
      corRMark * first_mark1 = (((tmp_mark_offset = markOffsets[marklist_index])
                  < 0) ? NULL : rm_array->markList[tmp_mark_offset]) ;
      if (first_mark1) {
        if (first_mark1->rel_p) {
          if (!first_mark1->rel_p->reln->count){
              if (delete_s)
                  delete s;
              return 0;
          }
        }
        else {
          if (!(*(first_mark1->lastptr))){
              if (delete_s)
                  delete s;
              return 0;
          }
        }
      }
    }
   
#    ifdef DO_PROFILE
    // Initialize profiling information
    pinfo = &(corCurrModuleInfo->corGlobalProfileInfo);
    pinfo->set_rule_name(rInfo->pInfo.rule);
    profile_rule = ((corexEnv->GlobalRelationOptions |
             (F->rels[rInfo_relations[rInfo_num_literals]].relPtr)->
             local_options) & REL_PROFILE_STATS);
    if (profile_rule) pinfo->init_rule();
#    endif

    //StackMark (s->stackmark0);

    //    StackMark *stack_marks;

    if (rule_id < 0) {
      // imperative rule
      (s->rule_env) = new corRuleEnv(this);
    }
    else {
      (s->rule_env) = F->module_env->scc_envs[this->scc->scc_id]->
                  snrule_envs[this->rule_id];
    }

    //stack_marks = (s->rule_env)->rmarks;
    (s->tuple_its)    = (s->rule_env)->tuple_its;
    //(s->arglist_its) = (s->rule_env)->arglist_its;
    (s->tuples) = (s->rule_env)->tuples;
    it_bindenvs = (s->rule_env)->bindenvs;
    env = & ((s->rule_env)->bindenv);

    if (rule_id < 0) {
      if (rm_array && rm_array->array_size) {
        for (index = 0; index < rInfo->num_literals; index++) {
          (s->tuple_its)[index] =
            new corTupleIterator(F->rels[rInfo->offsets[index]].relPtr,
                      rInfo->arg_list[index], env, MARKLIST(2*index),
                      MARKLIST(2*index+1), &(indices[index]));
        }
      }
      else {
        for (index = 0; index < rInfo->num_literals; index++) {
          (s->tuple_its)[index] =
            new corTupleIterator (F->rels[rInfo->offsets[index]].relPtr,
                       rInfo->arg_list[index],env, NULL, NULL,
                       &(indices[index]));
        }
      }
    }
    else {
      if (rm_array && rm_array->array_size) {
        for (index = 0; index < rInfo->num_literals; index++) {
          (s->tuple_its)[index]->
            reset_fields(F->rels[rInfo->offsets[index]].relPtr, env,
                 MARKLIST(2*index), MARKLIST(2*index+1),
                 &(indices[index]));
        }
      }
      else {
        for (index = 0; index < rInfo->num_literals; index++) {
          (s->tuple_its)[index]->
            reset_fields(F->rels[rInfo->offsets[index]].relPtr, env,
                 NULL, NULL, &(indices[index]));
        }
        // NOTE:this should be moved into reset_fields!
        //F->rels[rInfo->offsets[index]].relPtr->scan_count++;
      }
    }

    int (s->direction) = FORWARD;
  } else { 
#ifdef KEN_DEBUG
    fprintf(corexEnv->error_file, " resume sn evaluation SN::eval\n");
#endif
    //some other variables derived from the above;
    env = & ((s->rule_env)->bindenv);
    it_bindenvs = (s->rule_env)->bindenvs;
  }  
  // Nested Loops Join evaluation using backtracking on failure.
  
  for(; s->current >= 0;) {

#ifdef KEN_DEBUG
    fprintf(corexEnv->error_file," stack %p-%d\n", corCurTrail, corCurTrail->TrailSize);
#endif        
    // Exit statement from for loop. When current < 0, it implies that
    // no more facts can be generated from this rule evaluation, i.e.
    // it has backtracked past its starting point.

    // Check to see if rule evaluation has been succesful.
    // This happens when a valid tuple for the last rhs literal
    // has been found. When this happens, the head fact(s) should be
    // created.

    if ( s->current >= rInfo_num_literals ) {
      index = s->current;

      // The following code allows multiple heads per rule
      for( ; rInfo_relations[index] != -1; index ++) {
        offset = rInfo_relations[index];
        if (offset < 0) { 
          // A negative offset implies a negated literal
          // Eventually in such a case maybe we should delete the
          // tuple
          offset = (-2)-offset;
          rel = F->rels[offset].relPtr ;
          corTuple del_tup(&rInfo->arg_list[index], env);
          delete_tuple(rel, &del_tup);
          continue;
        }
    
        // Find the relation in which the head fact should be inserted.
        rel = F->rels[offset].relPtr ;
    
#       ifdef DO_PROFILE
        (pinfo->i_succesful)++ ;
#       endif
    
        if (!corCurrModuleInfo->NonGroundFacts) {
          // Insert the new head fact into the relation
          if (rel->insert_new( rInfo->arg_list[index], env,
                      &corUnusedEnv, (s->tuples)[0], 0))
            {
              HANDLE_SUCCESS;
              s->answer_count++;
            }
        } else {        /** NON_GROUND FACTS **/
          INSERT_NG_FACT;
        }
      }

      // Continue with the nested loops join
      s->current = rInfo_num_literals -1;
      s->backtrack_index = successBacktrack;            
      (s->direction) = BACKWARD;

      if (allow_sus && corsuspension.set>=0) { //suspended query
          if(corsuspension.set==1 || (corsuspension.set == 0 &&
              corsuspension.byte_count < 128)) { //suspend now

            s->answer = corcurUser->num_answers;

          // This needs to specify both the user and the connection
            s->conn_id = corcurConn->conn_id;    // who suspended it
            s->user_id = corcurUser->user_id;    

            s->name = corCurDB->name;
            s->table = F;
            s->semi_naive = this;
            s->trail = corCurTrail;       //current stack saved
            corCurTrail = new corTrail(200); //create new stack
            
            corsuspension.suspend(s);
            return s->answer_count;
          } else {
            continue;
          }
      }
      continue;
    } else {     /**    the entire rule has not been instantiated yet **/
      index = predOrders[s->current];
      tuple_its_index = (s->tuple_its)[index];

      // To do a scan of each relation on the rhs, a TupleIterator is
      // needed on that relation. If this is the first time the relation
      // is reached during the evaluation, a TupleIterator needs to be
      // created for it.

      if ((s->direction)) {        /**    FORWARD **/
        tuple_its_index->reset();
        s->backtrack_index = s->current;
      }
      
      if (UsingRetUnify) {
          if ((s->direction)) {
            //    Return unify modifies the iterator bindenv in some
            //    cases.    So the rest of the rule evaluation should
            //    use the new bindenv rather than the original iterator
            //    bindenv.
            //    The iterator's bindenv is reset when backtracking 
            //    to the first literal.
            if (index !=0)
              it_bindenvs[index] = (s->tuple_its)[index-1]->bindenv;
            else
              it_bindenvs[index] = env;
            tuple_its_index->bindenv = it_bindenvs[index];
          }
          else {
            // Restore bindenv as it was when 
            // the literal was reached in the forward
            // direction.
            tuple_its_index->bindenv = it_bindenvs[index];
          }
          
          if (s->current ==s->backtrack_index) {
            //    Set the following flag to indicate the literal
            //    in rule body that is being looked at; it is
            //    needed to implement return unification optimizations
      
            tuple_its_index->unification_number = s->current+1;
            
            //    Set the prev_tuple and the prev_bindenv fields.    These
            //    are distinct since prev_tuple is the original tuple
            //    with the original bindenv, while prev_bindenv is the
            //    versioned copy of that bindenv.
            int prev_index;
            
            if (s->current != 0) {
              prev_index = predOrders[s->current-1];
              tuple_its_index->prev_tuple = (s->tuples)[prev_index];
              tuple_its_index->prev_bindenv =
                (s->tuple_its)[prev_index]->tuple_env;
            }
            else {
              tuple_its_index->prev_tuple = NULL;
              tuple_its_index->prev_bindenv = NULL;
            }
          }
      }
      
      if (s->current ==s->backtrack_index) {
        //    Get the next tuple for the current relation in
        //    the nested loops join.
   
        (s->tuples)[index] = tuple_its_index->get_next_tuple();

        if (tuple_its_index->no_match() ^ negated[index]) {
          s->backtrack_index = btDecision[(s->direction)][s->current];
        } else {
          s->current++;
          (s->direction) = FORWARD;
          continue;
        }    
      }
      
      // else on intelligent backtracking failure or
      // on failure to find a valid tuple, backtrack to the previous
      // rhs literal.
      
      // Undo bindings BEFORE
      // deleting tuple_iterator (and the tuple_env in it).
      
      tuple_its_index->release(); /* Releases state of tuple iterator*/
      (s->tuples)[index] = NULL;
      
      // Decrement current so that the join evaluation backtracks to the
      // previous rhs relation, and a fresh tuple for that relation will
      // be fetched in the next iteration through the for loop.
      s->current--;
      (s->direction) = BACKWARD;
    }
  }

  // Cleaning up at the end of the nested loops rule evaluation.

  if (rule_id < 0) {
    for (index = 0; index < rInfo_num_literals; index++) {
      delete (s->tuple_its)[index];
      (s->tuple_its)[index] = NULL;
    }
  }
  (s->stackmark0).pop_to();

#    ifdef DO_PROFILE
  if (profile_rule) {
    pinfo->exit_rule();
    _PROFILE(rInfo->pInfo, predOrders, rInfo_num_literals) ;
  }
#    endif
  
  if (rule_id < 0) delete (s->rule_env);

  // Set up break points for rules
  if (scc) {
    int outcome =
      strcmp(corb.module_name,(scc->module->module_info.name)->string());

    if ((!outcome) && (corb.rule_flags[scc->scc_id])) {
      fprintf(corexEnv->error_file,"rule number is %d\n",rInfo->rule->rule_number);
      if ((corb.rule_flags[scc->scc_id])->test(rInfo->rule->rule_number+1)) {
          coral_break(corRule_String, rInfo->rule->rule_number+1);
      }
    } else if ((!outcome) && (corb.all)) {
      coral_break(corRule_String, rInfo->rule->rule_number+1);
    }
  }

  int ret = s->answer_count;
  if (delete_s)
      delete s;
  return(ret);
}

//*************************************************************************
